#include "config.h"

#include "Flat_Matrix.h"
#include "MPI_Events.h"

#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include <math.h>
#include <assert.h>
#include <stdlib.h>

#if HAVE_OMPT_H
#include "OMPT_Events.h"
#endif /* HAVE_OMPT_H */

#if HAVE_MPI_T
#include "MPIT_Events.h"
#include "MALP_MPI_T_data.h"
#endif /* HAVE_MPI_T */

/*
 * This source code is a mashup of Time_Matrix therefore
 * if you encounter any bugs they might be there too... 
 */

static void
Flat_Matrix_init_coll_MALP_EVENT_MPI(struct Flat_Matrix *flat_matrix, const uint64_t time_slices, const uint64_t num_slots)
{
    int i, j;
    assert(flat_matrix);

    for (i = 0; i < T_MPI_COUNT; i++)
    {
        for (j = 0; j < FLAT_GRID_COUNT; j++)
        {
            Sample_Grid_init(&(flat_matrix->mpi_grids[i][j]), time_slices, num_slots);
        }
    }
}

static void
Flat_Matrix_init_coll_MALP_EVENT_WRAPPED(struct Flat_Matrix *flat_matrix, const uint64_t time_slices, const uint64_t num_slots)
{
    int i, j;
    assert(flat_matrix);

    for (i = 0; i < T_POSIX_COUNT; i++)
        for (j = 0; j < FLAT_GRID_COUNT; j++)
            Sample_Grid_init(&(flat_matrix->posix_grids[i][j]), time_slices, num_slots);
}

static void
Flat_Matrix_init_coll_MALP_EVENT_OMPT(struct Flat_Matrix *flat_matrix, const uint64_t time_slices, const uint64_t num_slots)
{
#if HAVE_OMPT_H
    Sample_2D_Grid_t *new_2D_grids;
    assert(flat_matrix);

    new_2D_grids = Sample_2D_Grid_alloc(T_OMPT_COUNT, FLAT_GRID_COUNT);
    assert(new_2D_grids);
    flat_matrix->ompt_grids = new_2D_grids;
    Sample_2D_Grid_init(new_2D_grids, time_slices, num_slots);
#endif /* HAVE_OMPT_H */
}

static void
Flat_Matrix_init_coll_MALP_EVENT_MPI_T(struct Flat_Matrix *flat_matrix, const uint64_t time_slices, const uint64_t num_slots)
{
#if HAVE_MPI_T
    const int __cvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_CVAR);

    flat_matrix->mpi_t_cvar_grids = malloc(sizeof(struct Sample_Grid) * __cvar_count);

    int i;
    for (i = 0; i < __cvar_count; i++)
    {
        Sample_Grid_init(&(flat_matrix->mpi_t_cvar_grids[i]), time_slices, num_slots);
    }

    const int __pvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_PVAR);

    flat_matrix->mpi_t_pvar_grids = malloc(sizeof(struct Sample_Grid) * __pvar_count);

    for (i = 0; i < __pvar_count; i++)
    {
        Sample_Grid_init(&(flat_matrix->mpi_t_pvar_grids[i]), time_slices, num_slots);
    }

#endif /* HAVE_MPI_T */
}

int Flat_Matrix_init_coll(struct Collector *coll)
{
    int rank;
    struct Flat_Matrix *flat_matrix = NULL;

    assert(coll);

    const uint64_t time_slices = 2048;
    const uint64_t partition_size = 1;
    PMPI_Comm_rank(VMPI_Get_partition_comm(), &rank);

    flat_matrix = malloc(sizeof(struct Flat_Matrix));

    coll->internal_data = flat_matrix;

    Flat_Matrix_init_coll_MALP_EVENT_MPI(flat_matrix, time_slices, 1);
    Flat_Matrix_init_coll_MALP_EVENT_WRAPPED(flat_matrix, time_slices, 1);
    Flat_Matrix_init_coll_MALP_EVENT_OMPT(flat_matrix, time_slices, 1);
    Flat_Matrix_init_coll_MALP_EVENT_MPI_T(flat_matrix, time_slices, 1);

    return 0;
}

static void
Flat_Matrix_release_coll_MALP_EVENT_MPI(struct Flat_Matrix *flat_matrix)
{
    int i, j;
    assert(flat_matrix);
    for (i = 0; i < T_MPI_COUNT; i++)
        for (j = 0; j < FLAT_GRID_COUNT; j++)
            Sample_Grid_release(&(flat_matrix->mpi_grids[i][j]));
}

static void
Flat_Matrix_release_coll_MALP_EVENT_WRAPPED(struct Flat_Matrix *flat_matrix)
{
    int i, j;
    assert(flat_matrix);
    for (i = 0; i < T_POSIX_COUNT; i++)
        for (j = 0; j < FLAT_GRID_COUNT; j++)
            Sample_Grid_release(&(flat_matrix->posix_grids[i][j]));
}

static void
Flat_Matrix_release_coll_MALP_EVENT_OMPT(struct Flat_Matrix *flat_matrix)
{
#if HAVE_OMPT_H
    Sample_2D_Grid_fini(flat_matrix->ompt_grids);
    Sample_2D_Grid_desalloc(flat_matrix->ompt_grids);
    flat_matrix->ompt_grids = NULL;
#endif /* HAVE_OMPT_H */
}

static void
Flat_Matrix_release_coll_MALP_EVENT_MPI_T(struct Flat_Matrix *flat_matrix)
{
    return;
#if HAVE_MPI_T

    flat_matrix->mpi_t_cvar_grids = NULL;

    flat_matrix->mpi_t_pvar_grids = NULL;
#endif /* HAVE_MPI_T */
}

int Flat_Matrix_release_coll(struct Collector *coll)
{
    struct Flat_Matrix *flat_matrix;

    assert(coll && coll->internal_data);
    flat_matrix = (struct Flat_Matrix *)coll->internal_data;

    Flat_Matrix_release_coll_MALP_EVENT_MPI(flat_matrix);
    Flat_Matrix_release_coll_MALP_EVENT_WRAPPED(flat_matrix);
    Flat_Matrix_release_coll_MALP_EVENT_OMPT(flat_matrix);
    Flat_Matrix_release_coll_MALP_EVENT_MPI_T(flat_matrix);

    free(coll->internal_data);
    coll->internal_data = NULL;
    return 0;
}

static void
Flat_Matrix_push_event_MALP_EVENT_MPI(struct MALP_Trace_Event *event, struct Flat_Matrix *flat_matrix)
{
    struct MPI_Event *mpievt;

    assert(event && flat_matrix);
    assert(flat_matrix->mpi_grids);
    mpievt = &(event->event.mpi_event);

    const int type = mpievt->type;
    assert(type < T_MPI_COUNT);
    double __duration = event->duration / Timer_second();
    double __timestamp = event->timestamp / Timer_second();

    /* Update the global profile */
    Sample_Grid_accumulate(&(flat_matrix->mpi_grids[type][FLAT_GRID_HITS]), 0, __timestamp, 1);
    Sample_Grid_accumulate(&(flat_matrix->mpi_grids[type][FLAT_GRID_TIME]), 0, __timestamp, __duration);

    if (mpievt->size)
        Sample_Grid_accumulate(&(flat_matrix->mpi_grids[type][FLAT_GRID_SIZE]), 0, __timestamp, mpievt->size);
}

static void
Flat_Matrix_push_event_MALP_EVENT_WRAPPED(struct MALP_Trace_Event *event, struct Flat_Matrix *flat_matrix)
{
    struct Wrapped_Event *we;

    assert(event && flat_matrix);
    assert(flat_matrix->posix_grids);

    we = &(event->event.wrapped_event);

    const int type = we->type;
    double __duration = event->duration / Timer_second();
    double __timestamp = event->timestamp / Timer_second();

    /* Update the global profile */
    Sample_Grid_accumulate(&(flat_matrix->posix_grids[type][FLAT_GRID_HITS]), 0, __timestamp, 1);
    Sample_Grid_accumulate(&(flat_matrix->posix_grids[type][FLAT_GRID_TIME]), 0, __timestamp, __duration);

    if (we->size)
        Sample_Grid_accumulate(&(flat_matrix->posix_grids[type][FLAT_GRID_SIZE]), 0, __timestamp, we->size);
}

static void
Flat_Matrix_push_event_MALP_EVENT_OMPT(struct MALP_Trace_Event *event, struct Flat_Matrix *flat_matrix)
{
#if HAVE_OMPT_H
    MALP_OMPT_event_t *ompt;
    struct Sample_Grid **grids;

    assert(event && flat_matrix);
    assert(flat_matrix->ompt_grids);
    assert(flat_matrix->ompt_grids->grids);

    grids = flat_matrix->ompt_grids->grids;
    ompt = &(event->event.ompt);

    const int type = ompt->type;
    double __duration = event->duration / Timer_second();
    double __timestamp = event->timestamp / Timer_second();

    /* Update the global profile */
    Sample_Grid_accumulate(&(grids[type][FLAT_GRID_HITS]), 0, __timestamp, 1);
    Sample_Grid_accumulate(&(grids[type][FLAT_GRID_TIME]), 0, __timestamp, __duration);

    if (ompt->size)
        Sample_Grid_accumulate(&(grids[type][FLAT_GRID_SIZE]), 0, __timestamp, ompt->size);
#endif /* HAVE_OMPT_H */
}

static void
Flat_Matrix_push_event_MALP_EVENT_MPI_T(struct MALP_Trace_Event *event, struct Flat_Matrix *flat_matrix)
{
#if HAVE_MPI_T
    struct Sample_Grid *grid = NULL;
    struct MALP_MPI_T_event_s *mpit_evt;

    assert(event && flat_matrix);

    mpit_evt = &(event->event.mpit);


    if (!MALP_mpi_t_is_numeric(mpit_evt))
    {
        return;
    }

    const int type = mpit_evt->logical_var_idx;

    double __timestamp = event->timestamp / Timer_second();

    switch (mpit_evt->type)
    {
    case MALP_MPI_T_TYPE_POLL_CVAR:
        grid = flat_matrix->mpi_t_cvar_grids;
        break;
    case MALP_MPI_T_TYPE_POLL_PVAR:
        grid = flat_matrix->mpi_t_pvar_grids;
        break;
    default: /* Skip unknown MPI_T Event */
        grid = NULL;
    }

    if (!grid)
        return;

    double val = MALP_mpi_t_to_double(mpit_evt);


    Sample_Grid_accumulate(grid, 0, __timestamp, val);

#endif /* HAVE_MPI_T */
}

void Flat_Matrix_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
    struct Collector *coll = (struct Collector *)pcoll;
    struct Flat_Matrix *flat_matrix = (struct Flat_Matrix *)coll->internal_data;

    switch (event->type)
    {

    case MALP_EVENT_MPI:
        Flat_Matrix_push_event_MALP_EVENT_MPI(event, flat_matrix);
        break;

    case MALP_EVENT_WRAPPED:
        Flat_Matrix_push_event_MALP_EVENT_WRAPPED(event, flat_matrix);
        break;
#if HAVE_OMPT_H
    case MALP_EVENT_OMPT:
        Flat_Matrix_push_event_MALP_EVENT_OMPT(event, flat_matrix);
        break;
#endif /* HAVE_OMPT_H */

#if HAVE_MPI_T
    case MALP_EVENT_MPI_T:
        Flat_Matrix_push_event_MALP_EVENT_MPI_T(event, flat_matrix);
        break;
#endif /* HAVE_MPI_T */
    default:
        /* Nothing to do */
        break;
    }
}


static void Flat_Matrix_perform_reduce_single(struct Sample_Grid *grid, const int partition_size, const int partition_id, int flatten)
{
    /* Do a '0' hit over the matrix in order to make all matrix having the same quantum */
    double end_time = Duration_infos_get_end(partition_id) / Timer_second();
    Sample_Grid_scale(grid, end_time * 1.05);

    /* We only flatten temporal grids */
    if (flatten)
        Sample_Grid_flatten(grid, partition_size);

    Sample_Grid_reduce(grid);

}

static void
Flat_Matrix_perform_reduce(struct Sample_Grid *grid, const int partition_size, const int partition_id)
{
    int i;

    for (i = 0; i < FLAT_GRID_COUNT; i++)
    {
        Flat_Matrix_perform_reduce_single( &(grid[i]), partition_size, partition_id, i == FLAT_GRID_TIME);
    }
}

static void
Flat_Matrix_reduce_MALP_EVENT_MPI(struct Flat_Matrix *flat_matrix, const int partition_size, const int partition_id)
{
    int i;
    assert(flat_matrix && flat_matrix->mpi_grids);
    for (i = 0; i < T_MPI_COUNT; i++)
        Flat_Matrix_perform_reduce(flat_matrix->mpi_grids[i], partition_size, partition_id);
}

static void
Flat_Matrix_reduce_MALP_EVENT_WRAPPED(struct Flat_Matrix *flat_matrix, const int partition_size, const int partition_id)
{
    int i;
    assert(flat_matrix && flat_matrix->mpi_grids);
    for (i = 0; i < T_POSIX_COUNT; i++)
        Flat_Matrix_perform_reduce(flat_matrix->posix_grids[i], partition_size, partition_id);
}

static void
Flat_Matrix_reduce_MALP_EVENT_OMPT_T(struct Flat_Matrix *flat_matrix, const int partition_size, const int partition_id)
{
#if __HAVE_OMPT_H
    int i;
    struct Sample_Grid **grids;
    assert(flat_matrix && flat_matrix->ompt_grids);
    assert(flat_matrix->ompt_grids->grids);
    grids = flat_matrix->ompt_grids->grids;
    for (i = 0; i < flat_matrix->ompt_grids->dim_x; i++)
        Flat_Matrix_perform_reduce(grids[i], partition_size, partition_id);
#endif /* __HAVE_OMPT_H */
}

static void
Flat_Matrix_reduce_MALP_EVENT_MPI_T(struct Flat_Matrix *flat_matrix, const int partition_size, const int partition_id)
{
#if HAVE_MPI_T
    int i;

    const int __cvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_CVAR);
    const int __pvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_PVAR);

    for (i = 0; i < __cvar_count; i++)
        Flat_Matrix_perform_reduce_single(&(flat_matrix->mpi_t_cvar_grids[i]), partition_size, partition_id, 0);


    for (i = 0; i < __pvar_count; i++)
        Flat_Matrix_perform_reduce_single(&(flat_matrix->mpi_t_pvar_grids[i]), partition_size, partition_id, 0);
#endif /* HAVE_MPI_T */
}

int Flat_Matrix_reduce(struct Collector *coll)
{
    struct Flat_Matrix *flat_matrix = (struct Flat_Matrix *)coll->internal_data;

    const int partition_id = coll->partition_id;
    const int partition_size = Collector_partition_size(coll);

    Flat_Matrix_reduce_MALP_EVENT_MPI(flat_matrix, partition_size, partition_id);
    Flat_Matrix_reduce_MALP_EVENT_WRAPPED(flat_matrix, partition_size, partition_id);
    Flat_Matrix_reduce_MALP_EVENT_OMPT_T(flat_matrix, partition_size, partition_id);
    Flat_Matrix_reduce_MALP_EVENT_MPI_T(flat_matrix, partition_size, partition_id);

    return 0;
}

static json_t *
Flat_Matrix_gen_entry(struct Sample_Grid *grids, char *name)
{
    int i;
    json_t *ret, *jname, *measures;

    ret = json_object();
    assert(ret);

    jname = json_string(name);
    assert(jname);
    json_object_set(ret, "name", jname);
    json_decref(jname);

    measures = json_object();
    assert(measures);

    for (i = 0; i < FLAT_GRID_COUNT; i++)
    {
        if (!(grids[i].has_data))
            continue;

        json_t *tmp = Sample_Grid_dump_JSON(&(grids[i]));
        json_object_set(measures, (char *)flat_grid_types_names[i], tmp);
        json_decref(tmp);
    }

    json_object_set(ret, "measures", measures);
    json_decref(measures);
    return ret;
}

static void
Flat_Matrix_accumulate_MALP_EVENT_MPI(struct Flat_Matrix *flat_matrix, json_t *measures)
{
    int i;
    json_t *__tmp_ret;

    assert(measures);
    assert(flat_matrix && flat_matrix->mpi_grids);

    for (i = 0; i < T_MPI_COUNT; i++)
    {
        if (!(flat_matrix->mpi_grids[i][0].has_data))
            continue;

        __tmp_ret = Flat_Matrix_gen_entry(flat_matrix->mpi_grids[i], (char *)MPI_Event_names[i]);
        json_array_push(measures, __tmp_ret);
        json_decref(__tmp_ret);
        __tmp_ret = NULL;
    }
}

static void
Flat_Matrix_accumulate_MALP_EVENT_WRAPPED(struct Flat_Matrix *flat_matrix, json_t *measures)
{
    int i;
    json_t *__tmp_ret;

    assert(measures);
    assert(flat_matrix && flat_matrix->posix_grids);

    for (i = 0; i < T_POSIX_COUNT; i++)
    {
        if (!(flat_matrix->posix_grids[i][0].has_data))
            continue;

        __tmp_ret = Flat_Matrix_gen_entry(flat_matrix->posix_grids[i], (char *)Wrapped_symbol_names[i]);
        json_array_push(measures, __tmp_ret);
        json_decref(__tmp_ret);
        __tmp_ret = NULL;
    }
}

static void
Flat_Matrix_accumulate_MALP_EVENT_OMPT(struct Flat_Matrix *flat_matrix, json_t *measures)
{
#if HAVE_OMPT_H
    int i;
    json_t *__tmp_ret;
    struct Sample_Grid **grids;

    assert(measures);
    assert(flat_matrix && flat_matrix->ompt_grids);
    assert(flat_matrix->ompt_grids->grids);

    grids = flat_matrix->ompt_grids->grids;

    for (i = 0; i < flat_matrix->ompt_grids->dim_x; i++)
    {
        if (!(grids[i][0].has_data))
            continue;

        __tmp_ret = Flat_Matrix_gen_entry(grids[i], (char *)OMPT_Event_names[i]);
        json_array_push(measures, __tmp_ret);
        json_decref(__tmp_ret);
        __tmp_ret = NULL;
    }
#endif /* HAVE_OMPT_H */
}

static void
Flat_Matrix_accumulate_MALP_EVENT_MPI_T(struct Flat_Matrix *flat_matrix, json_t *measures)
{
#if HAVE_MPI_T

    int i;
    json_t *__tmp_ret;

    assert(measures);
    assert(flat_matrix && flat_matrix->mpi_t_cvar_grids);

    for (i = 0; i < __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_CVAR); i++)
    {
        if (!(flat_matrix->mpi_t_cvar_grids[i].has_data))
            continue;

        __tmp_ret = Flat_Matrix_gen_entry(&flat_matrix->mpi_t_cvar_grids[i], (char *)__malp_mpi_t_get_var_name(MALP_MPI_T_TYPE_POLL_CVAR, i));
        json_array_push(measures, __tmp_ret);
        json_decref(__tmp_ret);
        __tmp_ret = NULL;
    }

    for (i = 0; i < __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_PVAR); i++)
    {
  
        if (!(flat_matrix->mpi_t_pvar_grids[i].has_data))
            continue;

        __tmp_ret = Flat_Matrix_gen_entry(&flat_matrix->mpi_t_pvar_grids[i], (char *)__malp_mpi_t_get_var_name(MALP_MPI_T_TYPE_POLL_PVAR, i));
        json_array_push(measures, __tmp_ret);
        json_decref(__tmp_ret);
        __tmp_ret = NULL;
    }

#endif /* HAVE_MPI_T */
}

int Flat_Matrix_accumulate(struct Collector *coll)
{
    int rank;
    char uri[100];
    struct Flat_Matrix *flat_matrix;
    json_t *flat_matrices, *measures;

    PMPI_Comm_rank(VMPI_Get_partition_comm(), &rank);
    if (rank)
        return 0;

    /* Only process 0 from here */
    assert(coll && coll->internal_data);
    flat_matrix = (struct Flat_Matrix *)coll->internal_data;

    flat_matrices = json_object();
    assert(flat_matrices);

    measures = json_array();
    assert(measures);

    Flat_Matrix_accumulate_MALP_EVENT_MPI(flat_matrix, measures);
    Flat_Matrix_accumulate_MALP_EVENT_WRAPPED(flat_matrix, measures);
    Flat_Matrix_accumulate_MALP_EVENT_OMPT(flat_matrix, measures);
    Flat_Matrix_accumulate_MALP_EVENT_MPI_T(flat_matrix, measures);

    json_object_set(flat_matrices, "measures", measures);
    json_decref(measures);

    snprintf(uri, 100, "%s", coll->analysis_cache_name);
    nodeSharedCache_set(&nsc, uri, flat_matrices);
    json_decref(flat_matrices);
    flat_matrices = NULL;

    return 0;
}

void Flat_Matrix_init(struct Collector *coll, uint64_t partition_id)
{

    Collector_init(coll, MALP_EVENT_ANY,
                   partition_id,
                   "Flat_Matrix",
                   Flat_Matrix_init_coll,
                   Flat_Matrix_release_coll,
                   Flat_Matrix_push_event,
                   Flat_Matrix_reduce,
                   Flat_Matrix_accumulate);
}
