#include "MPIT_Profile.h"

#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include "VMPI.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>


#if HAVE_MPI_T

static void __MPI_T_init_Profile_acc( struct MPI_T_Profile_acc* prof, const int count )
{
   assert( prof && count >= 0 );

   new_Profile_acc->hits = (uint64_t) malloc( sizeof( uint64_t ) * count ); 
   assert( new_Profile_acc->hits );
   memset( new_Profile_acc->hits, 0, sizeof( uint64_t ) * count );
  
   new_Profile_acc->size = (uint64_t) malloc( sizeof( uint64_t ) * count ); 
   assert( new_Profile_acc->size );
   memset( new_Profile_acc->size, 0, sizeof( uint64_t ) * count );

}

static void __MPI_T_Init_Global_Profil_const_cvars( struct MPI_T_Profile* prof )
{
    int index_ignored, i, err, count, val;
    MPI_T_cvar_handle handle;

    for( index_ignored = 0, i = 0; i < up; i++ ) 
    {
       if( i == prof->mpit_cvars_not_const_ids[index_ignored] )
       {
           index_ignored++;
           continue;
       }

       /* ASSUME MPI_INT TYPE TODO: extend ...*/
       err = MPI_T_cvar_handle_alloc( i, NULL, &handle, &count ); 
       assert( err == MPI_SUCCESS );
      
       /* ASSUME single value */
       assert( count == 1 ); 

       err = MPI_T_cvar_read( handle, &val );
       assert( err == MPI_SUCCESS );

       /* Update Global Infos */
       prof->global_acc.cvars_hits[i] += 1;
       prof->global_acc.cvars_size[i] += val;

       err=MPI_T_cvar_handle_free(&handle);
       assert( err == MPI_SUCCESS );
    }
}

int  MPI_T_Profile_init_coll( struct Collector * coll )
{
   
   int rank, err, threadsupport;
   int cvars_num, pvars_num, cvars_poll_num, pvars_poll_num;
   int *cvars_poll_ids, *pvars_poll_ids;
   struct MPI_T_Profile * prof;

   assert( coll );

   coll->internal_data = malloc( sizeof( struct MPI_T_Profile ) );
   assert( coll->internal_data );
   memset( coll->internal_data, 0, sizeof( struct MPI_T_Profile ) );
   
   err = MPI_T_init_thread( MPI_THREAD_SINGLE, &threadsupport ); 
   assert( err == MPI_SUCCESS );

   err = MPI_T_cvar_get_num( &cvars_num ); 
   assert( err == MPI_SUCCESS );
   cvars_poll_ids = MPI_T_Filter_constant_cvars( cvars_num, MALP_MPI_T_CVARS ); 

   err = MPI_T_pvar_get_num( &pvars_num ); 
   assert( err == MPI_SUCCESS );

   const int total_num = cvars_num + pvars_num;
   const int total_poll_num = cvars_poll_num + pvars_poll_num;

   prof = (struct MPI_T_Profile *) coll->internal_data;
   Hash_Table_init( &prof->per_stream_profiles, sizeof(struct MPI_T_Profile_acc_cell), 64, MPI_T_PROF_COLL_AVOID );

   // Init Global Profile arrays
   __MPI_T_init_Profile_acc( &global_acc );

   // Get constant value
   __MPI_T_Init_Global_Profil_const_cvars( prof );

   // Create the stream array
   PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

   if( !rank )
   {
      char uri[100];
      json_t * array = json_array();
      snprintf( uri, 100, "%s/streams", coll->analysis_cache_name);
      nodeSharedCache_set( &nsc, uri , array );
   }

   return 0;
}

int  MPI_T_Profile_release_coll( struct Collector * coll )
{
   struct MPI_T_Profile * prof;
   assert( coll );
   
   prof = (struct MPI_T_Profile *) coll->internal_data;
   assert( prof );

   Hash_Table_release( &prof->per_stream_profiles, NULL );
   free( prof );

   coll->internal_data = NULL; 
   return 0;
}

static int
MPI_T_Profile_acc_test(void *elem, uint64_t key, void *dummy) 
{
   assert( elem );
   struct MPI_T_Profile_acc *acc = (struct MPI_T_Profile_acc *) elem;
   if( acc->associated_tid == key ) return 1;
   return 0; 
}

static void 
MPI_T_Profile_push_event(struct MALP_Trace_Event *event, void *coll)
{
   int type;
   struct Collector * __coll;
   struct MPI_T_Profile * prof;
   struct MPI_T_Profile_acc *acc;

   if( event->type != MALP_EVENT_MPI_T )
      return;
   
   assert( coll );
   assert( event );
   
   __coll = ( struct Collector * ) coll;
   prof = (struct MPI_T_Profile *) __coll->internal_data;

   type = event->event.mpit.type; 
   
   // Update the global profile
   prof->global_acc.hits[type]++;
   prof->global_acc.time[type]+= event->duration;
   
   // Update the perrank profile
   acc = Hash_Table_get( &prof->per_stream_profiles, event->tid , NULL, MPI_T_Profile_acc_test);
 
   if(! acc )
   {
      struct MPI_T_Profile_acc tmp;
      memset( &tmp, 0, sizeof( struct MPI_T_Profile_acc ) );
      tmp.associated_tid = event->tid;
      acc = Hash_Table_push( &prof->per_stream_profiles, (void *)&tmp , event->tid );
   }

   acc->hits[type]++;
}

static int MPI_T_Profile_reduce( struct Collector * coll )
{
   uint64_t *__tmp_send_buff, *__tmp_recv_buff;
	
   struct MPI_T_Profile *prof = (struct MPI_T_Profile *)coll->internal_data;
   struct MPI_T_Profile_acc *global_acc = &prof->global_acc;
   
   const int not_const_cvars = prof->mpit_cvars_not_const;
   __tmp_recv_buff = ( uint64_t* ) malloc( sizeof( uint64_t ) * not_const_cvars );  
   assert( __tmp_recv_buff ); 
   __tmp_send_buff = ( uint64_t* ) malloc( sizeof( uint64_t ) * not_const_cvars );  
   assert( __tmp_send_buff ); 

   // HITS
   __MPI_T_pack_not_const_value( prof, global_acc->cvars_hits, __tmp_send_buff  )
   PMPI_Reduce(__tmp_send_buff, __tmp_recv_buff, not_const_cvars, MPI_UNSIGNED_LONG_LONG , MPI_SUM, 0, VMPI_Get_partition_comm());
   __MPI_T_pack_not_const_value( prof, global_acc->cvars_hits, __tmp_recv_buff  )
	
   // SIZE
   __MPI_T_pack_not_const_value( prof, global_acc->cvars_size, __tmp_send_buff  )
   PMPI_Reduce(__tmp_send_buff, __tmp_recv_buff, not_const_cvars, MPI_UNSIGNED_LONG_LONG , MPI_SUM, 0, VMPI_Get_partition_comm());
   __MPI_T_pack_not_const_value( prof, global_acc->cvars_size, __tmp_recv_buff  )

   free( __tmp_recv_buff ); __tmp_recv_buff = NULL;
   free( __tmp_send_buff ); __tmp_send_buff = NULL;

   const int pvars_tot = prof->mpit_pvars_tot;
   const int cvars_tot = prof->mpit_cvars_tot;

   /* first pvars is after last cvars */
   __tmp_send_buff = &( global_acc->hit[cvars_tot] );
   __tmp_recv_buff = ( uint64_t* ) malloc( sizeof( uint64_t ) * pvars_tot ); 
   assert( __tmp_recv_buff );

   // HITS
   PMPI_Reduce( __tmp_send_buff, __tmp_recv_buff, pvars_tot, MPI_UNSIGNED_LONG_LONG , MPI_SUM, 0, VMPI_Get_partition_comm());
   memcpy( __tmp_send_buff, __tmp_recv_buff, pvars_tot * sizeof(uint64_t ) );

   __tmp_send_buff = &( global_acc->size[cvars_tot] );
   memset( __tmp_recv_buff, 0, sizeof( uint64_t ) * pvars_tot );	

   // SIZE
   PMPI_Reduce( __tmp_send_buff, __tmp_recv_buff, pvars_tot, MPI_UNSIGNED_LONG_LONG , MPI_SUM, 0, VMPI_Get_partition_comm());
   memcpy( __tmp_send_buff, __tmp_recv_buff, pvars_tot * sizeof(uint64_t ) );

   return 0;
}

static json_t * MPI_T_Profile_gen( struct MPI_T_Profile_acc * acc, int render_min_max )
{
   int i;
   json_t * profile = json_array();
   
   const int cvars_num = 128;

   for(i = 0 ; i < cvars_num; i++ )
   {
      if( !( acc->hits[i] ) ) continue; 

      json_t * function_entry = json_object();
      json_t * name = json_string( (char *) MPI_T_Event_names[i] );
      json_t * hits = json_int( acc->hits[i] );
      json_t * time = json_real( (double) acc->time[i] / Timer_second() );
      json_t * size = json_real( acc->size[i] );

      json_object_set( function_entry, "name", name);
      json_object_set( function_entry, "hits", hits);
      json_object_set( function_entry, "size", size);
			
      json_decref(name);
      json_decref(hits);
      json_decref(size);
			
      json_array_push( profile, function_entry );
      json_decref( function_entry );
			
   }
   return profile;
}

static void MPI_T_Profile_render_per_rank_profile(void *acc, void *pcoll)
{
   json_t *profile, *entry, *tid;
   struct MPI_T_Profile_acc* __acc;
   char entry_name[100], stid[100];
   struct Stream_Descriptor *stream_desc;

   __acc = (struct MPI_T_Profile_acc *) acc;

   const uint64_t associated_tid = __acc->associated_tid;
  
   profile = MPI_T_Profile_gen( __acc, 0 );  
   stream_desc = Meta_Data_Handler_get_desc( __acc->associated_tid );

   if( !stream_desc ) return;

   const uint64_t partition_id = stream_desc->partition_id;

   entry = json_object();
   snprintf(entry_name, 100, "%ld/MPI_T_Profile/streams/%lu", partition_id, associated_tid );
   snprintf( stid, 100, "%lu", associated_tid );
   
   tid = json_string( stid );
   json_object_set( entry, "tid", tid );
   json_decref( tid );
   json_object_set( entry, "profile", profile );
   json_decref( profile );

   nodeSharedCache_reduce( &nsc, entry_name, entry);

   json_decref( entry );
}

static int MPI_T_Profile_accumulate( struct Collector * coll )
{
   int rank;
   struct MPI_T_Profile * prof;

   assert( coll );
   prof = (struct MPI_T_Profile *) coll->internal_data;
   assert( prof ); 

   PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
	
   // Each rank renders its own profiles
   Hash_Table_walkthrought_2p( &prof->per_stream_profiles, (void *)coll, MPI_T_Profile_render_per_rank_profile  );
	
   if( !rank )
   {
      char uri[100];
      json_t * profile;
      struct MPI_T_Profile_acc *global_acc;

      global_acc = &prof->global_acc;
      assert( global_acc );

      // Do the JSON rendering
      profile =  MPI_T_Profile_gen( global_acc, 1 );
      snprintf(uri, 100, "%s/profile", coll->analysis_cache_name );
      nodeSharedCache_set( &nsc, uri, profile );
      json_decref( profile );
   }

   return 0;
}

#endif /* HAVE_MPI_T */


void MPI_T_Profile_init( struct Collector* coll, uint64_t partition_id )
{
#if HAVE_MPI_T
        Collector_init( coll, MALP_EVENT_MPI_T,
                                        partition_id,
                                        "MPI_T_Profile",
                                        MPI_T_Profile_init_coll,
                                        MPI_T_Profile_release_coll,
                                        MPI_T_Profile_push_event,
                                        MPI_T_Profile_reduce,
                                        MPI_T_Profile_accumulate );
#endif /* HAVE_MPI_T */
}
