#include "MPI_Section_col.h"

#include "MPI_Events.h"

#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include "VMPI.h"

#include <stdlib.h>
#include <string.h>


int MPI_Section_coll_entry_init( struct MPI_Section_coll_entry *ent, uint64_t rank )
{
	char label[64];
	
	snprintf( label, 64,  "%lld", rank );
	if( MPI_Section_tree_init_static( &ent->tree, label, 4096, (void *)ent ) < 0 )
		return -1;

	ent->lock = 0;
	printf("Setting rank to %d\n", rank );
	ent->rank = rank;
	ent->current_tree = &ent->tree;
	
	return 0;
}

int MPI_Section_coll_entry_release( struct MPI_Section_coll_entry *ent )
{
	if( MPI_Section_tree_release( &ent->tree ) < 0 )
		return -1;
	
	ent->rank = (uint64_t)-1;
	ent->current_tree = NULL;
	
	return 0;
}



void MPI_Section_coll_ctx_init( struct MPI_Section_coll_ctx * ctx )
{
	Hash_Table_init( &ctx->ht, sizeof( struct MPI_Section_coll_entry_cell ), 32, MPI_SECTION_COLL_AVOID );
	ctx->ht_lock = 0;
}

void MPI_Section_coll_ctx_release( struct MPI_Section_coll_ctx * ctx )
{
	
}


int  MPI_Section_init_coll( struct Collector * coll )
{
	coll->internal_data = malloc( sizeof( struct MPI_Section_coll_ctx ) );
	
	if( !coll->internal_data )
	{
		perror("malloc");
		abort();
	}
	
	memset( coll->internal_data, 0, sizeof( struct MPI_Section_coll_ctx ) );
	
	struct MPI_Section_coll_ctx * sect = (struct MPI_Section_coll_ctx *)coll->internal_data;
	
	MPI_Section_coll_ctx_init( sect );
	
		
	/* Create the stream array */
	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
	
	if( rank == 0 )
	{
		json_t * array = json_array();
		
		char uri[100];
		snprintf( uri, 100, "%s/ranks", coll->analysis_cache_name);
		printf("URI : %s\n", uri );
		nodeSharedCache_set( &nsc, uri , array );
		
		json_decref(array);
	}
	
	return 0;
}


int  MPI_Section_release_coll( struct Collector * coll )
{
	struct MPI_Section_coll_ctx * sect = (struct MPI_Section_coll_ctx *)coll->internal_data;

	MPI_Section_coll_ctx_release( sect );
	
	return 0;
}

int MPI_Section_coll_entry_test(void * a , uint64_t key, void * arg)
{
	struct MPI_Section_coll_entry * ent = (struct MPI_Section_coll_entry*)a;
	
	if( ent->rank == key )
		return 1;
	
	return 0;
}

void MPI_Section_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
	struct Collector *coll = (struct Collector *)pcoll;
	struct MPI_Section_coll_ctx * sect = (struct MPI_Section_coll_ctx *)coll->internal_data;

	//printf("%lu Event from %ld type %s(%d)\n", event->timestamp, event->rank, MALP_Trace_event_type_names[event->type], event->type );

	if( event->type != MALP_EVENT_MPI_SECTION )
	{
		return;
	}
	
	struct MPI_Section_event *me = &event->event.mpi_section;
	//printf("[%d] MPI_SECTION [%s %s]\n", event->rank, me->label , MPI_Section_type_name[ me->type ] );

	MALP_Spinlock_lock( &sect->ht_lock );
	/* Get or create CTX */
	void * ctx =  Hash_Table_get( &sect->ht, event->rank , NULL, MPI_Section_coll_entry_test);

	if( !ctx )
	{
		struct MPI_Section_coll_entry ent;
		ctx = Hash_Table_push( &sect->ht, &ent , event->rank );
		MPI_Section_coll_entry_init( (struct MPI_Section_coll_entry *)ctx, event->rank );

	}

	struct MPI_Section_coll_entry * current_ctx = (struct MPI_Section_coll_entry *)ctx;

	/* Bind initial tree to current section */
	
	MALP_Spinlock_unlock( &sect->ht_lock );
	
	
	MALP_Spinlock_lock( &current_ctx->lock );
	
	if( me->type == MPI_SECTION_BEGIN )
	{
		current_ctx->current_tree = MPI_Section_tree_enter( current_ctx->current_tree, me->label, event->timestamp );
	}
	else if( me->type == MPI_SECTION_END )
	{
		current_ctx->current_tree = MPI_Section_tree_leave( current_ctx->current_tree, me->label, event->timestamp );
	}

	MALP_Spinlock_unlock( &current_ctx->lock );
}

int  MPI_Section_reduce( struct Collector * coll )
{
	
	
	return 0;
}

void render_per_rank_section(void *pctx, void *pcoll)
{
	struct MPI_Section_coll_entry * entry = (struct MPI_Section_coll_entry *) pctx;
	struct Collector *coll = (struct Collector *)pcoll;
	
	
	char entry_name[100];
	snprintf(entry_name, 100, "%s/ranks/%d", coll->analysis_cache_name , entry->rank );

	int crank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &crank );
		


	printf("===>[%d] %s COLL @ %p\n", crank, entry_name , pcoll);
	
	json_t * section_info = json_object();
	
	json_t * rank = json_int( entry->rank );
		
	json_object_set( section_info, "rank", rank );

	json_t * sections = json_array();
	
	MPI_Section_tree_compute_stats( &entry->tree );
	MPI_Section_tree_render( &entry->tree, sections );
	
	json_object_set( section_info, "sections", sections );
	
	nodeSharedCache_reduce( &nsc, entry_name, section_info);

	json_decref( rank );
	json_decref( section_info );
	json_decref( sections );
}



int  MPI_Section_accumulate( struct Collector * coll )
{
	struct MPI_Section_coll_ctx * sect = (struct MPI_Section_coll_ctx *)coll->internal_data;

		
	/* Each rank render its sections */
	
	Hash_Table_walkthrought_2p( &sect->ht, (void *)coll, render_per_rank_section  );
	
	
	
	
	return 0;
}




void MPI_Section_init(struct Collector *coll, uint64_t partition_id)
{
		
	Collector_init( coll, MALP_EVENT_ANY,
				partition_id,
				"MPI_Section",
				MPI_Section_init_coll,
				MPI_Section_release_coll,
				MPI_Section_push_event,
				MPI_Section_reduce,
				MPI_Section_accumulate );
	
}
