#include "Map.h"

#include "MPI_Events.h"
#include "VMPI.h"
#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include "map_plugin.h"
#include "Flat_Matrix.h"




static struct MAP_Plugin_array __map_plugin_array;


int  MAP_init_coll( struct Collector * coll )
{
	char * plugin_prefix = NULL;
	
	plugin_prefix = getenv( "MAP_PLUGIN_PREFIX" );
	
	if( !plugin_prefix )
	{
		return 0;
	}
	
	int ret = MAP_Plugin_array_load( &__map_plugin_array, plugin_prefix );
	
	if( ret )
	{
		abort();
	}
	
	
	/* Prepare the payloads */
	int i,j;
	
	for( i = 0 ; i < __map_plugin_array.plugin_count; i++ )
	{
		/* For each plugin */
		struct MAP_Plugin * plugin = MAP_Plugin_array_get( &__map_plugin_array, i );
		 
		for( j = 0 ; j < plugin->metric_count ; j++ )
		{
			/* For each metric in the plugin */
			struct MAP_Metric * metric = MAP_Plugin_get_metric( plugin, j );
			 
			struct MAP_Metric_payload * pl = malloc( sizeof( struct MAP_Metric_payload ) );
			
			if( !pl )
			{
				perror("malloc");
				abort();	
			}
			 
			Sample_Grid_init( &pl->grid , 4096, 1 );
			
			/* Save the payload pointer inside the metric */
			MAP_Metric_set_payload( metric, (void *)pl );
			 
		}
		
	}
	
	
	return 0;
}


int  MAP_release_coll( struct Collector * coll )
{
	return 0;
}



void MAP_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
	
	if( event->type != MALP_EVENT_MAP )
		return;
	
	struct MAP_Event *me = &event->event.map;

	
	//printf("MAP [%d, %d] %g %lld\n",  me->plugin_id,  me->metric_id, me->dvalue, me->ulivalue );

	
	struct MAP_Metric * metric =  MAP_Plugin_array_get_metric( &__map_plugin_array, me->plugin_id,  me->metric_id );
	
	if( !metric )
	{
		fprintf(stderr, "ERROR no such metric");
		abort();
	}
	
	struct MAP_Metric_payload * pl = (struct MAP_Metric_payload *)MAP_Metric_get_payload( metric );
	
	if( !pl )
		return;
	
	double value = 0.0;
	
	if( metric->data_type == MAP_METRIC_UINT64 )
	{
		value = (double) me->ulivalue;
	}
	else if( metric->data_type == MAP_METRIC_DOUBLE )
	{
		value = me->dvalue;
	}
	else
	{
		return;
	}
	
	/* Set the point in the sampled grid */
	Sample_Grid_accumulate( &pl->grid ,  0, event->timestamp / Timer_second(), value);
}

int  MAP_reduce( struct Collector * coll )
{
	int i,j;
	
	/* Just walk the payloads and reduce the grids */
	for( i = 0 ; i < __map_plugin_array.plugin_count; i++ )
	{
		/* For each plugin */
		struct MAP_Plugin * plugin = MAP_Plugin_array_get( &__map_plugin_array, i );
		 
		for( j = 0 ; j < plugin->metric_count ; j++ )
		{
			/* For each metric in the plugin */
			struct MAP_Metric * metric = MAP_Plugin_get_metric( plugin, j );
			 
			struct MAP_Metric_payload * pl = (struct MAP_Metric_payload *)MAP_Metric_get_payload( metric );
			
			if( !pl )
				continue;
			
			/* Do a '0' hit over the matrix in order to make all matrix having the same quantum */
			double end_time = Duration_infos_get_end( coll->partition_id )/ Timer_second(); 
			Sample_Grid_scale( &pl->grid, end_time*1.05 );
			
			Sample_Grid_reduce( &pl->grid );
		 
		}
		
	}
	
	return 0;
}


#define STRING_CHECK_NULL_SUBST_EMPTY( a ) do{ if( a == NULL ) a = ""; } while(0)


int  MAP_accumulate_metric( struct Collector * coll, struct MAP_Metric * metric,  json_t * metric_table )
{
	struct MAP_Metric_payload * pl = (struct MAP_Metric_payload *)MAP_Metric_get_payload( metric );
	
	if( !pl )
		return 0;
	
	if(!pl->grid.has_data )
	{
		/* Do not save empty metrics */
		return 0;
	}
	
	json_t * metric_descriptor = json_object();
	
	char * cid = metric->id;
	STRING_CHECK_NULL_SUBST_EMPTY( cid );
	json_t * id = json_string( cid );
	
	int cis_derivative = metric->derivative;
	json_t * is_derivative = json_bool( cis_derivative );
	
	char * ctype_name = (char *)MAP_Metric_type_name[metric->data_type];
	STRING_CHECK_NULL_SUBST_EMPTY( ctype_name );
	json_t * type_name = json_string( ctype_name );
	
	char * cdomain = (char *)MAP_Metric_domain_name[metric->domain];
	STRING_CHECK_NULL_SUBST_EMPTY( cdomain );
	json_t * domain = json_string( cdomain );
	
	char * cdata_coll = metric->data_collector_ref;
	STRING_CHECK_NULL_SUBST_EMPTY( cdata_coll );
	json_t * data_coll = json_string( cdata_coll );
	
	int cdivide_by_sample = metric->data_collector_divide_by_sample;
	json_t * divide_by_sample = json_bool( cdivide_by_sample );
	
	char * cdisplay_name = metric->display_name;
	STRING_CHECK_NULL_SUBST_EMPTY( cdisplay_name );
	json_t * display_name = json_string( cdisplay_name );
	
	char * cdescription = metric->description;
	STRING_CHECK_NULL_SUBST_EMPTY( cdescription );
	json_t * description = json_string( cdescription );
	
	char * ctype = metric->type;
	STRING_CHECK_NULL_SUBST_EMPTY( ctype );
	json_t * type = json_string( ctype );
	
	char * ccolor = metric->color;
	STRING_CHECK_NULL_SUBST_EMPTY( ccolor );
	json_t * color = json_string( ccolor );
	
	json_object_set( metric_descriptor, "id", id );
	json_decref( id );
		
	json_object_set( metric_descriptor, "is_derivative", is_derivative );
	json_decref( is_derivative );
	
	json_object_set( metric_descriptor, "type_name", type_name );
	json_decref( type_name );
		
	json_object_set( metric_descriptor, "domain", domain );
	json_decref( domain );
		
	json_object_set( metric_descriptor, "data_coll", data_coll );
	json_decref( data_coll );
	
	json_object_set( metric_descriptor, "divide_by_sample", divide_by_sample );
	json_decref( divide_by_sample );
	
	json_object_set( metric_descriptor, "display_name", display_name );
	json_decref( display_name );
		
	json_object_set( metric_descriptor, "description", description );
	json_decref( description );
	
	json_object_set( metric_descriptor, "type", type );
	json_decref( type );
	
	json_object_set( metric_descriptor, "color", color );
	json_decref( color );
	
	
	json_t * measures = Sample_Grid_dump_JSON( &pl->grid );
	json_object_set( metric_descriptor, "measures", measures );
	json_decref( measures );
	
	/* Add ourselves the the array */
	json_array_push( metric_table, metric_descriptor);
}



int  MAP_accumulate_metric_group( struct Collector * coll, struct MAP_Metric_group * metricg,  json_t * metricg_table )
{
	json_t * metricg_descriptor = json_object();

	char * cid = metricg->id;
	STRING_CHECK_NULL_SUBST_EMPTY( cid );
	char * cname = metricg->name;
	STRING_CHECK_NULL_SUBST_EMPTY( cname );
	char * cdescription = metricg->description;
	STRING_CHECK_NULL_SUBST_EMPTY( cdescription );


	json_t * id = json_string( cid );
	json_t * name = json_string( cname );
	json_t * description = json_string( cdescription );
	
	json_t * group_metrics = json_array();
	
	int i;
	
	/* Create the metric group list */
	for( i = 0 ; i < metricg->group_metric_count; i++ )
	{
		struct MAP_Metric * m = metricg->group_metrics[i];
		
		if( !m )
			continue;
		
		json_t * mid = json_string( m->id );
		
		json_array_push( group_metrics, mid );
		
		json_decref( mid );
	}
	
	
	json_object_set( metricg_descriptor, "id", id );
	json_decref( id );
	
	json_object_set( metricg_descriptor, "name", name );
	json_decref( name );
	
	json_object_set( metricg_descriptor, "description", description );
	json_decref( description );
	
	json_object_set( metricg_descriptor, "group", group_metrics );
	json_decref( group_metrics );
	
	/* Add ourselves the the array */
	json_array_push( metricg_table, metricg_descriptor);
}






int  MAP_accumulate_plugin( struct Collector * coll, struct MAP_Plugin * plugin,  json_t * plugin_table )
{
	json_t * plugin_descriptor = json_object();
	
	
	/* Overall Desc */
	char * csource_id = plugin->source_id;
	STRING_CHECK_NULL_SUBST_EMPTY( csource_id );
	
	char * cpath_to_source_so = plugin->path_to_source_so;
	STRING_CHECK_NULL_SUBST_EMPTY( cpath_to_source_so );

	
	
	json_t * source_id = json_string( csource_id );
	json_object_set( plugin_descriptor, "source_id", source_id );
	json_decref( source_id );
	
	
	json_t * path_to_source_so = json_string( cpath_to_source_so );
	json_object_set( plugin_descriptor, "path_to_source_so", path_to_source_so );
	json_decref( path_to_source_so );	
	
	/* METRICS */
	json_t * metric_table = json_array();
	
	int i;
	
	for( i = 0 ; i < plugin->metric_count ; i++ )
	{
		struct MAP_Metric * metric = MAP_Plugin_get_metric( plugin, i );
		MAP_accumulate_metric( coll, metric, metric_table );
	}
	
	json_object_set( plugin_descriptor, "metrics", metric_table );
	json_decref( metric_table );
	
	/* METRIC GROUPS */
	json_t * metricg_table = json_array();
	
	for( i = 0 ; i < plugin->metric_group_count ; i++ )
	{
		struct MAP_Metric_group * metricg = &plugin->metric_groups[i];
		MAP_accumulate_metric_group( coll, metricg, metricg_table );
	}
		
	json_object_set( plugin_descriptor, "groups", metricg_table );
	json_decref( metricg_table );	
	
	/* Add ourselves the the array */
	json_array_push( plugin_table, plugin_descriptor);
}





int  MAP_accumulate( struct Collector * coll )
{

	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

	if( rank )
		return 0;
		
	/* Only 0 is here */

	json_t * map_measures = json_array();
	
	int i,j;
	
	/* Just walk the payloads and reduce the grids */
	for( i = 0 ; i < __map_plugin_array.plugin_count; i++ )
	{
		/* For each plugin */
		struct MAP_Plugin * plugin = MAP_Plugin_array_get( &__map_plugin_array, i );
		 
		MAP_accumulate_plugin( coll, plugin, map_measures );
	}
	
	
	
	char uri[100];
	snprintf(uri, 100, "%s", coll->analysis_cache_name );
	nodeSharedCache_set( &nsc, uri, map_measures );
	
	
	return 0;
}




void MAP_init(struct Collector *coll, uint64_t partition_id)
{
		
	Collector_init( coll, MALP_EVENT_MAP,
					partition_id,
					"map",
					MAP_init_coll,
					MAP_release_coll,
					MAP_push_event,
					MAP_reduce,
					MAP_accumulate );
	
}
