#include "OMPT_Profile.h"

#include "OMPT_Events.h"

#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include "VMPI.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>


#if HAVE_OMPT_H

int  OMPT_Profile_init_coll( struct Collector * coll )
{
   int rank;
   struct OMPT_Profile * prof;

   assert( coll );

   coll->internal_data = malloc( sizeof( struct OMPT_Profile ) );
   assert( coll->internal_data );
   memset( coll->internal_data, 0, sizeof( struct OMPT_Profile ) );

   prof = (struct OMPT_Profile *) coll->internal_data;

   Hash_Table_init( &prof->per_stream_profiles, sizeof(struct OMPT_Profile_acc_cell), 64, OMPT_PROF_COLL_AVOID );

   // Create the stream array
   PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

   if( !rank )
   {
      char uri[100];
      json_t * array = json_array();
      snprintf( uri, 100, "%s/streams", coll->analysis_cache_name);
      nodeSharedCache_set( &nsc, uri , array );
   }

   return 0;
}

int  OMPT_Profile_release_coll( struct Collector * coll )
{
   struct OMPT_Profile * prof;
   assert( coll );
   
   prof = (struct OMPT_Profile *) coll->internal_data;
   assert( prof );

   Hash_Table_release( &prof->per_stream_profiles, NULL );
   free( prof );

   coll->internal_data = NULL; 
   return 0;
}

static int
OMPT_Profile_acc_test(void *elem, uint64_t key, void *dummy) 
{
   assert( elem );
   struct OMPT_Profile_acc *acc = (struct OMPT_Profile_acc *) elem;
   if( acc->associated_tid == key ) return 1;
   return 0; 
}

static void 
OMPT_Profile_push_event(struct MALP_Trace_Event *event, void *coll)
{
   int type;
   struct Collector * __coll;
   struct OMPT_Profile * prof;
   struct OMPT_Profile_acc *acc;

   if( event->type != MALP_EVENT_OMPT )
      return;
   
   assert( coll );
   assert( event );
   
   __coll = ( struct Collector * ) coll;
   prof = (struct OMPT_Profile *) __coll->internal_data;

   type = event->event.ompt.type; 

   // Update the global profile
   prof->global_acc.hits[type]++;
   prof->global_acc.time[type]+= event->duration;
   
   // Update the perrank profile
   acc = Hash_Table_get( &prof->per_stream_profiles, event->tid , NULL, OMPT_Profile_acc_test);
 
   if(! acc )
   {
      struct OMPT_Profile_acc tmp;
      memset( &tmp, 0, sizeof( struct OMPT_Profile_acc ) );
      tmp.associated_tid = event->tid;
      acc = Hash_Table_push( &prof->per_stream_profiles, (void *)&tmp , event->tid );
   }

   acc->hits[type]++;
   acc->time[type]+= event->duration;
}

static int OMPT_Profile_reduce( struct Collector * coll )
{
   uint64_t tmp_buff[T_OMPT_COUNT];
	
   struct OMPT_Profile *prof = (struct OMPT_Profile *)coll->internal_data;
   struct OMPT_Profile_acc *global_acc = &prof->global_acc;

   // HITS
   PMPI_Reduce(global_acc->hits, tmp_buff, T_OMPT_COUNT, MPI_UNSIGNED_LONG_LONG , MPI_SUM, 0, VMPI_Get_partition_comm());
   memcpy( global_acc->hits, tmp_buff, T_OMPT_COUNT * sizeof(uint64_t ) );

   // TIME
   PMPI_Reduce(global_acc->time, tmp_buff, T_OMPT_COUNT, MPI_UNSIGNED_LONG_LONG , MPI_SUM, 0, VMPI_Get_partition_comm());
   memcpy( global_acc->time, tmp_buff, T_OMPT_COUNT * sizeof(uint64_t ) );
	
   // SIZE
   PMPI_Reduce(global_acc->size, tmp_buff, T_OMPT_COUNT, MPI_UNSIGNED_LONG_LONG , MPI_SUM, 0, VMPI_Get_partition_comm());
   memcpy( global_acc->size, tmp_buff, T_OMPT_COUNT * sizeof(uint64_t ) );
   return 0;
}

static json_t * OMPT_Profile_gen( struct OMPT_Profile_acc * acc, int render_min_max )
{
   int i;
   json_t * profile = json_array();
   
   for(i = 0 ; i < T_OMPT_COUNT ; i++ )
   {
      if( !( acc->hits[i] ) ) continue; 

      json_t * function_entry = json_object();
      json_t * name = json_string( (char *) OMPT_Event_names[i] );
      json_t * hits = json_int( acc->hits[i] );
      json_t * time = json_real( (double) acc->time[i] / Timer_second() );
      json_t * size = json_real( acc->size[i] );

      json_object_set( function_entry, "name", name);
      json_object_set( function_entry, "hits", hits);
      json_object_set( function_entry, "time", time);
      json_object_set( function_entry, "size", size);
			
      json_decref(name);
      json_decref(hits);
      json_decref(time);
      json_decref(size);
			
      json_array_push( profile, function_entry );
      json_decref( function_entry );
			
   }
   return profile;
}

static void OMPT_Profile_render_per_rank_profile(void *acc, void *pcoll)
{
   json_t *profile, *entry, *tid;
   struct OMPT_Profile_acc* __acc;
   char entry_name[100], stid[100];
   struct Stream_Descriptor *stream_desc;

   __acc = (struct OMPT_Profile_acc *) acc;

   const uint64_t associated_tid = __acc->associated_tid;
  
   profile = OMPT_Profile_gen( __acc, 0 );  
   stream_desc = Meta_Data_Handler_get_desc( __acc->associated_tid );

   if( !stream_desc ) return;

   const uint64_t partition_id = stream_desc->partition_id;

   entry = json_object();
   snprintf(entry_name, 100, "%ld/OMPT_Profile/streams/%lu", partition_id, associated_tid );
   snprintf( stid, 100, "%lu", associated_tid );
   
   tid = json_string( stid );
   json_object_set( entry, "tid", tid );
   json_decref( tid );
   json_object_set( entry, "profile", profile );
   json_decref( profile );

   nodeSharedCache_reduce( &nsc, entry_name, entry);

   json_decref( entry );
}

static int OMPT_Profile_accumulate( struct Collector * coll )
{
   int rank;
   struct OMPT_Profile * prof;

   assert( coll );
   prof = (struct OMPT_Profile *) coll->internal_data;
   assert( prof ); 

   PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
	
   // Each rank renders its own profiles
   Hash_Table_walkthrought_2p( &prof->per_stream_profiles, (void *)coll, OMPT_Profile_render_per_rank_profile  );
	
   if( !rank )
   {
      char uri[100];
      json_t * profile;
      struct OMPT_Profile_acc *global_acc;

      global_acc = &prof->global_acc;
      assert( global_acc );

      // Do the JSON rendering
      profile =  OMPT_Profile_gen( global_acc, 1 );
      snprintf(uri, 100, "%s/profile", coll->analysis_cache_name );
      nodeSharedCache_set( &nsc, uri, profile );
      json_decref( profile );
   }

   return 0;
}

#endif /* HAVE_OMPT_H */


void OMPT_Profile_init( struct Collector* coll, uint64_t partition_id )
{
#if HAVE_OMPT_H
        Collector_init( coll, MALP_EVENT_OMPT,
                                        partition_id,
                                        "OMPT_Profile",
                                        OMPT_Profile_init_coll,
                                        OMPT_Profile_release_coll,
                                        OMPT_Profile_push_event,
                                        OMPT_Profile_reduce,
                                        OMPT_Profile_accumulate );
#endif /* HAVE_OMPT_H */
}
