#include "config.h"

#include "Time_Matrix.h"
#include "MPI_Events.h"

#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include <math.h>

#if HAVE_MPI_T
   #include "MPIT_Events.h"
   #include "MALP_MPI_T_data.h"
#endif /* HAVE_MPI_T */

static void
Time_Matrix_init_coll_MALP_EVENT_MPI( struct Time_Matrix *tm, const uint64_t time_slices, const uint64_t partition_size )
{
   int i,j;
   assert( tm );
	
   for( i = 0 ; i < T_MPI_COUNT ; i++ )
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_init( &tm->mpi_grids[i][j] , time_slices, partition_size);
}

void
Time_Matrix_init_coll_MALP_EVENT_WRAPPED( struct Time_Matrix* time_matrix, const uint64_t time_slices, const uint64_t partition_size )
{
   int i,j;
   assert( time_matrix );

   for( i = 0 ; i < T_POSIX_COUNT ; i++ )
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_init( &( time_matrix->posix_grids[i][j] ), time_slices, partition_size);	
}
#if HAVE_OMPT_H
static void 
Time_Matrix_init_coll_MALP_EVENT_OMPT( struct Time_Matrix* time_matrix, const uint64_t time_slices, const uint64_t partition_size )
{
   int i,j;
   struct Sample_Grid** grids;

   assert( tm );
   grids = t->ompt_grids;

   for( i = 0 ; i < T_OMPT_COUNT ; i++ )
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_init( &( grids[i][j] ) , time_slices, partition_size);
}
#endif /* HAVE_OMPT_H */

#if HAVE_MPI_T
static void
Time_Matrix_init_coll_MALP_EVENT_MPI_T( struct Time_Matrix* time_matrix, const uint64_t time_slices, const uint64_t partition_size )
{
   int i,j;
   struct Sample_Grid** grids;

   assert( time_matrix );

   const int __cvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_CVAR);
   grids = (struct Sample_Grid**) malloc( sizeof(struct Sample_Grid*) * __cvar_count );
   assert( grids );
   time_matrix->mpi_t_cvar_grids = grids;

   for( i = 0 ; i < __cvar_count ; i++ )
   {
      grids[i] = (struct Sample_Grid*) malloc( sizeof(struct Sample_Grid) * GRID_COUNT );
      assert( grids[i] );
           
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_init( &( grids[i][j] ), time_slices, partition_size);
   }

   const int __pvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_PVAR);
   grids = (struct Sample_Grid**) malloc( sizeof(struct Sample_Grid*) * __pvar_count );
   assert( grids );
   time_matrix->mpi_t_pvar_grids = grids;

   for( i = 0 ; i < __pvar_count ; i++ ) 
   {
      grids[i] = (struct Sample_Grid*) malloc( sizeof(struct Sample_Grid) * GRID_COUNT ); 
      assert( grids[i] );

      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_init( &( grids[i][j] ), time_slices, partition_size);
   }
}
#endif /* HAVE_MPI_T */

int  Time_Matrix_init_coll( struct Collector * coll )
{
   int rank;
   struct Time_Matrix* time_matrix = NULL;

   assert( coll );
   PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

   const uint64_t time_slices = 512;
   const uint64_t partition_size = Collector_partition_size( coll );

   assert( partition_size > 0 );

   time_matrix = malloc( sizeof( struct Time_Matrix ) );
   assert( time_matrix );	
   coll->internal_data = time_matrix;

   Time_Matrix_init_coll_MALP_EVENT_MPI( time_matrix, time_slices, partition_size );
   Time_Matrix_init_coll_MALP_EVENT_WRAPPED( time_matrix, time_slices, partition_size );
#if HAVE_OMPT_H
   Time_Matrix_init_coll_MALP_EVENT_OMPT( time_matrix, time_slices, partition_size );
#endif /* HAVE_OMPT_H */
#if HAVE_MPI_T
   Time_Matrix_init_coll_MALP_EVENT_MPI_T( time_matrix, time_slices, partition_size );
#endif /* HAVE_MPI_T */
   return 0;
}

static void
Time_Matrix_release_coll_MALP_EVENT_MPI( struct Time_Matrix *tm )
{
   int i,j;
   assert( tm );

   for( i = 0 ; i < T_MPI_COUNT ; i++ )
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_release( &tm->mpi_grids[i][j] );
}

static void
Time_Matrix_release_coll_MALP_EVENT_WRAPPED( struct Time_Matrix *tm )
{
   int i,j;
   assert( tm );

   for( i = 0 ; i < T_POSIX_COUNT ; i++ )
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_release( &tm->posix_grids[i][j] );
}

#if HAVE_MPI_T
static void
Time_Matrix_release_coll_MALP_EVENT_MPI_T( struct Time_Matrix *tm )
{
   int i,j;
   assert( tm );

   const int __t_mpi_t_cvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_CVAR);
   for( i = 0 ; i < __t_mpi_t_cvar_count ; i++ )
   {
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_release( &tm->mpi_t_cvar_grids[i][j] ); 

      free( tm->mpi_t_cvar_grids[i] );
   }
   free( tm->mpi_t_cvar_grids ); tm->mpi_t_cvar_grids = NULL; 

   const int __t_mpi_t_pvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_PVAR);
   for( i = 0 ; i < __t_mpi_t_pvar_count ; i++ )
   {
      for( j = 0 ; j < GRID_COUNT ; j++ )
         Sample_Grid_release( &tm->mpi_t_pvar_grids[i][j] ); 

      free( tm->mpi_t_pvar_grids[i] );
   }
   free( tm->mpi_t_pvar_grids ); tm->mpi_t_pvar_grids = NULL; 
}
#endif /* HAVE_MPI_T */

int Time_Matrix_release_coll( struct Collector * coll )
{
   struct Time_Matrix *tm = (struct Time_Matrix *)coll->internal_data;
   assert( tm );

   Time_Matrix_release_coll_MALP_EVENT_MPI( tm );
   Time_Matrix_release_coll_MALP_EVENT_WRAPPED( tm );
#if HAVE_MPI_T
   Time_Matrix_release_coll_MALP_EVENT_MPI_T( tm );
#endif /* HAVE_MPI_T */
   free( coll->internal_data ); coll->internal_data = NULL;

   return 0;
}

static void 
Time_Matrix_push_event_MALP_EVENT_MPI(struct MALP_Trace_Event *event, struct Time_Matrix *tm )
{
   struct MPI_Event *mpievt;

   assert( event->type == MALP_EVENT_MPI );

   mpievt = &( event->event.mpi_event );

   const int type = mpievt->type;
   double __duration = event->duration / Timer_second();
   double __timestamp = event->timestamp / Timer_second();

   //fprintf(stderr, "TS %lld TD %lld  TSS %lld\n", __timestamp, __duration, Timer_second() );

   /* Update the global profile */
   Sample_Grid_accumulate( &( tm->mpi_grids[type][GRID_HITS] ), event->rank, __timestamp, 1 );
   Sample_Grid_accumulate( &( tm->mpi_grids[type][GRID_TIME] ), event->rank, __timestamp, __duration );

   if( mpievt->size )
	Sample_Grid_accumulate( &( tm->mpi_grids[type][GRID_SIZE] ), event->rank, __timestamp, mpievt->size );
}

static void 
Time_Matrix_push_event_MALP_EVENT_WRAPPED(struct MALP_Trace_Event *event, struct Time_Matrix *tm )
{
   struct Wrapped_Event *we;

   assert( event->type == MALP_EVENT_WRAPPED );

   we = &( event->event.wrapped_event );

   const int type = we->type;  
   double __duration = event->duration / Timer_second();
   double __timestamp = event->timestamp / Timer_second();

   /* Update the global profile */
   Sample_Grid_accumulate( &( tm->posix_grids[type][GRID_HITS] ), event->rank, __timestamp, 1 );
   Sample_Grid_accumulate( &( tm->posix_grids[type][GRID_TIME] ), event->rank, __timestamp, __duration );

   if( we->size )
	Sample_Grid_accumulate( &(tm->posix_grids[type][GRID_SIZE]), event->rank, __timestamp, we->size );

}

#if HAVE_MPI_T
static void 
Time_Matrix_push_event_MALP_EVENT_MPI_T(struct MALP_Trace_Event *event, struct Time_Matrix *tm )
{
   struct Sample_Grid** mpit_grids; 
   struct MALP_MPI_T_event_s* mpit_evt;

   assert( event->type == MALP_EVENT_MPI_T );

   mpit_evt = &( event->event.mpit );

   if( mpit_evt->type != MALP_MPI_T_TYPE_POLL_PVAR ) return;

  if (!MALP_mpi_t_is_numeric(mpit_evt))
  {
      return;
  }

  double val = MALP_mpi_t_to_double(mpit_evt);


   const int type = mpit_evt->logical_var_idx;  
   double __duration = event->duration / Timer_second();
   double __timestamp = event->timestamp / Timer_second();

    //fprintf(stderr, "TS IS %lld\n", __timestamp);

   mpit_grids = ( mpit_evt->type == MALP_MPI_T_TYPE_POLL_PVAR ) ? tm->mpi_t_pvar_grids : tm->mpi_t_cvar_grids; 
  
   /* Update the global profile */
   Sample_Grid_accumulate( &( mpit_grids[type][GRID_TIME] ), event->rank, __timestamp, 1 );
   Sample_Grid_accumulate( &( mpit_grids[type][GRID_HITS] ), event->rank, __timestamp, 1 );
   Sample_Grid_accumulate( &( mpit_grids[type][GRID_SIZE] ), event->rank, __timestamp, val );
}
#endif /* HAVE_MPI_T */


void Time_Matrix_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
  struct Collector *coll = (struct Collector *)pcoll;
  struct Time_Matrix *tm = (struct Time_Matrix *)coll->internal_data;
	
   switch( event->type )
   {
      case MALP_EVENT_MPI: 
         Time_Matrix_push_event_MALP_EVENT_MPI( event, tm );
         break;
      case MALP_EVENT_WRAPPED:
         Time_Matrix_push_event_MALP_EVENT_WRAPPED( event, tm );
         break;
#if HAVE_MPI_T
      case MALP_EVENT_MPI_T:
         Time_Matrix_push_event_MALP_EVENT_MPI_T( event, tm );
         break;
#endif /* HAVE_MPI_T */
      default:
         break;
   }
}


void Time_Matrix_perform_reduce( struct Collector * coll , struct Sample_Grid *grid )
{
	int j;
	
	for( j = 0 ; j < GRID_COUNT ; j++ )
	{
			if( grid[j].has_data )
			{
				/* Do a '0' hit over the matrix in order to make all matrix having the same quantum */
				double end_time = Duration_infos_get_end( coll->partition_id )/ Timer_second(); 
				Sample_Grid_scale( &grid[j], end_time );
				
				/* We only flatten temporal grids */
				if( j == GRID_TIME )
					Sample_Grid_flatten( &grid[j], 1 );
			}
			
			
			Sample_Grid_reduce( &grid[j] );
			
	}
}

static void 
Time_Matrix_reduce_MALP_EVENT_MPI( struct Collector * coll, struct Time_Matrix *tm )
{
   int i;
   assert( coll && tm );
   for( i = 0 ; i < T_MPI_COUNT ; i++ )
     Time_Matrix_perform_reduce( coll, tm->mpi_grids[i] );
}

static void 
Time_Matrix_reduce_MALP_EVENT_WRAPPED( struct Collector * coll, struct Time_Matrix *tm )
{
   int i;
   assert( coll && tm );
   for( i = 0 ; i < T_POSIX_COUNT ; i++ )
     Time_Matrix_perform_reduce( coll, tm->posix_grids[i] );
}

#if HAVE_MPI_T
static void
Time_Matrix_reduce_MALP_EVENT_MPI_T( struct Collector * coll, struct Time_Matrix *tm )
{
   int i;
   assert( coll && tm );

   const int __t_mpi_t_cvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_CVAR);
   for( i = 0 ; i < __t_mpi_t_cvar_count ; i++ )
     Time_Matrix_perform_reduce( coll, tm->mpi_t_cvar_grids[i] );

   const int __t_mpi_t_pvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_PVAR);
   for( i = 0 ; i < __t_mpi_t_pvar_count ; i++ )
     Time_Matrix_perform_reduce( coll, tm->mpi_t_pvar_grids[i] );
}
#endif /* HAVE_MPI_T */

int Time_Matrix_reduce( struct Collector * coll )
{
   struct Time_Matrix *tm = (struct Time_Matrix *)coll->internal_data;
   assert( tm );

   Time_Matrix_reduce_MALP_EVENT_MPI( coll, tm );
   Time_Matrix_reduce_MALP_EVENT_WRAPPED( coll, tm );	
#if HAVE_MPI_T
   Time_Matrix_reduce_MALP_EVENT_MPI_T( coll, tm );	
#endif /* HAVE_MPI_T */
   return 0;
}

json_t * Time_Matrix_gen_entry( struct Sample_Grid * grids, char *name )
{
	json_t * ret = json_object();
	
	json_t * jname = json_string( name );
	json_object_set( ret, "name", jname );
	json_decref( jname );
	
	json_t * measures = json_object();
	
	int i;
	
	for( i = 0 ; i < GRID_COUNT ; i++ )
	{
		if( !grids[i].has_data )
		{
			continue;
		}
		
		json_t * tmp = Sample_Grid_dump_JSON( &grids[i] );
		
		json_object_set( measures, (char *)grid_types_names[i], tmp );
		json_decref( tmp );
	}
	
	
	json_object_set( ret, "measures", measures );
	
	json_decref( measures );
	
	return ret;
}

static void
Time_Matrix_accumulate_MALP_EVENT_MPI( struct Time_Matrix *tm, json_t* measures_array )
{
   int i;
   json_t *__tmp_ret;
	
   for( i = 0 ; i < T_MPI_COUNT ; i++ )
   {
      if( !( tm->mpi_grids[i][0].has_data ) )
         continue;

      __tmp_ret = Time_Matrix_gen_entry( tm->mpi_grids[i], (char *)MPI_Event_names[ i ]  );
      json_array_push( measures_array, __tmp_ret );
      json_decref( __tmp_ret ); 
   }
}

static void
Time_Matrix_accumulate_MALP_EVENT_WRAPPED( struct Time_Matrix *tm, json_t* measures_array )
{
   int i; 
   json_t *__tmp_ret;

   for( i = 0 ; i < T_POSIX_COUNT ; i++ )
   {
      if( !( tm->posix_grids[i][0].has_data ) )
         continue;
      
      __tmp_ret = Time_Matrix_gen_entry( tm->posix_grids[i], (char *) Wrapped_symbol_names[ i ]  );
      json_array_push( measures_array, __tmp_ret );	
      json_decref( __tmp_ret );
   }
}

#if HAVE_MPI_T
static void
Time_Matrix_accumulate_MALP_EVENT_MPI_T( struct Time_Matrix *tm, json_t* measures_array )
{
   int i; 
   json_t *__tmp_ret;
   char* __tmp_var_name;

   const int __cvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_CVAR);
   for( i = 0; i < __cvar_count; i++ )
   {
       if( !( tm->mpi_t_cvar_grids[i][0].has_data ) )
          continue;
       
       __tmp_var_name = __malp_mpi_t_get_var_name( MALP_MPI_T_TYPE_POLL_CVAR, i); 
       if( !( __tmp_var_name ) )
          continue;

       __tmp_ret = Time_Matrix_gen_entry( tm->mpi_t_cvar_grids[i], __tmp_var_name );
       free( __tmp_var_name ); __tmp_var_name = NULL;
       json_array_push( measures_array, __tmp_ret );	
   }

   /* MPI_T PVARS */   
   const int __pvar_count = __malp_mpi_t_get_num_vars(MALP_MPI_T_TYPE_POLL_PVAR);
   for( i = 0; i < __pvar_count; i++ )
   {

       if( !( tm->mpi_t_pvar_grids[i][0].has_data ) )
          continue;

       __tmp_var_name = __malp_mpi_t_get_var_name( MALP_MPI_T_TYPE_POLL_PVAR, i); 
       if( !( __tmp_var_name ) )
          continue;

       __tmp_ret = Time_Matrix_gen_entry( tm->mpi_t_pvar_grids[i], __tmp_var_name );
       free( __tmp_var_name ); __tmp_var_name = NULL;
       json_array_push( measures_array, __tmp_ret );	
   }
}
#endif /* HAVE_MPI_T */

int Time_Matrix_accumulate( struct Collector * coll )
{
   int rank;
   json_t *time_matrices, *measures;
   struct Time_Matrix *tm = (struct Time_Matrix *)coll->internal_data;

   PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

   if( rank ) return 0;

   /* Only process 0 from here */
   measures = json_array();
   time_matrices = json_object();
	
   Time_Matrix_accumulate_MALP_EVENT_MPI( tm, measures );	
   Time_Matrix_accumulate_MALP_EVENT_WRAPPED( tm, measures );
#if HAVE_MPI_T
   Time_Matrix_accumulate_MALP_EVENT_MPI_T( tm, measures );
#endif /* HAVE_MPI_T */
   json_object_set( time_matrices, "measures", measures );
   json_decref( measures );
	
   char uri[100];
   snprintf(uri, 100, "%s", coll->analysis_cache_name );
   nodeSharedCache_set( &nsc, uri, time_matrices );
	
   json_decref( time_matrices );
   return 0;
}

void Time_Matrix_init(struct Collector *coll, uint64_t partition_id)
{
		
	Collector_init( coll, MALP_EVENT_ANY,
					partition_id,
					"Time_Matrix",
					Time_Matrix_init_coll,
					Time_Matrix_release_coll,
					Time_Matrix_push_event,
					Time_Matrix_reduce,
					Time_Matrix_accumulate );
	
}
