#include "Topology.h"

#include "Topology.h"
#include "MPI_Events.h"

#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include "MPI_Event_class.h"



void Topology_acc_init(struct Topology_acc *acc, uint64_t psize)
{
	
	Array_init_2D( &acc->hits, psize , psize ,1);
	Array_init_2D( &acc->time, psize , psize ,1);
	Array_init_2D( &acc->size, psize , psize ,1);
	
	acc->filled = 1;
}

void Topology_acc_release(struct Topology_acc *acc)
{
	if(!acc->filled )
		return;
		
	Array_release( &acc->hits );
	Array_release( &acc->time );
	Array_release( &acc->size );
	
	acc->filled = 0;
}

void Topology_acc_reduce(struct Topology_acc *acc, uint64_t psize)
{
	uint64_t array_size = psize * psize;


	/* Does everybody has the buffer ? */
	int allocated = acc->filled;
	int to_allocate = 0;
	PMPI_Allreduce( &allocated, &to_allocate, 1, MPI_INT, MPI_SUM, VMPI_Get_partition_comm());
	
	if( !allocated && to_allocate )
	{
		Topology_acc_init(acc, psize);
	}
	
	allocated = to_allocate;
	
	if( allocated )
	{
		Array_reduce( &acc->hits, MPI_SUM );
		Array_reduce( &acc->time, MPI_SUM);
		Array_reduce( &acc->size, MPI_SUM );
	}
}


int  Topology_init_coll( struct Collector * coll )
{
	coll->internal_data = malloc( sizeof( struct Topology ) );
	
	if( !coll->internal_data )
	{
		perror("malloc");
		abort();
	}
	
	memset( coll->internal_data, 0, sizeof( struct Topology ) );
	
	struct Topology *topo = (struct Topology *)coll->internal_data;
	VMPI_Partition_desc * desc = VMPI_Get_desc_by_id(coll->partition_id);
	
	if( !desc )
	{
		printf("Failed to retrieve desc\n ");
		return 1;
	}
	
	uint64_t psize = desc->size;
	topo->psize = psize;
	
	return 0;
}


int  Topology_release_coll( struct Collector * coll )
{
	struct Topology * topo = (struct Topology *)coll->internal_data;

	int i;
	
	for( i = 0 ; i < COMM_MAT_T_MPI_COUNT; i++ )
	{
			Topology_acc_release(&topo->acc[i]);
	}


	free( coll->internal_data );

	return 0;
}

void Topology_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
	struct Collector *coll = (struct Collector *)pcoll;
	struct Topology *topo = (struct Topology *)coll->internal_data;


	if( event->type != MALP_EVENT_MPI )
	{
		printf("BAD event type\n");
		return;
	}
	

	struct MPI_Event *mpievt = &event->event.mpi_event;
	int type = mpievt->type;

	if( !Event_is_MPI_p2p( type ) )
		return;
	

	uint64_t source = mpievt->source;
	uint64_t dest =  mpievt->dest;

	if( (topo->psize < source) || (topo->psize < dest) )
		return;

	int p2p_offset =  MPI_Comm_translate_mpi_event_p2p_type( type );
	
	if( !topo->acc[p2p_offset].filled )
		Topology_acc_init( &topo->acc[p2p_offset], topo->psize );

	/* Update topology */

	Array_inc_2D( &topo->acc[p2p_offset].hits, source, dest );
	Array_inc_val_2D( &topo->acc[p2p_offset].time, event->duration, source, dest );
	
	if( mpievt->size )
		Array_inc_val_2D( &topo->acc[p2p_offset].size, mpievt->size, source, dest );

}

int  Topology_reduce( struct Collector * coll )
{
	struct Topology *topo = (struct Topology *)coll->internal_data;

	int i;

	for( i = 0 ; i < COMM_MAT_T_MPI_COUNT; i++ )
	{
		Topology_acc_reduce(&topo->acc[i], topo->psize);
	}
	
	return 0;
}



int  Topology_accumulate( struct Collector * coll )
{
	struct Topology * topo = (struct Topology *)coll->internal_data;

	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
	

	if( rank != 0 )
		return 1;

	printf("ACCUM\n");
	
		
	/* Here 0 has everything thanks to the reduction */
	int i;

	for( i = 0 ; i < COMM_MAT_T_MPI_COUNT; i++ )
	{
		struct Topology_acc *currentacc = &topo->acc[i];
		
		if( !currentacc->filled )
			continue;
		
		json_t * matrix = json_array();
		
		
		uint64_t x,y;
		int not_empty = 0;
		
		for( x = 0 ; x < topo->psize; x++ )
		{
			for( y = 0 ; y < topo->psize; y++ )
			{
				if( Array_get_2D( &currentacc->hits, x, y ) )
				{
						not_empty = 1;
						
						json_t * matrix_entry = json_object();
					
						json_t *src = json_int( x );
						json_t *dest = json_int( y );
						
						json_t *hits = json_real(Array_get_2D( &currentacc->hits, x, y ));
						json_t *time = json_real(Array_get_2D( &currentacc->time, x, y )/Timer_second());
						json_t *size = json_real(Array_get_2D( &currentacc->size, x, y ) );
						
						
						json_object_set( matrix_entry, "src", src );
						json_object_set( matrix_entry, "dest", dest );
						
						json_object_set( matrix_entry, "hits", hits );
						json_object_set( matrix_entry, "time", time );
						json_object_set( matrix_entry, "size", size );
						
						json_array_push( matrix, matrix_entry );
						
						json_decref( src );
						json_decref( dest );
						
						json_decref( hits );
						json_decref( time );
						json_decref( size );
					
						json_decref( matrix_entry );
				}
				
			}	
		}
		
		
		if( not_empty )
		{
			char uri[100];
			snprintf(uri, 100, "%s/%s", coll->analysis_cache_name, MPI_Event_names[MPI_Comm_translate_comm_mat_to_type(i)] );
			
			printf("Setting %s\n", uri );
			
			nodeSharedCache_set( &nsc, uri, matrix );
		}
		
		
		json_decref( matrix );

	}

	return 0;
}



void Topology_init(struct Collector *coll, uint64_t partition_id)
{
		
	Collector_init( coll, MALP_EVENT_MPI,
					partition_id,
					"Topology",
					Topology_init_coll,
					Topology_release_coll,
					Topology_push_event,
					Topology_reduce,
					Topology_accumulate );
	
}
