#include "MPI_Section_tree.h"
#include <string.h>
#include <stdio.h>

/*
 * SECTION Ballance
 */

int MPI_Section_balance_init( struct MPI_Section_balance * bal , uint64_t cells )
{
	if( !bal )
		return 1;

	if( cells == 0 )
	{
		fprintf(stderr, "Error cannon allocate empty ballancing\n");
		return -1;
	}
	
	bal->ballance = calloc( cells, sizeof( uint64_t ) );
	
	if( !bal->ballance )
	{
		perror("calloc");
		return -1;
	}
	
	bal->cell_count = cells;
	
	bal->start_offset = 0;
	bal->quantum = 1;
	
	return 0;
}

int MPI_Section_balance_release( struct MPI_Section_balance * bal )
{
	if( !bal )
		return 1;
	
	free( bal->ballance );
	
	memset( bal, 0, sizeof( struct MPI_Section_balance ) );
	
	return 0;
}

int MPI_Section_balance_hit( struct MPI_Section_balance * bal , uint64_t value )
{
	/* Is the target in the interval ? */
	if( (bal->start_offset <= value) 
	&& ( value < ( bal->start_offset + bal->cell_count * bal->quantum ) ) )
	{
		/* Direct HIT */
		uint64_t target_cell = (value - bal->start_offset) / bal->quantum;

		/* For security */
		if( target_cell < bal->cell_count )
		{
			bal->ballance[ target_cell ] ++;
		}
		else
		{
			fprintf(stderr, "Error Target cell overflows array\n");
			return -1;
		}
		
	}
	else
	{
		uint64_t diff = 0;
		uint64_t * new_array = calloc( bal->cell_count, sizeof( uint64_t ) );
		
		if( !new_array )
		{
			perror("calloc");
			return -1;
		}
		
		uint64_t old_start_offset = bal->start_offset;
		uint64_t old_quantum = bal->quantum;
		
		/* We need to resample */
		if( value < bal->start_offset )
		{
			//printf("Ballance Underflow \n");
			/* Underflow */
			diff = bal->start_offset - value;
			
			/* Fix start offset and quantum */
			bal->start_offset -= diff;
			
			/* recompute quantum so that the
			 * range remains unchanged */
			uint64_t old_range = old_start_offset + bal->cell_count * old_quantum;
			bal->quantum = (old_range - bal->start_offset) / bal->cell_count;
		}
		else
		{
			//printf("Ballance Overfow [%lld over %lld] \n", value , bal->start_offset + bal->cell_count * bal->quantum);
			/* Overflow */
			diff = value - bal->start_offset + bal->cell_count * bal->quantum;
			/* Update only the quantum */
			uint64_t new_range = bal->start_offset + diff + 1 + bal->cell_count * bal->quantum;
			bal->quantum = (new_range - bal->start_offset) / bal->cell_count; 
		}
		
		
		/* Repush sampled points */
		uint64_t i;
		
		for( i = 0 ; i < bal->cell_count ; i++ )
		{
			if( bal->ballance[i] )
			{
				uint64_t new_target_cell = (i * old_quantum) / bal->quantum;
				
				/* Should not happen */
				if( new_target_cell < bal->cell_count )
				{
					new_array[ new_target_cell ] += bal->ballance[i];
				}
				else
				{
					fprintf(stderr, "Error Target cell overflows array when resampling\n");
				}
			}
		}
		
		/* Replace array */
		free( bal->ballance );
		bal->ballance = new_array;

		/* Repush on the resampled array */
		return MPI_Section_balance_hit( bal, value );
	}
	
	return 0;
}



int MPI_Section_balance_walk( struct MPI_Section_balance * bal, 
				void (*call)( uint64_t off, uint64_t val, void * arg),
				void * arg )
{
	uint64_t i;
	
	if( !call || !bal )
	{
		return -1;
	}
	
	for( i = 0 ; i < bal->cell_count ; i++ )
	{
		if( bal->ballance[i] )
		{
			uint64_t off = bal->start_offset + i * bal->quantum;
			(call)( off, bal->ballance[i], arg );
		}
	}
	
	return 0;
}



void MPI_Section_balance_render_push( uint64_t off, uint64_t val, void * arg)
{
	json_t * ret = json_array();
	
	//printf("%g :: %lld \n", (double)off, val);
	
	json_t * time_off = json_real( (double)off/Timer_second() );
	json_t * value = json_real( (double) val );
	
	json_array_push(ret, time_off);
	json_decref( time_off );
	
	json_array_push(ret, value);
	json_decref( value );
	
	/* Append to output */
	json_t * out = (json_t *)arg;
	json_array_push(out, ret);
	json_decref( ret );
}

json_t * MPI_Section_balance_render( struct MPI_Section_balance * bal )
{
	json_t * ret = json_array();
	
	MPI_Section_balance_walk( bal, MPI_Section_balance_render_push, (void *)ret );
	
	return ret;
}


/*
 * SECTION TREE
 */


int MPI_Section_tree_init_static( struct MPI_Section_tree * tree, char * label, uint64_t cells, void * payload )
{

	tree->lock = 0;
	
	tree->generic_payload = payload;
	tree->cell_count = cells;
	
	tree->hits = 0;
	tree->total_time = 0;
	tree->exclusive_time = 0;
	tree->current_ts = 0;
	
	if( MPI_Section_balance_init( &tree->bal, tree->cell_count ) < 0 )
	{
		return -1;
	}
	
	tree->siblings_count = 0;
	tree->siblings = NULL;
	tree->parent = NULL;
	
	snprintf( tree->label, 64, "%s", label );
	
	return 0;
}


struct MPI_Section_tree * MPI_Section_tree_new( struct MPI_Section_tree * parent, char * label )
{
	struct MPI_Section_tree * ret = calloc( 1, sizeof( struct MPI_Section_tree ) );
	
	if( !ret )
	{
		perror("calloc");
		return NULL;
	}

	
	if( !parent )
	{
		fprintf(stderr, "MPI_Section_tree_new should be called with a parent\n");
		return NULL;
	}

	/* Initialize the static part */
	printf("Inheriting %lld from %p [%s]\n", parent->cell_count ,parent, parent->label);
	if( MPI_Section_tree_init_static( ret, label, parent->cell_count, parent->generic_payload ) < 0 )
	{
		return NULL;
	}

	ret->parent = parent;
	
	return ret;
}

int MPI_Section_tree_release( struct MPI_Section_tree * tree )
{
	if( tree->parent )
	{
		fprintf(stderr, "MPI_Section_tree_release should be called on root element\n");
		return -1;
	}
	
	int i;
	
	/* First go down */
	for( i = 0 ; i < tree->siblings_count ; i++ )
	{
		if( MPI_Section_tree_release( tree->siblings[i] ) < 0 )
		{
			return -1;
		}
	}
	
	/* Now free content */
	if( MPI_Section_balance_release( &tree->bal ) < 0 )
	{
		return -1;
	}
	
	int had_parent = (tree->parent)?1:0;
	
	memset( tree, 0, sizeof( struct MPI_Section_tree ) );
	
	/* We do not free the root (static init) */
	if( had_parent )
		free( tree );
	
	return 0;
}

struct MPI_Section_tree * MPI_Section_tree_enter( struct MPI_Section_tree * tree, char * section_label, uint64_t timestamp )
{
	struct MPI_Section_tree * ret = NULL;
	
	/* First do we have a previous tree ? */
	if( !tree )
	{
		fprintf(stderr, "Error any MPI_Section tree must have a root (static init)\n");
		return NULL;
	}
	
	/* Now is the tree already present in the siblings */
	int i;
	
	for( i = 0 ; i < tree->siblings_count ; i++ )
	{
		if( !strcmp( tree->siblings[i]->label , section_label ) )
		{
			/* Found ! */
			ret = tree->siblings[i];
			break;
		}
	}
	
	/* Allocate a new siblings if not found */
	if( !ret )
	{
		ret = MPI_Section_tree_new( tree, section_label );
		
		if( !ret )
		{
			return NULL;
		}
		
		/* Add it to the siblings table */
		tree->siblings_count++;
		
		tree->siblings = realloc( tree->siblings, tree->siblings_count * sizeof( struct MPI_Section_tree * ) );
		
		if( !tree->siblings )
		{
			perror("calloc");
			return NULL;
		}
		
		tree->siblings[tree->siblings_count-1] = ret;
		
	}
	
	/* Now set entry TS */
	
	ret->current_ts = timestamp;
	ret->hits++;
	
	return ret;
}

struct MPI_Section_tree * MPI_Section_tree_leave( struct MPI_Section_tree * tree, char * section_label, uint64_t timestamp )
{
	if( !tree )
	{
		return NULL;
	}
	
	if( !tree->parent )
	{
	  fprintf(stderr, "Error MPI_Section underflow detected\n");
	  return NULL;
	}
	
	/* Check that Labels are Matching */
	if( strcmp( tree->label, section_label ) )
	{
	  fprintf(stderr, "Error MPI_Section mismatch (%s , %s)\n", tree->label,  section_label);
	  return NULL;
	}
	
	/* Accumulate Time */
	uint64_t duration = timestamp - tree->current_ts;
	
	//printf("DURATION is %lld (%lld - %lld)\n", duration , timestamp, tree->current_ts);
	
	tree->total_time += duration;
	
	tree->current_ts = timestamp;
	
	/* Hit the Ballance array */
	if( MPI_Section_balance_hit( &tree->bal , duration ) < 0 )
	{
	  fprintf(stderr, "Error failled to hitorize MPI_Section duration (%s)\n", section_label);
	}
	
	/* Return parent to move UP */
	return tree->parent;
}

uint64_t __MPI_Section_tree_compute_stats( struct MPI_Section_tree * tree )
{
	if( tree->siblings_count )
	{
		int i;
		uint64_t child_time = 0;
		
		for( i = 0 ; i < tree->siblings_count ; i++ )
		{
			 child_time += __MPI_Section_tree_compute_stats( tree->siblings[i] );
		}
		
		/* Now set exclusive time */
		tree->exclusive_time = tree->total_time - child_time;
	}
	
	return tree->total_time;
}

void MPI_Section_tree_compute_stats( struct MPI_Section_tree * tree )
{
	__MPI_Section_tree_compute_stats( tree );
}



int MPI_Section_tree_render( struct MPI_Section_tree * tree, json_t * parent_array )
{
	json_t * label = json_string( tree->label );
	
	json_t * hits = json_int( tree->hits );
	json_t * total_time = json_real( (double)tree->total_time / Timer_second()  );
	json_t * exclusive_time = json_real( (double)tree->exclusive_time / Timer_second()  );
	json_t * children = json_array();
	
	json_t * balance = MPI_Section_balance_render( &tree->bal );
	
	json_t * ret = json_object();
	
	json_object_set( ret, "label", label );
	json_decref( label );
	
	json_object_set( ret, "hits", hits );
	json_decref( hits );
	
	json_object_set( ret, "total_time", total_time );
	json_decref( total_time );
	
	json_object_set( ret, "exclusive_time", exclusive_time );
	json_decref( exclusive_time );
	
	json_object_set( ret, "balance", balance );
	json_decref( balance );
	
	json_object_set( ret, "children", children );
	json_decref( children );

	if( tree->siblings_count )
	{
		int i;
		for( i = 0 ; i < tree->siblings_count ; i++ )
		{
			MPI_Section_tree_render( tree->siblings[i] , children );
		}	
	}
	
	json_array_push( parent_array, ret );
	json_decref( ret );
	
	return 1;
}








