/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"

#include <mpi.h>
#include <string.h>
#include <stdlib.h>
#include "Timer.h"

/*
 * 
 *	DURATION HANDLER
 * 
 */

struct Duration_infos * ___duration_infos = NULL;


void Duration_infos_init()
{
	___duration_infos = calloc( VMPI_Get_partition_count() , sizeof( struct Duration_infos ) );
	
	if( !___duration_infos )
	{
		perror( "calloc") ;
		exit(1);
	}
}

void Duration_infos_reduce()
{
	uint64_t min, max, duration;
	int i;

	for( i = 0 ; i < VMPI_Get_partition_count() ; i++ )
	{
		struct Duration_infos *infos = Duration_infos_get( i );
		min = Duration_infos_get_begin( i );
		max = Duration_infos_get_end( i );
		duration = Duration_infos_get_duration( i );

		PMPI_Allreduce(&min, &infos->begin, 1, MPI_LONG_LONG_INT, MPI_MIN,  VMPI_Get_partition_comm());
		PMPI_Allreduce(&max, &infos->end, 1, MPI_LONG_LONG_INT, MPI_MAX,  VMPI_Get_partition_comm());
		PMPI_Allreduce(&duration, &infos->duration, 1, MPI_LONG_LONG_INT, MPI_MAX,  VMPI_Get_partition_comm());
	}
	
	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
	
	if( rank == 0 )
	{
		/* Now that we have the durations rank 0 updates the cache */
		json_t * partitions = nodeSharedCache_get( &nsc, "partitions" );
		
		if( !partitions )
		{
			printf("Error could not retrieve partition list\n");
			abort();
		}
		
		int i;
		int len = json_to_array(partitions)->size;
		
		for( i = 0 ; i < len ; i++ )
		{
			struct Duration_infos *current_infos = Duration_infos_get( i );
			json_t * current_partition = json_array_get( partitions, i );
			
			if( !current_partition || (current_partition->type == JSON_NULL) )
				continue;
			
			json_t * js_start_time = json_real( current_infos->begin / Timer_second() );
			json_t * js_end_time = json_real( current_infos->end / Timer_second() );
			json_t * js_duration = json_real( current_infos->duration / Timer_second() );
			
			json_object_set( current_partition, "start_time", js_start_time );
			json_object_set( current_partition, "end_time", js_end_time );
			json_object_set( current_partition, "duration", js_duration );
			
			json_decref( js_start_time );
			json_decref( js_end_time );
			json_decref( js_duration );
			
			json_decref( current_partition );
			
		}
		
		nodeSharedCache_commit( &nsc, "partitions" );

		json_decref( partitions );
		
	}

}


void Duration_infos_release()
{
	free ( ___duration_infos );
	___duration_infos = NULL;
}



uint64_t Duration_infos_get_begin( uint64_t partition_id )
{
	if( VMPI_Get_partition_count() < partition_id )
			return 0;
	
	return ___duration_infos[partition_id].begin;
}

uint64_t Duration_infos_get_end( uint64_t partition_id )
{
	if( VMPI_Get_partition_count() < partition_id )
		return 0;
	
	
	return ___duration_infos[partition_id].end;
}

uint64_t Duration_infos_get_duration( uint64_t partition_id )
{
	if( VMPI_Get_partition_count() < partition_id )
		return 0;
	
	return ___duration_infos[partition_id].duration;
}

/*
 * 
 *	META DATA 
 * 
 */


struct Meta_Data_Handler __meta_data_handler;


void _Meta_Data_Handler_push(struct MALP_Trace_Event *container , void *arg)
{
	struct Stream_Descriptor *desc = &container->event.stream_descriptor;
	Meta_Data_Handler_push( desc );
}


void Meta_Data_Handler_init(int rank, int size)
{
	memset( Meta_Data_Handler(), 0, sizeof( struct Meta_Data_Handler) );
	Hash_Table_init( &Meta_Data_Handler()->entries, sizeof( struct Meta_Data_Handler_cell), 4241, MALP_MDH_COLL_AVOID );
	Duration_infos_init();
}

void Meta_Data_Handler_release()
{
	Hash_Table_release( &Meta_Data_Handler()->entries, NULL );
	memset( Meta_Data_Handler(), 0, sizeof( struct Meta_Data_Handler) );
	Duration_infos_release();
}

void Meta_Data_Handler_unify()
{
	Duration_infos_reduce();
}


void Update_Stream_cache( struct Stream_Descriptor *desc )
{
	char stid[100];
	snprintf( stid, 100, "%lu", desc->tid );	
	json_t * tid = json_string( stid );

	json_t * parent_tid = json_int( desc->parent_tid );
	json_t * partition_id = json_int( desc->partition_id );
	json_t * type = json_string( (char *)Stream_type_names[desc->type] );

	json_t * pid = json_int( desc->pid );
	json_t * thread_id = json_int( desc->thread_id );
	
	json_t * start_time = json_real( desc->start_time / Timer_second() );
	json_t * end_time = json_real( desc->end_time  / Timer_second() );
	
	json_t * mpi_rank = json_int( desc->mpi_rank );
	json_t * omp_thread = json_int( desc->omp_thread );
	
	json_t * hostname = json_string( desc->hostname );
	
	json_t* jdesc = json_object();

	
	json_object_set( jdesc, "tid" , tid );
	json_object_set( jdesc, "parent_tid" , parent_tid );
	json_object_set( jdesc, "partition_id" , partition_id );
	json_object_set( jdesc, "type" , type );
	
	json_object_set( jdesc, "pid" , pid );
	json_object_set( jdesc, "thread_id" , thread_id );
	
	json_object_set( jdesc, "start_time" , start_time );
	json_object_set( jdesc, "end_time" , end_time );
	
	json_object_set( jdesc, "mpi_rank" , mpi_rank );
	json_object_set( jdesc, "omp_thread" , omp_thread );
	
	json_object_set( jdesc, "hostname" , hostname );
	
	json_decref( tid );
	json_decref( parent_tid );
	json_decref( partition_id );
	json_decref( type );
	
	json_decref( pid );
	json_decref( thread_id );
	
	json_decref( start_time );
	json_decref( end_time );
	
	json_decref( mpi_rank );
	json_decref( omp_thread );
	
	json_decref( hostname );

	snprintf( stid, 100, "streams/%lu", desc->tid );
	nodeSharedCache_reduce( &nsc, stid, jdesc);
}





int Meta_Data_Handler_push( struct Stream_Descriptor *desc )
{
	void *pret;
	
	MALP_Spinlock_lock(&Meta_Data_Handler()->lock);

	if( desc->desc_type == DESC_BEGIN )
	{

		pret = Hash_Table_push( &Meta_Data_Handler()->entries, (void *)desc , desc->tid );
		
		if( !pret )
		{
			MALP_Spinlock_unlock(&Meta_Data_Handler()->lock);
			return 0;
		}
		
		Meta_Data_Handler()->entries_count++;
	}
	else /* DESC_END => Update duration infos */
	{
		struct Stream_Descriptor *tmp = Meta_Data_Handler_get_desc( desc->tid );
		

		if( !tmp )
		{
			printf("Warning end event with no associated desc !\n");
			MALP_Spinlock_unlock(&Meta_Data_Handler()->lock);
			return -1;
		}
		
		tmp->end_time = desc->end_time;
		
		/* Update Partitions duration info if we have a process */
		if( tmp->type == MALP_PROCESS )
		{
			struct Duration_infos *duration = Duration_infos_get( tmp->partition_id  );
			
			if( !duration )
			{
				printf("Warning unknown partition in end event\n");
				MALP_Spinlock_unlock(&Meta_Data_Handler()->lock);
				return -1;
			}
			
			/* Here we update start and end times if needed */
			if( !duration->begin  || tmp->start_time < duration->begin )
				duration->begin = tmp->start_time;
			if( !duration->end  || duration->end < tmp->end_time  )
				duration->end = tmp->end_time;
			
			//printf("Duration is now %lld for %lld\n", duration->duration, tmp->partition_id );
			
			duration->duration = duration->end - duration->begin;
		}
		
				
		Update_Stream_cache( tmp );
	}
	
	MALP_Spinlock_unlock(&Meta_Data_Handler()->lock);
	
	return 1;
}

int Meta_Data_Handler_test_func(void *pelem, uint64_t id, void *dummy)
{
	struct Stream_Descriptor *sd = (struct Stream_Descriptor *)pelem;
	
	if( sd->tid == id )
		return 1;
	
	return 0;
}


void Stream_Descriptor_print( struct Stream_Descriptor *desc )
{
	
	printf("======================\n");
	printf("TID : %llu\n", (long long unsigned int)desc->tid );
	printf("PAR : %llu\n", (long long unsigned int)desc->parent_tid );
	printf("TYPE : %d\n", desc->desc_type );
	printf("STREAM TYPE : %d\n", desc->type );
	printf("START : %llu\n",(long long unsigned int) desc->start_time );
	printf("END : %llu\n", (long long unsigned int)desc->end_time );
	printf("RANK : %llu\n",(long long unsigned int) desc->mpi_rank );
	printf("OMP : %llu\n", (long long unsigned int)desc->omp_thread );
	printf("HOST : %s\n", desc->hostname );
	printf("======================\n");
	
}
