/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#ifndef METADATA_HANDLER_H
#define METADATA_HANDLER_H

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdint.h>

#include "Spinlock.h"
#include "Hash_Table.h"
#include "Event_Desc.h"
#include "VMPI.h"


struct Duration_infos
{
	uint64_t begin;
	uint64_t end;
	uint64_t duration;
};


void Duration_infos_init();
void Duration_infos_release();
void Duration_infos_reduce();

extern struct Duration_infos * ___duration_infos;

uint64_t Duration_infos_get_begin( uint64_t partition_id );
uint64_t Duration_infos_get_end( uint64_t partition_id );
uint64_t Duration_infos_get_duration( uint64_t partition_id );


static inline struct Duration_infos * Duration_infos_get( uint64_t partition_id  )
{
	if( VMPI_Get_partition_count() < partition_id )
		return NULL;

	return &___duration_infos[partition_id];
}




#define MALP_MDH_COLL_AVOID 10


/**
 * @ingroup Meta_Data_Handler_cell_
 * @brief Structure of a meta data handler cell
 *
 * Used to store stream descriptors
 */
struct Meta_Data_Handler_cell
{
	struct Hash_Table_Internal _____;
	struct Stream_Descriptor descriptors[MALP_MDH_COLL_AVOID];
};

/**
 * @ingroup Meta_Data_Handler_
 * @brief Structure of a meta data handler
 *
 * Used to store placement/hierachy/position/identification data and stream descriptor information
 */
struct Meta_Data_Handler
{
	MALP_Spinlock lock;			/* Spinlock */
	struct Hash_Table entries;	/**< Hash table grouping all stream infos */
	uint64_t entries_count;		/**< Number of entries stored */
};

/**
 * @ingroup Meta_Data_Handler_
 * @brief current meta data handler
 */
extern struct Meta_Data_Handler __meta_data_handler;

/**
 * @ingroup Meta_Data_Handler_
 * @brief getter for the meta data handler
 * @return a reference to the meta data handler
 */
static inline struct Meta_Data_Handler *Meta_Data_Handler()
{
	return &__meta_data_handler;
}


/**
 * @ingroup Meta_Data_Handler_
 * @brief meta data handler initialisation
 * @param rank
 * @param size
 */
void Meta_Data_Handler_init( int rank, int size);

/**
 * @ingroup Meta_Data_Handler_
 * @brief meta data handler release/freeing by releasing the hash map entries and freeing the allocated memory for it.
 */
void Meta_Data_Handler_release();

/**
 * @ingroup Meta_Data_Handler_
 * @brief Update a meta data handler entries with a new stream descriptor or update the duration of an existing one
 * @param desc
 * @return update status (-1: error, 1: success)
 */
int Meta_Data_Handler_push( struct Stream_Descriptor *desc );

/**
 * @ingroup Meta_Data_Handler_
 * @brief Update a metata handler entries using the stream descriptor of a MALP trance event
 * @param container
 * @param arg (unused)
 */
void _Meta_Data_Handler_push(struct MALP_Trace_Event *container , void *arg);

/**
 * @ingroup Meta_Data_Handler_
 * @brief Display the main informations (id, rank, parent, duration etc...) of a Stream descriptor
 * @param desc
 */
void Stream_Descriptor_print( struct Stream_Descriptor *desc );

/* Internals */
/**
 * @ingroup Meta_Data_Handler_
 * @brief [Internals] Test if an object is a stream descriptor with a given id
 * @param pelem
 * @param id
 * @param dummy (unused)
 * @return test status (1: desc is a stream descriptor with id as an identifier, 0: wrong identifier or object)
 */
int Meta_Data_Handler_test_func(void *pelem, uint64_t id, void *dummy);

/**
 * @ingroup Meta_Data_Handler_
 * @brief [Internals] Get a stream descriptor from the entries of a meta data handler using its id.
 * @param id
 * @return a stream descriptor from the entries hash table
 */
static inline struct Stream_Descriptor * Meta_Data_Handler_get_desc( uint64_t id )
{
	struct Stream_Descriptor *ret;
	

	ret = (struct Stream_Descriptor *)Hash_Table_get( &Meta_Data_Handler()->entries, id , NULL, Meta_Data_Handler_test_func);
	
	return ret;
}



#ifdef __cplusplus
}
#endif


#endif /* METADATA_HANDLER_H */
