#include "Sample_Grid.h"

#include <mpi.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>

#include "VMPI.h"
#include "Bit_Array.h"
#include "Binary_Tree.h"


void Sample_Grid_init( struct Sample_Grid *g , uint64_t time_slices, uint64_t slots )
{
	g->grid = calloc( slots , sizeof( struct Sample_Grid * ) );
	
	if( !g->grid )
	{
		perror("calloc");
		abort();
	}
	
	g->grid_size_x = time_slices;
	g->slot_count = slots;
        assert( g->slot_count > 0 );
	
	g->grid = calloc( slots, sizeof( double ) );
	
	if( !g->grid )
	{
		perror("calloc");
		abort();
	}
	
	g->min = 0;
	g->max = 0;
	
	g->qx = 0;
	
	g->has_data = 0;
}

void Sample_Grid_release( struct Sample_Grid *g )
{
	int i;
	
	for( i = 0 ; i  < g->slot_count; i++ )
	{
		free( g->grid[i] );
	}
	
	
	free( g->grid );
	memset( g, 0, sizeof( struct Sample_Grid ) );
}


void Sample_Grid_flatten( struct Sample_Grid *g, int multiplicity )
{
	if( !g->has_data )
		return;
	
	
	double q = g->qx;
	
	int i,j;
	
	/* Over lines (ranks) */
	for( i = 0 ; i < g->slot_count; i++ )
	{
		/* Over colls (samples) */
		
		double *line = g->grid[i];

		if( ! line )
		{
			continue;
		}
		
		double *new_line = calloc( sizeof(double), g->grid_size_x );
		
		if( !new_line )
		{
			perror("malloc");
			abort();
		}
		
		for( j = 0 ; j < g->grid_size_x; j++ )
		{
			double current_val = line[j]/multiplicity;
			
			int k;
			
			for( k = j ; k < g->grid_size_x; k++ )
			{
					if( current_val <= 0 )
						break;
				
					double to_write = (q < current_val ) ? ( q ) : current_val;
		
					new_line[k] += to_write;
						
					current_val -= to_write;	

			}
		}
		
		memcpy( line, new_line, sizeof(double) * g->grid_size_x );
		
		free( new_line );	
	}
	
	
	
	
}


void Sample_Grid_accumulate( struct Sample_Grid *g ,  unsigned int slot, double offset, double value )
{
	int64_t grid_offset = -1;
	
	if( g->qx !=  0.0 )
	{
		/* Try to fit in the grid */
		grid_offset = roundf( offset / g->qx );
	}
	
	if( g->slot_count <= slot )
	{
		//printf("Slot overflow %g\n", offset );
		return;
                //abort();
	}
	
		 	
	double *sample_line = NULL;
	
	
	if( (g->grid_size_x <= grid_offset)||(grid_offset == -1) )
	{

		/* We do not fit or we have no quantum 
		 * 
		 * Then we compute a new quantum such
		 * as offset * qx == 0.77 grid_size_x
		 * 
		 * */
		
		double old_qx = g->qx;
		
		/* New QX */
		if( offset )
		{
			g->qx = offset/(0.9 * g->grid_size_x);
		}
		
		/* By definition if we have to resample the array is too small
		 * thus we can resample in one pass by acumulating cells to the
		 * left and computing its offset
		 * if it does cross cell border then we accumulate to the left.
		 */
		 
		 int i, j = 0;
		 
		 
		 if( old_qx )
		 {
			 for( j = 0 ; j < g->slot_count ; j++ )
			 {
					sample_line = g->grid[j];
					
					if( !sample_line )
					{
						continue;
					}

					/* We put the Y offsetting outside as it is only supposed to happen
					 * at reduction time */
			
					 for( i = 0 ; i < g->grid_size_x ; i++ )
					 {
						 uint64_t new_offset = roundf( (i*old_qx) / g->qx );
						 
						 if( new_offset != i )
						 {

							 if( i < new_offset) 
							 {
								 printf("Error you are pushing to the right (%g --> %g) !\n",old_qx, g->qx );
								 abort();
							 }

							/*Aggregate in the offset */
							sample_line[new_offset] = (sample_line[i] + sample_line[new_offset] );
							sample_line[i] = 0;
						 } 
					 }
			
			 }
		 
		}
		 
		 /* Now that we resampled simply recompute the offset */
		 grid_offset = offset / g->qx;

	}
	
	/* After resampling target sample line */

	/* Check if this line is allocated */
	if( !g->grid[slot] )
	{
		/* If not do allocate IT */
		g->grid[slot] = calloc( g->grid_size_x, sizeof( double ) );
		
		if( !g->grid[slot] )
		{
			perror("calloc");
			abort();
		}
		
	}

	g->has_data |= 1;
	

	sample_line = g->grid[slot];
	
	
	/* Set value */
	sample_line[grid_offset] += value;
}

void Sample_Grid_scale( struct Sample_Grid *g ,double offset )
{
	int i;
	
	for( i = 0 ; i  < g->slot_count; i++ )
	{
		if( g->grid[i] )
		{
			Sample_Grid_accumulate( g , i, offset, 0.0 );
			break;
		}
	}
}


void Sample_Grid_get_min_max( struct Sample_Grid *g , double *min, double *max )
{
	if( !min || !max )
		return;
	
	if( !g->qx )
	{
		
		return;
	}
	
	double current_min = -1.0;
	double current_max =  -1.0;
	
	int i,j;
	
	for( j=0 ; j < g->slot_count ; j++ )
	{
		double *current_line = g->grid[j];
		
		for( i=0 ; i < g->grid_size_x ; i++ )
		{
			if( !current_line )
				continue;
				
			if( (current_min == -1.0) || current_line[i] < current_min )
				current_min = current_line[i];
			
			if( (current_max == -1.0) || current_max < current_line[i] )
				current_max = current_line[i];
		}
	}
	
	*min = current_min;
	*max = current_max;
}


void Sample_Grid_dump_text( struct Sample_Grid *g, char *dest )
{
	FILE *out = fopen( dest, "w");
	
	if( !out )
	{
		perror( "fopen" );
		return;
	}
	
	int i,j;
	
	if( !g->qx )
	{
		fprintf( out, "0.0 0.0\n");
		return;
	}
	
	for( j=0 ; j < g->slot_count ; j++ )
	{
		double *current_line = g->grid[j];
		
		for( i=0 ; i < g->grid_size_x ; i++ )
		{
			if( current_line )
				fprintf( out, "%g %d %g\n", (i)*g->qx, j, current_line[i] );
			else
				fprintf( out, "%g %d 0.0\n", (i)*g->qx, j );
		}
		
		fprintf( out, "\n");
	}
	
	fclose( out );
}


double Sample_Grid_get( struct Sample_Grid *g ,  unsigned int slot, double offset )
{
	if( ! g->grid[slot] )
		return 0.0;

	int64_t grid_offset = -1;
	
	if( g->qx !=  0.0 )
	{
		/* Try to fit in the grid */
		grid_offset = roundf( offset / g->qx );
	}

	if( grid_offset < 0 )
		return 0.0;

	double *current_line = g->grid[slot];
	
	return current_line[grid_offset];
}


json_t * Sample_Grid_dump_JSON( struct Sample_Grid *g )
{
	json_t * ret = json_object();
	json_t * array = json_array();
	
	int i,j;
	
	if( !g->qx )
	{
		return ret;
	}
	
				
	double min = 0.0;
	double max = 0.0;
	
	Sample_Grid_get_min_max( g , &min, &max );
	
	json_t * min_js = json_real( min );
	json_t * max_js = json_real( max );
	json_t * quantum = json_real( g->qx );
	
	json_object_set( ret, "min", min_js );
	json_object_set( ret, "max", max_js );
	json_object_set( ret, "q", quantum );
	
	json_decref( min_js );
	json_decref( max_js );
	json_decref( quantum );
	
	for( j=0 ; j < g->slot_count ; j++ )
	{
		double *current_line = g->grid[j];
		
		json_t * line_array = json_array();
		json_array_push( array , line_array );
		
		for( i=0 ; i < g->grid_size_x ; i++ )
		{
			
			
			json_t * elem = NULL;
			
			if( current_line )
				elem = json_real( current_line[i] );
			else
				elem = json_real( 0.0 );
			
			json_array_push( line_array, elem );
			json_decref( elem );
		}
		
		json_decref( line_array );
	}
	
	json_object_set( ret, "data", array );
	json_decref( array );
	
	return ret;
}




void Sample_Grid_reduce( struct Sample_Grid *g )
{
	int comm_size, comm_rank;
	int local_count = 0;
	
	PMPI_Comm_size( VMPI_Get_partition_comm(), &comm_size );
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &comm_rank );


	
	/* First we check is the grid has been hit */
	int tmp;
	
	PMPI_Allreduce( &g->has_data , &tmp, 1, MPI_INT, MPI_SUM, VMPI_Get_partition_comm() );
	
	g->has_data = tmp;
	
	if( !g->has_data )
	{
		/* Nothing to do no one has been hit */
		return;
	}
	
	//if( !comm_rank )
	//	printf("Digraph G{\n");
	PMPI_Barrier( VMPI_Get_partition_comm() );
		
			
	/* Then we allocate a bit array in order to describle our accounted
	 * children. We will use them to send only children from whom
	 * we have data */
	 
	struct Bit_Array known_slots;
	 
	Bit_Array_init( &known_slots, g->slot_count );
	 
	 /* We now flag the slots that we are accounting for already */
	 
	 int i,j;
	 

	 for( j = 0 ; j < g->slot_count ; j++ )
	 {
		 if( g->grid[j] != NULL )
		 {
			Bit_Array_set( &known_slots, j, 1);
			local_count++;
		 }

	 }

	
	
	/* Now we build a binary tree with ranks */
	struct Binary_Tree btree;
	Binary_Tree_init(&btree, comm_rank, comm_size);
	
	
	/*
	 * COMMUNICATION PHASE
	 *
	 * We first send the sampling rate store it
	 * then we send the data toward a temporary buffer
	 * then we push the data to the local array
	 * according to the sampling rate
	 *  
	 */
	
	/* Create a temporary buffer */
	double * temp_buffer = malloc( sizeof( double ) * g->grid_size_x );
	
	if( !temp_buffer )
	{
		perror("malloc");
		abort();
	}
	
	
	int lc = Binary_Tree_left_child( &btree );
	int rc = Binary_Tree_right_child( &btree );
	int pa = Binary_Tree_parent( &btree );
	
	double remote_sampling_rate = 0.0;
	
	int remote_count = 0;
	int target_rank = 0;
	
	PMPI_Barrier( VMPI_Get_partition_comm() );
	
	/* Left child */
	if( 0 <= lc )
	{
		/* Sampling rate */
		PMPI_Recv( &remote_sampling_rate, 1, MPI_DOUBLE, lc, 129, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );
		
		/* Recv number of remote entries */
		PMPI_Recv( &remote_count, 1, MPI_INT, lc, 130, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );

		int j = 0;

		for( j = 0 ; j < remote_count ; j++ )
		{
			/* Target Rank */
			PMPI_Recv( &target_rank, 1, MPI_INT, lc, 128, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );

			/* Buffer */
			PMPI_Recv( temp_buffer, g->grid_size_x , MPI_DOUBLE, lc, 128, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );
			
			/* Push */
			for( i = 0 ; i < g->grid_size_x; i++ )
			{
				 Sample_Grid_accumulate( g ,  target_rank, i * remote_sampling_rate , temp_buffer[i] );
			}
			
			if( !Bit_Array_get( &known_slots, target_rank ) )
			{
				Bit_Array_set( &known_slots, target_rank, 1);
				local_count++;
			}
		}
		
	}

	/* Right child */
	if( 0 <= rc )
	{
		/* Sampling rate */
		PMPI_Recv( &remote_sampling_rate, 1, MPI_DOUBLE, rc, 129, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );
		
		/* Recv number of remote entries */
		PMPI_Recv( &remote_count, 1, MPI_INT, rc, 130, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );

		for( j = 0 ; j < remote_count ; j++ )
		{
			/* Target Rank */
			PMPI_Recv( &target_rank, 1, MPI_INT, rc, 128, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );
			
			/* Buffer */
			PMPI_Recv( temp_buffer, g->grid_size_x , MPI_DOUBLE, rc, 128, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE );
			
			/* Push */
			for( i = 0 ; i < g->grid_size_x; i++ )
			{
				Sample_Grid_accumulate( g ,  target_rank, i * remote_sampling_rate , temp_buffer[i] );
			}

			if( !Bit_Array_get( &known_slots, target_rank ) )
			{
				Bit_Array_set( &known_slots, target_rank, 1);
				local_count++;
			}
		}
		
	}
	
	
	/* Send to parent */
	if( 0 <= pa )
	{
		/* Send Local sampling rate */
		PMPI_Send( &g->qx, 1 , MPI_DOUBLE, pa, 129, VMPI_Get_partition_comm() );
		
		/* Send number of entries */
		PMPI_Send( &local_count, 1 , MPI_INT, pa, 130, VMPI_Get_partition_comm() );

		/* Send individual entries */
		for( j = 0 ; j <  g->slot_count  ; j++ )
		{
			
			if( g->grid[j] )
			{
	
				//printf("%d -> %d [label=\"%d\"]\n", comm_rank, pa, j );
				/* Send target rank */
				PMPI_Send( &j, 1 , MPI_INT, pa, 128, VMPI_Get_partition_comm() );

				/* Send Buffer */
				PMPI_Send( g->grid[j], g->grid_size_x, MPI_DOUBLE, pa, 128, VMPI_Get_partition_comm() );
			
			}
		}
		
		
		
	}
	
	/* Free the temp buffer */
	free( temp_buffer );
	/* Release the Btree */
	Binary_Tree_release(&btree);
	/* Release known slots */
	Bit_Array_release( &known_slots );
	
	/* At this point root rank shall have the whole matrix */
	//if( !comm_rank )
	//	printf("}\n");
}


