/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

/**
 * @addtogroup internal_chained_list_
 * @{
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#ifndef CHAINED_LIST
#define CHAINED_LIST

#ifdef __cplusplus
  extern "C"
  {
#endif

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/**
 * @brief Basic buffered chained list
 *
 * This structure is not intended to be used directly. Chained_List must be used instead. 
 */
struct chained_list {
	void *p_payload;                /**< @brief payload array of trunk */
	uint32_t current_offset;        /**< @brief Current offset in trunk */
	uint32_t size_of_trunk;         /**< @brief Size of a trunk */
    size_t size_of_payload;         /**< @brief Size of the payload */
	struct chained_list *p_prev;    /**< @brief previous trunk */
};


/**
 * @brief Ininitalizes a buffered chained list
 * @param size_of_trunk size of the trunk in number of elems
 * @param size_of_payload size of a payload elem eg sizeof( TYPE )
 * @return pointer to an allocated chained_list
 */
struct chained_list *chained_list_init(uint32_t size_of_trunk, size_t size_of_payload);

/**
 * @brief Release a buffered chained list
 * @param list double pointer on a chained_list (set to NULL uppon release )
 */
void chained_list_release(struct chained_list **list);

/**
 * @brief Alocate a new chunk pointing to current_chunk (Recopy MODE )
 * @param current_chunk pointer to current_chunk
 * @return newly allocated chunk
 */
struct chained_list *chained_list_alloc_next_chunk(struct chained_list
        *current_chunk);

/**
 * @brief Alocate a new chunk  ( first chunk )
 * @param size_of_trunk in number of elems
 * @param size_of_payload elem eg sizeof( TYPE )
 * @return newly allocated chunk
 */
struct chained_list *chained_list_alloc_chunk(uint32_t size_of_trunk,
        size_t size_of_payload);

/**
 * @brief Count elements in the chained list
 * @param list pointer to an initialized chained list
 * @return number of elements in the list
 */
int chained_list_count(struct chained_list *list);

/**
 * @brief Push an element in the chained list
 * @param list double Pointer to an initialized chained list ( may be modified uppon push )
 * @param p_payload pointer to the elem
 * @return pointer to the pushed elem
 */
void *chained_list_push(struct chained_list **list, void *p_payload);

/**
 * @brief Push all element of source satisfying condition in dest
 * @param dest pointer to destination list
 * @param source pointer to source list
 * @param condition test function (takes a source elem as param) returns 1 for mergin
 */
void cond_chained_list_merge(struct chained_list **dest,
                             struct chained_list *source,
                             int (*condition) (void *));

/**
 * @brief Call a function uppon each element of the chained list
 * @param list pointer to a chained list list
 * @param action function (takes element in param as void *)
 */
void chained_list_walkthrough(struct chained_list *list,
                              void (*action) (void *));

/**
 * @brief Call a function uppon each element of the chained list
 * @param list pointer to a chained list list
 * @param action function (takes element in param as void *)
 * @param arg arguments to give to action (second void*)
 */
void chained_list_walkthrough_arg(struct chained_list *list,
                                  void (*action) (void *, void *), void *arg);
/**
 * @brief This struct is used to improve performance when sequentially walking through
 * 
 * It is used in Chained_List_get_nth_seq()
 */
struct chained_list_view {
	struct chained_list *last_chunk;    /**< @brief the chunk containing last element */
	uint64_t last_chunk_offset;         /**< @brief the offset of last element in last chunk */
};

/**
 * @brief Initializes a chained_list_view
 * @param clv the list view to initialize
 */
static inline void chained_list_view_init( struct chained_list_view *clv )
{
	clv->last_chunk = NULL;
	clv->last_chunk_offset = 0;
}

/**
 * @brief Get a pointer to the first element which matches a value
 * @param list pointer to a chained list list
 * @param test_func comparison function
 * @param value the wanted value
 * @return the found element (NULL if not found)
 */
void *chained_list_get_elem( struct chained_list *list, int (*test_func)(void *, void *), void *value );


/* Inlining of hot getters **/

/**
 * @brief Get a reference to the nth element of the chained list
 * @param list pointer to a chained list
 * @param n element index
 */
static inline void *chained_list_get_nth(struct chained_list *list, uint64_t n)
{
	uint64_t local_n = n + 1;

	//printf("BEGIN PROCESS %d ( in %p)\n", n, list);
	struct chained_list *tmp = list;

	while (tmp) {

		//printf("tmp % p OFF %d, local_n %d\n",tmp, tmp->current_offset, local_n);

		if ( (local_n <= tmp->current_offset) && (tmp->current_offset) ) {

			int applied_offset = ( tmp->current_offset *tmp->size_of_payload) - (local_n * tmp->size_of_payload);

			//printf("AP OFF FOR %lld is %d over %d\n", local_n , applied_offset,  tmp->current_offset  * tmp->size_of_payload );
			return (void *) ((char *)tmp->p_payload + applied_offset);
		}

		else {

			//printf("Going to next %p\n", tmp->p_prev);
			local_n -= tmp->current_offset;
			tmp = tmp->p_prev;
		}
	}

	// printf ("Get nth failed \n");
	return NULL;
}


/**
 * @brief Get a reference to the nth element of the chained list (optimised for sequential)
 * @param list pointer to a chained list list
 * @param view enhance sequential walking of the list by storing last chunk
 * @param n element index
 *
 * This function must be used in case the elements are get sequentially. 
 *
 * The chunk of the last found item is stored int view and the search starts from it, so the next call found next element much faster (especially when the list is hugh).
 */
static inline void *chained_list_get_nth_seq(struct chained_list *list, struct chained_list_view *view, uint64_t n)
{
	struct chained_list *tmp = list;
	n++;

	uint64_t tmp_n = 0;

	if( !list )
		return NULL;

	//If we already known an offset and it is lower than n
	if( view->last_chunk ) {
		if( view->last_chunk_offset < n ) {
			//printf("FAST FORWARD TO OFFSET %lld (%p) \n", view->last_chunk_offset, view->last_chunk);
			tmp = view->last_chunk;

			tmp_n = view->last_chunk_offset;
			n -= view->last_chunk_offset;
		}
	}


	if(  tmp->current_offset <= tmp->current_offset  ) {

		while( tmp ) {
			if(  (n <= tmp->current_offset) && (tmp->current_offset) ) {
				//Data is in this chunk
				//Save this chunk infos
				view->last_chunk = tmp;
				view->last_chunk_offset = tmp_n;

				//printf("SAVING CHUNK %p  OFF : %lld  \n", view->last_chunk, tmp_n );

				break;
			} else {
				n -= tmp->current_offset;
				tmp_n += tmp->current_offset ;
				tmp = tmp->p_prev;
			}
		}
	}

	//printf("CANDIDANDE N : %d over %d \n", n, tmp->current_offset );
	/* n was too large */
	if( !tmp )
		return NULL;

	int applied_offset = ( tmp->current_offset * tmp->size_of_payload ) - (n * tmp->size_of_payload);

	//printf("AP OFF FOR %lld is %d over %d\n", n , applied_offset,  tmp->current_offset  * tmp->size_of_payload );
	return (void *) ((char *)tmp->p_payload + applied_offset);
}

/**
 * @}
 */

/**
 * @addtogroup Chained_List_
 * @{
 */

/**
 * @brief This a structure wrapps the use of chained_list
 * This is the main chain list structure to use. 
 */
struct Chained_List {
	struct chained_list *list;  /**< @brief the head chained_list chunk */
	uint64_t count;             /**< @brief the number of elements in the list */
	uint64_t chunk_size;        /**< @brief the size of the chunks */
	size_t payload_size;        /**< @brief the size of elements that can be stored in the list */
};

/**
 * @brief Initializes a Chained_List
 * @param list the list ti initialize
 * @param chunk_size the wanted chunks size
 * @param payload_size the size of stored elements
 */
static inline void Chained_List_init( struct Chained_List *list, uint64_t chunk_size, size_t payload_size )
{
	list->list = NULL;
	list->count = 0;
	list->chunk_size = chunk_size;
	list->payload_size = payload_size;
}

/**
 * @brief Clears a Chained_List
 * @param list the list to clear
 */
static inline void Chained_List_release( struct Chained_List *list )
{
	chained_list_release( &list->list );
	list->count = 0;
}

/**
 * @brief Adds an element to the list
 * @param list the list where to add the element
 * @param payload a pointer on the element to add
 * @return a pointer on the pushed element
 *
 * The elements are pushed. It means that they are added at the head of the list. 
 */
static inline void* Chained_List_push( struct Chained_List *list, void *payload )
{
	if( list->list == NULL )
		list->list = chained_list_init( list->chunk_size, list->payload_size );

	void *ret = chained_list_push( &(list->list), payload );
	if( ret )
		list->count++;
	return ret;
}

/**
 * @brief Counts the elements in the list
 * @param list the list where to coune elements
 * @return the number of elements in the list
 */
static inline uint64_t Chained_List_count( struct Chained_List *list )
{
	return list->count;
}

/**
 * @brief Push all element of source satisfying condition in dest
 * @param dest pointer to destination list
 * @param source pointer to source list
 * @param condition test function (takes a source elem as param) returns 1 for mergin
 */
static inline void Chained_List_cond_merge(struct Chained_List *dest, struct Chained_List *source, int (*condition) (void *))
{
	if( source->list == NULL )
		return;
	
	if( dest->list == NULL )
		dest->list = chained_list_init( source->chunk_size, source->payload_size );

	cond_chained_list_merge( &dest->list, source->list, condition);
}

/**
 * @brief Call a function uppon each element of the chained list
 * @param list pointer to a chained list list
 * @param action function (takes element in param as void *)
 */
static inline void Chained_List_walkthrough(struct Chained_List *list, void (*action) (void *))
{
	if( list->list == NULL || list->count == 0 )
		return;

	chained_list_walkthrough( list->list, action);
}

/**
 * @brief Call a function uppon each element of the chained list
 * @param list pointer to a chained list
 * @param action function (takes element in param as void *)
 * @param arg arguments to give to action (second void*)
 */
static inline void Chained_List_walkthrough_arg(struct Chained_List *list, void (*action) (void *, void *), void *arg)
{
	if( list->list == NULL || list->count == 0)
		return;

	chained_list_walkthrough_arg( list->list, action, arg);
}

/**
 * @brief Gets the nth element of the list
 * @param list the list where to search the element
 * @param n the index of wanted element
 * @return a pointer to the element (NULL if not found)
 */
static inline void * Chained_List_get_nth(struct Chained_List *list, uint64_t n)
{
	if( list->list == NULL || list->count == 0)
		return NULL;

	return chained_list_get_nth( list->list, n);
}

/**
 * @brief Get a reference to the nth element of the chained list (optimised for sequential)
 * @param cl pointer to a chained list list
 * @param view enhance sequential walking of the list by storing last chunk
 * @param n element index
 *
 * This function must be used in case the elements are get sequentially. 
 *
 * The chunk of the last found item is stored int view and the search starts from it, so the next call found next element much faster (especially when the list is hugh).
 */
static inline void *Chained_List_get_nth_seq(struct Chained_List *cl, struct chained_list_view *view, uint64_t n)
{
	if( cl->list == NULL || cl->count == 0 )
		return NULL;

	return chained_list_get_nth_seq( cl->list, view, n);
}

/**
 * @brief Get a pointer to the first element which matches a value
 * @param list pointer to a chained list list
 * @param test_func comparison function
 * @param value the wanted value
 * @return the found element (NULL if not found or if list is empty)
 */
static inline void *Chained_List_get_elem( struct Chained_List *list, int (*test_func)(void *, void *), void *value )
{
	if( list->list == NULL )
		return NULL;

	return chained_list_get_elem( list->list, test_func, value );
}
#ifdef __cplusplus
}
#endif
#endif
  
/**
 * @}
 */
