/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

/**
 * @addtogroup Hash_Table_
 * @{
 */

#ifndef HASH_TABLE_H
#define HASH_TABLE_H


#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#include "Chained_List.h"
#include "Spinlock.h"
#include "CRC64.h"

/**
 * @brief This is the structure used to store the clusters (items that have the same hash)
 */
struct Hash_Table_Internal {
	MALP_Spinlock lock;             /**< @brief a lock for concurrent access */
	struct Chained_List next_pl;    /**< @brief the actual chained list (containing elements) */
	int count;                      /**< @brief the number of elements */
};

/**
 * @brief This is the main hask table structure. 
 */
struct Hash_Table {
	void *payload;                                      /**< @brief where data is stored (Hash_Table_Internal + cluster static array) */
	size_t sizeof_payload;                              /**< @brief the size of stored elements */
	size_t intern_payload_size;                         /**< @brief the size of the internal static array (coll_avoid * sizeof_payload) */
	uint64_t table_size;                                /**< @brief the size of the table */
	uint8_t need_lock;                                  /**< @brief set to 1 if internals need to use a lock before access */
	uint8_t __attribute__ ((aligned (64))) coll_avoid;  /**< @brief maximum number of elements stored in the cluster static array */
};

/**
 * @brief Indicates that the hashtable must use a lock for operations (get/push)
 * @param ht the hashtable where to indicate. 
 */
static inline void Hash_Table_use_lock( struct Hash_Table *ht )
{
	ht->need_lock = 1;
}

/* Interface */

/**
 * @addtogroup HashTableInitRelease
 * @{
 */

/**
 * @brief Hash_Table initialization
 * @param ht the hashtable to initialize
 * @param sizeof_payload the size of stored elements
 * @param table_size the size of the table
 * @param coll_avoid The size of the cluster static array
 */
void Hash_Table_init( struct Hash_Table *ht, size_t sizeof_payload, uint64_t table_size, uint8_t coll_avoid );

/**
 * @brief Hash_Table release
 * @param ht the hashtable to release
 * @param free_f release function for internal (must free what init_f passed to Hash_Table_init() allocates)
 */
void Hash_Table_release(  struct Hash_Table *ht, void (*free_f)(void *) );

/**
 * @}
 */

/**
 * @addtogroup HashTableAccessors
 * @{
 */

/**
 * @brief retrieves an element from the hashtable
 * @param ht where to retrieve the element
 * @param key the key of the wanted element
 * @param extra_arg extra argument to pass to test_func
 * @param test_func key comparison function. Takes a pointer to stored data, the key and an extra arg. returns 1 if test success
 * @return the found element
 */
void * Hash_Table_get( struct Hash_Table *ht, uint64_t key , void *extra_arg, int (*test_func)(void *, uint64_t , void *));

/**
 * @brief adds an element to a hashtable
 * @param ht where to add the element
 * @param payload the element to add
 * @param key the key where to add it
 * @return the added element
 */
void * Hash_Table_push(  struct Hash_Table *ht, void *payload , uint64_t key );

/**
 * @brief Retrieves the first found element of the hash table
 * @param ht the hash table where to get the element
 * @return the element (NULL if the table is empty)
 * @warning Only used in Unit Tests and Red_Ht_Table_get_an_entry() (which is never used...)
 */
void* Hash_Table_get_an_entry( struct Hash_Table *ht );

/**
 * @brief Adds an element into the hashtable. Replaces already existing element at key
 * @param ht where to add the element
 * @param payload the element to add
 * @param key where to add it
 * @param extra_arg extra argument to pass to test_func
 * @param test_func the same test function used with Hash_Table_get()
 * @return the added element
 */
void * Hash_Table_push_unique( struct Hash_Table *ht, void *payload , uint64_t key, void *extra_arg, int (*test_func)(void *, uint64_t, void * ) );

/**
 * @brief counts the elements in the hashtable by walking through all the elements
 * @param ht where to count elements
 * @return the number of elements
 */
uint64_t Hash_Table_count_by_walk(struct Hash_Table *ht);
/**
 * @}
 */

/**
 * @addtogroup HashTableWalkers
 * @{
 */

/**
 * @brief Executes a function on every stored elements
 * @param ht the hashtable where to walk through
 * @param action the function to execute on all elements (takes an element)
 */
void Hash_Table_walkthrought( struct Hash_Table *ht, void (*action)(void *) );

/**
 * @brief Executes a function on every stored elements
 * @param ht the hashtable where to walk through
 * @param second_arg an other argument to pass to action
 * @param action the function to execute on all elements (takes an element and an other argument)
 */
void Hash_Table_walkthrought_2p( struct Hash_Table *ht, void *second_arg,  void (*action)(void *, void *) );

/**
 * @brief Executes a function on every stored elements
 * @param ht the hashtable where to walk through
 * @param second_arg an other argument to pass to action
 * @param third_arg an other argument to pass to action
 * @param action the function to execute on all elements (takes an element and two other arguments)
 */
void Hash_Table_walkthrought_3p( struct Hash_Table *ht, void *second_arg, void *third_arg,  void (*action)(void *, void *, void *) );

/**
 * @brief Executes a function on every stored elements that satisfy a condition
 * @param ht the hashtable where to walk through
 * @param action what to execute on satisfying elements (takes an element)
 * @param test_func condition function. takes an element and the test_value. return 1 if success
 * @param test_value a test value to give to test_func
 */
void Hash_Table_walkthrought_cond( struct Hash_Table *ht, void (*action)(void *), int (*test_func)(void *, void * ), void *test_value );

/**
 * @brief Executes a function on every stored elements that satisfy a condition
 * @param ht the hashtable where to walk through
 * @param second_arg the second argument to pass to action
 * @param action what to execute on satisfying elements (takes an element and an other argument)
 * @param test_func condition function. takes an element and the test_value. return 1 if success
 * @param test_value a test value to give to test_func
 */
void Hash_Table_walkthrought_cond_2p( struct Hash_Table *ht, void *second_arg, void (*action)(void *, void *), int (*test_func)(void *, void * ), void *test_value );
/**
 * @}
 */

/**
 * @addtogroup HashTableInternals
 * @{
 */

/**
 * @brief Locks a cell at given key
 * @param ht the hashtable
 * @param key the key where to lock the cell
 */
void Hash_Table_lock_cell( struct Hash_Table *ht, uint64_t key );

/**
 * @brief Unlocks a cell at given key
 * @param ht the hashtable
 * @param key the key where to unlock the cell
 */
void Hash_Table_unlock_cell( struct Hash_Table *ht, uint64_t key );

/**
 * @brief Initializes internal hashtable data
 * @param phi the internal to initialize
 * @param coll_avoid  the number of elements in each chain list chunks
 * @param intern_payload_size the size of elements stored in the intern chained list
 */
void Hash_Table_init_Internal( void *phi, uint64_t coll_avoid, uint64_t intern_payload_size );

/**
 * @brief Releases internal hashtable data
 * @param phi the internal data to release
 */
void Hash_Table_release_Internal( void *phi );

/**
 * @}
 */


#ifdef __cplusplus
}
#endif

#endif

/**
 * @}
 */
