#ifndef __MALP_EMULE_ATOMICS_H__
#define __MALP_EMULE_ATOMICS_H__

#include "pthread.h"

extern pthread_mutex_t MALP_emule_atomics_lock;

typedef struct MALP_atomics_int_s
{
   volatile int val;
} MALP_atomics_int_t;

typedef struct MALP_atomics_ptr_s
{
   void* volatile val;
} MALP_atomics_ptr_t;

/* INTEGER FUNC */

static inline int 
MALP_atomics_load_int( MALP_atomics_int_t* ptr )
{
   int retval;
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   retval = ptr->val;
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
   return retval;
}
 
static inline void 
MALP_atomics_store_int( MALP_atomics_int_t* ptr, int val )
{
   int retval;
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   ptr->val = val;
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
}

static inline int
MALP_atomics_swap_int( MALP_atomics_int_t* ptr, int val )
{
   int prev;
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   prev = ptr->val;
   ptr->val = val;
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
   return (int) prev; 
}

static inline void
MALP_atomics_add_int( MALP_atomics_int_t* ptr, int val )
{
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   ptr->val += val;
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
}

static inline void
MALP_atomics_incr_int( MALP_atomics_int_t* ptr )
{
   MALP_atomics_add_int( ptr, 1 );
}

static inline void
MALP_atomics_decr_int( MALP_atomics_int_t* ptr )
{
   MALP_atomics_add_int( ptr, -1 );
}

static inline int
MALP_atomics_fetch_and_add_int( MALP_atomics_int_t* ptr, int val ) 
{
   int prev;
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   prev = ptr->val;
   ptr->val += val;
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
   return prev;
}

static inline int
MALP_atomics_fetch_and_incr_int( MALP_atomics_int_t* ptr ) 
{
   return MALP_atomics_fetch_and_add_int( ptr, 1 );
}

static inline int
MALP_atomics_fetch_and_decr_int( MALP_atomics_int_t* ptr ) 
{
   return MALP_atomics_fetch_and_add_int( ptr, -1 );
}

static inline int 
MALP_atomics_decr_and_test_int( MALP_atomics_int_t* ptr )
{
   int new_val;
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   ptr->val -= 1;
   new_val = ptr->val;
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
   return ( !new_val );
}

static inline int
MALP_atomics_cas_int( MALP_atomics_int_t* ptr, int old_val, int new_val )
{
   int prev;
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   prev = ptr->val;
   if( prev == old_val ) ptr->val = new_val;
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   return prev;
}

/* POINTER FUNC */
static inline void
MALP_atomics_store_ptr( MALP_atomics_ptr_t* ptr, void* val )
{
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   ptr->val = val; 
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
}

static inline void*
MALP_atomics_load_ptr( MALP_atomics_ptr_t* ptr )
{
   void* retval; 
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   retval = ptr->val; 
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
   return retval;
}

static inline void*
MALP_atomics_swap_ptr( MALP_atomics_ptr_t* ptr, void* val )
{
   void* prev; 
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   prev = ptr->val; 
   ptr->val = val;
   pthread_mutex_unlock( &MALP_emule_atomics_lock );
   return prev;
}

static inline void*
MALP_atomics_cas_ptr( MALP_atomics_ptr_t* ptr, void* old_val, void* new_val )
{
   void* prev;
   assert( ptr );
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   prev = ptr->val;
   if( prev == old_val ) ptr->val = new_val;
   pthread_mutex_lock( &MALP_emule_atomics_lock );
   return prev;
}

/* lock/unlock provides barrier */
#define MALP_atomics_write_barrier()      do {} while (0)
#define MALP_atomics_read_barrier()       do {} while (0)
#define MALP_atomics_read_write_barrier() do {} while (0)

#define MALP_ATOMICS_INT_T_INIT(val__) { (val__) }
#define MALP_ATOMICS_PTR_T_INIT(val__) { (val__) }

#endif /* __MALP_EMULE_ATOMICS_H__ */
