/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

/**
 * @addtogroup Common_Utils_
 * @{
 */

#include <sys/stat.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>

#ifndef TRACELIB_UTILS_H
#define TRACELIB_UTILS_H

/**
 * @brief Checks if a path is a directory
 * @param path the path to check
 * @return 1 if path is a directory, 0 otherwise
 */
static inline int MALP_Trace_check_dir(const char *path)
{
	struct stat statb;

	if (lstat(path, &statb) == -1) {
		return 0;
	}

	if (S_ISDIR(statb.st_mode)) {
		return 1;
	}

	return 0;
}

/**
 * @brief Creates a directory without failing if it already exists
 * @param path the path to the directory do create
 */
static inline void MALP_Trace_try_mkdir( char *path )
{
	if (mkdir(path, S_IRWXU) < 0) {
		if (!MALP_Trace_check_dir(path)) {
			printf("Failed to make directory : %s \n", path);
			abort();
		}
	}
}

/**
 * @brief Checks if a file exists
 * @param path the path to the file to check
 * @return 1 if the file exists, 0 otherwise
 */
static inline int MALP_Trace_file_exists( const char *path )
{
	struct stat statb;

	if (lstat(path, &statb) == -1) {
		return 0;
	}

	return 1;
}

/**
 * @brief Fast string comparison
 * @param haystack first string to compare
 * @param needle second string to compare
 * @return 1 if the strings are identical, 0 otherwise
 */
static inline int MALP_Trace_fast_strcmp( char *haystack, char *needle )
{
	if ( !haystack || !needle )
		return 0;

	while ( (*haystack != '\0') && (*needle !='\0') ) {
		if ( *haystack != *needle )
			return 0;

		haystack++;
		needle++;
	}

	if ( *haystack != *needle )
		return 0;

	return 1;
}

/**
 * @brief Checks if a string contains a numerical element
 * @param number the string ti check
 * @return 1 if the string represents a numeric, 0 otherwise
 */
static inline int MALP_Trace_is_numeric(char *number)
{
	while (*number != '\0' ) {
		if ( (*number < 48 || 57 < *number) && ( *number != '.') )
			return 0;

		number++;
	}
	return 1;
}

/**
 * @brief Checks if the first not space char of a string is a #
 * @param line the string to check
 * @return 1 if the string starts with a #, 0 otherwise
 */
static inline int MALP_Trace_line_begins_with_sharp(char *line)
{
	if ( !line )
		return 1;

	char *tmp = line;
	while (*tmp != '\0') {

		if (*tmp != '#' && *tmp != ' ')
			return 0;

		if (*tmp == '#' )
			return 1;

		tmp++;
	}

	return 0;
}

/**
 * @brief Getter on the total amount of memory available on current node 
 * @return the total memory
 */
static inline uint64_t MALP_Trace_total_mem()
{
	return (sysconf(_SC_PHYS_PAGES) *  sysconf(_SC_PAGE_SIZE) );
}

/**
 * @brief Getter on the per core memory on the current node
 * @return the per core memory
 */
static inline uint64_t MALP_Trace_mem_per_core()
{
	int core_count = sysconf(_SC_NPROCESSORS_CONF);
	return  MALP_Trace_total_mem()/core_count;
}

/**
 * @brief Removes starting and ending spaces from a string
 * @param val the string where to removes spaces
 * @return the same string as val with spaces removed
 */
static inline char * my_pad( char *val )
{

	int len;
	len = strlen(val);

	if( len == 0 )
		return val;


	while( val[len - 1] == ' ' )
	{
		val[len - 1] = '\0';
		len--;
	}

	while( *val == ' ' )
		val++;

	return val;
}


#endif

/**
 * @}
 */
