/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#ifndef MALP_MAP_PLUGIN
#define MALP_MAP_PLUGIN

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <stdint.h>
#include <time.h>
#include <stdlib.h>
#include "allinea_metric_plugin_types.h"
#include "MAP_Event.h"

/** Each metric can be in a domain 
 * Only TIME is valid for the moment 
 **/
typedef enum
{
	MAP_DOMAIN_NONE, /** Domain not set */
	MAP_DOMAIN_TIME, /** Metric is in TIME */
	MAP_DOMAIN_COUNT /** Number of metric domains */
}MAP_Metric_domain;

/** Convert a \ref MAP_Metric_domain to an human readable format */
static const char * const MAP_Metric_domain_name[ MAP_DOMAIN_COUNT ] =
{
	"none",
	"time"
};

/** This defines a MAP metric */
struct MAP_Metric
{
	char * id;								/** ID of this metric */
	char derivative;						/** ID of this metric */
	MAP_Metric_type data_type;				/** Return value type of the metric */
	MAP_Metric_domain domain;				/** Metric domain */
	void * data_collector;					/** Pointer to data collector */
	char * data_collector_ref;				/** Reference of data_collector */
	char data_collector_divide_by_sample;	/** Divide by sample the data */
	double previous_ts;						/** Has a previous TS entry */
	char * display_name;					/** Display name */
	char * description;						/** Description of the metric */
	char * type;							/** Type of this metric */
	char * color;							/** Display color of the metric */
	char * error_message;					/** Error message */
	int return_code;						/** Return code */
	void * payload;							/** Payload pointer */
};

/** Load a metric array from a given XML desc
 * @arg metric Pointer to the metric to be initialized
 * @arg desc XML node associated with the metric
 * @arg libhandle library handle to retrieve the source symbol
 * @return Non zero on error
 */
int MAP_Metric_init( struct MAP_Metric * metric, xmlNode * desc, void * libhandle );

/** Release a metric
 * @arg metric Pointer to the metric to be initialized
 */
int MAP_Metric_release( struct MAP_Metric * metric );


/** Abort while printing the error if present
 * @arg metric Pointer to the metric 
 */
static inline void MAP_Metric_check_error( struct MAP_Metric * metric )
{
	if( !metric )
		return;
	
	if( metric->error_message )
	{
		fprintf(stderr, "METRIC ERROR: %s (code %d)\n", metric->error_message, metric->return_code );
		abort();
	}
}

/** Set the payload in a Metric
 * @arg metric Pointer to the metric 
 * @arg payload Pointer to the payload
 */
static inline void MAP_Metric_set_payload( struct MAP_Metric * metric, void * payload )
{
	if( !metric )
		return;
	
	metric->payload = payload;
}

/** Get the payload in a Metric
 * @arg metric Pointer to the metric 
 * @return a pointer to the payload
 */
static inline void * MAP_Metric_get_payload( struct MAP_Metric * metric )
{
	if( !metric )
		return NULL;
	
	return metric->payload;
}

/** Forward declare \ref MAP_Metric_group */
struct MAP_Metric_group;

/** This defines a MAP Plugin gathering \ref MAP_Metric and \ref MAP_Metric_group */
struct MAP_Plugin
{
	char * source_id;							/** ID of the source */
	char * path_to_source_so;					/** Path to the .so */
	void * libhandle;							/** dlopen handle of the plugin .so */
	
	/** Functions to be loaded from the plugin */
	int (*allinea_plugin_initialize) (plugin_id_t plugin_id, void* data);
	int (*allinea_plugin_cleanup) (plugin_id_t plugin_id, void* data);
	int (*mymetric_getIntValue) (metric_id_t id, struct timespec *currentSampleTime, uint64_t *outValue);
	int (*mymetric_getDoubleValue) (metric_id_t id, struct timespec *currentSampleTime, double *outValue);
	
	struct MAP_Metric * metrics;				/** Extracted Metrics */
	unsigned int metric_count;					/** Number of metrics in this library */
	struct MAP_Metric_group * metric_groups;	/** Extracted Metric Groups */
	unsigned int metric_group_count;			/** Number of metric groups in this library */
	char * error_message; 						/** Error message */
	int return_code;							/** Return code */
};

/** Load a plugin from its description
 * @arg plugin pointer to the plugin
 * @arg xml_file path to the XML description file
 * @return Non zero on error */
int MAP_Plugin_load( struct MAP_Plugin * plugin, char * xml_file );

/** Release a plugin 
 * @arg plugin pointer to the plugin
 * @return Non zero on error */
int MAP_Plugin_release( struct MAP_Plugin * plugin );


/** Get a metric from the plgin using its id
 * @arg array Pointer to the plugin array 
 * @arg id Id of the metric to retrieve
 * @return Pointer to the metric or NULL
 */
static inline struct MAP_Metric * MAP_Plugin_get_metric( struct MAP_Plugin * plugin, unsigned int id )
{
	if( plugin->metric_count <= id )
	{
		return NULL;
	}
	
	return &plugin->metrics[ id ];
}


/** Abort while printing the error if present
 * @arg plugin pointer to the plugin
 */
static inline void MAP_Plugin_check_error( struct MAP_Plugin * plugin )
{		
	if( plugin->error_message )
	{
		fprintf(stderr, "PLUGIN ERROR: %s (code %d)\n", plugin->error_message, plugin->return_code );
		abort();
	}
}		



/** Retrieve a metric from its name (ID field)
 * @arg plugin pointer to the plugin
 * @arg name Name of the plugin to be queried
 * @return NULL if not found pointer to Metric otherwise */
struct MAP_Metric * MAP_Plugin_get_metric_by_name( struct MAP_Plugin * plugin, char * name );


/** This defines a MAP PLugin array gathering \ref MAP_Plugin */
struct MAP_Plugin_array
{
	/** Path where the plugins are located */
	char * plugin_prefix;
	
	/** Loaded plugins */
	struct MAP_Plugin * plugins;
	/** Number of loaded plugins */
	unsigned int plugin_count;
};

/** Load a plugin array from a given prefix
 * @arg array Pointer to the plugin array to be allocated
 * @arg plugin_prefix Path to the plugin storage directory 
 * @return Non zero on error
 */
int MAP_Plugin_array_load( struct MAP_Plugin_array * array, char * plugin_prefix );

/** Release the Plugin array
 * @arg array Pointer to the plugin array to be released
 * @return Non zero on error
 */
int MAP_Plugin_array_release( struct MAP_Plugin_array * array );


/** Get a plugin the array from its id
 * @arg array Pointer to the plugin array 
 * @arg id Id of the plugin to retrieve
 * @return Pointer to the plugin or NULL
 */
static inline struct MAP_Plugin * MAP_Plugin_array_get( struct MAP_Plugin_array * array, unsigned int id )
{
	if( array->plugin_count <= id )
	{
		return NULL;
	}
	
	return &array->plugins[ id ];
}


/** Get a metric from the array and plugins from their ids
 * @arg array Pointer to the plugin array 
 * @arg plugin_id Id of the plugin to retrieve
 * @arg metric_id Id of the metric to retrieve
 * @return Pointer to the metric or NULL
 */
static inline struct MAP_Metric * MAP_Plugin_array_get_metric( struct MAP_Plugin_array * array, unsigned int plugin_id,  unsigned int metric_id )
{
	struct MAP_Plugin * plugin = MAP_Plugin_array_get( array, plugin_id );
	
	if( !plugin )
	{
		return NULL;
	}
	
	return MAP_Plugin_get_metric( plugin, metric_id );
}


/** Retrieve all the data from the plugin array
 * @arg array Pointer to the plugin array to be collected
 * @arg getTS function returning a timestamp
 * @arg handler Function called to provide the data
 * 			- plugin : Pointer to the plugin
 * 			- metric : Pointer to the metric in the plugin
 *          - plugin_id : id of the plugin (offset in plugin array)
 *          - metric_id : id of the metric (offset in plugin local metric array)
 * 			- timestamp : the timestamp at event retrieval
 * 			- type : Type of the data being collected
 * 			- dvalue : Double value is matches type (0.0 otherwise)
 * 			- ulivalue : uint64_t value if matches type (0 otherwise)
 * 			- arg : possible void pointer to be passed around by the caller
 * @arg arg void pointer to be passed to the handler
 * @return Non zero on error
 */
int MAP_Plugin_array_trigger( struct MAP_Plugin_array * array , void (*handler)( struct MAP_Plugin * plugin,
																  struct MAP_Metric * metric,
																  int plugin_id,
																  int metric_id,
																  struct timespec * timestamp,
																  MAP_Metric_type type, 
																  double dvalue, 
																  uint64_t ulivalue,
																  void * arg),
																  void * arg );

/** Defines is a proces is the root in order to display text only in root
 *
 *  Not setting this variable is equivalent to deactivating display 
 * 
 **/
void MAP_Plugin_set_as_root();

/** This defines a MAP metric group tagging several \ref MAP_Metric */
struct MAP_Metric_group
{
	char * id;								/** Group id */
	char * name;							/** Display name */
	char * description;						/** Group Description */
	struct MAP_Metric ** group_metrics;	/** Metrics reffered to */
	unsigned int group_metric_count;		/** Number of metrics reffered to */
};

/** Init a metric group
 * @arg plugin pointer to the metric group
 * @arg desc Pointer to the XML node describing the metric group
 * @arg plugin Pointer to the plugin to backreference grouped metrics
 * @return NULL if not found pointer to Metric otherwise */
int MAP_Metric_group_init( struct MAP_Metric_group * group, xmlNode * desc, struct MAP_Plugin * plugin );


/** Release a metric group
 * @arg plugin pointer to the metric group
 * @return NULL if not found pointer to Metric otherwise */
int MAP_Metric_group_release( struct MAP_Metric_group * group );




#endif /* MALP_MAP_PLUGIN */
