/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#include "map_plugin_api.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <stdint.h>

#include "Timer.h"
#include "map_plugin.h"

/** ERROR reporting API */

/* Plugin Level */
void allinea_set_plugin_error_message(plugin_id_t plugin_id, int error_code, const char *error_message)
{
	struct MAP_Plugin * p = (struct MAP_Plugin *)plugin_id;
	
	p->error_message = strdup( error_message );
	p->return_code = error_code;
	
}

void allinea_set_plugin_error_messagef(plugin_id_t plugin_id, int error_code, const char *error_message, ...)
{
	struct MAP_Plugin * p = (struct MAP_Plugin *)plugin_id;
	char * buff = malloc( 1024 * 8 );
	
	if( !buff )
	{
		perror("malloc");
		return;
	}
	
	va_list args;
    va_start(args, error_message);
    vsnprintf(buff, 1024*8, error_message, args);
    va_end(args);
    
	p->error_message = buff;
	p->return_code = error_code;   
}


/* Metric Level */
void allinea_set_metric_error_message(metric_id_t metric_id, int error_code, const char *error_message)
{
	struct MAP_Metric * m = (struct MAP_Metric *)metric_id;
	
	m->error_message = strdup( error_message );
	m->return_code = error_code;
}

void allinea_set_metric_error_messagef(metric_id_t metric_id, int error_code, const char *error_message, ...)
{
	struct MAP_Metric * m = (struct MAP_Metric *)metric_id;
	char * buff = malloc( 1024 * 8 );
	
	if( !buff )
	{
		perror("malloc");
		return;
	}
	
	va_list args;
    va_start(args, error_message);
    vsnprintf(buff, 1024*8, error_message, args);
    va_end(args);
    
	m->error_message = buff;
	m->return_code = error_code;   
}



/** Machine discovery API */

/* Including HT */
int allinea_get_logical_core_count()
{
	return sysconf( _SC_NPROCESSORS_ONLN );
}


/* Logical Cores */
int allinea_get_physical_core_count()
{
	return sysconf( _SC_NPROCESSORS_ONLN );
}

/** Safe Syscall API */

/* Print */
void allinea_safe_printf(const char *format, ...)
{
	va_list args;
    va_start(args, format);
    vprintf( format, args);
    va_end(args);
}

void allinea_safe_fprintf(int fd, const char *format, ...)
{
	va_list args;
    va_start(args, format);
    allinea_safe_vfprintf(fd, format, args);
    va_end(args);
}

void allinea_safe_vfprintf(int fd, const char *format, va_list ap)
{
	FILE * f = fdopen( fd , "w" );
	
	if( !f )
		return;
	
	vfprintf( f, format, ap );
}

/* Open Close */
int allinea_safe_open (const char *file, int oflags, ...)
{
    mode_t mode = 0;
    if (oflags & O_CREAT)
    {
        va_list ap;
        va_start(ap, oflags);
        mode = va_arg(ap, mode_t);
        va_end(ap);
    }

    return open(file, oflags, mode);
}

int allinea_safe_close(int fd)
{
	return close( fd );
}

/* IO operations */
ssize_t allinea_safe_read(int fd, void *buf, size_t count)
{
	return read( fd, buf, count );
}

ssize_t allinea_safe_write(int fd, const void *buf, size_t count)
{
	return write( fd, buf, count );
}

ssize_t allinea_safe_read_all(int fd, void *buf, size_t count)
{
	
	struct stat stats;
	
	if( fstat(fd, &stats ) < 0 )
	{
		perror("fstat");
		return -1;
	}
	
	int to_read = (stats.st_size < count ) ? ( stats.st_size ) : ( count );
	
	int did_read = 0;

	while( did_read < to_read )
	{
		char dat;
		
		int ret = read( fd, buf + did_read, to_read);
		
		if( ret < 0 )
		{
			if( errno == EAGAIN )
				continue;
			
			perror("read");
			return -1;
		}
		
		if( ret == 0 )
		{
			break;
		}
		
		did_read += ret;
	}
	
	return did_read;
	
}

ssize_t allinea_safe_read_line(int fd, void *buf, size_t count)
{
	int did_read = 0;
	char * cbuf = ( char * ) buf;
	
	while( did_read < count )
	{
		char dat;
		
		int ret = read( fd, &dat, 1);
		
		if( ret < 0 )
		{
			if( errno == EAGAIN )
				continue;
			
			perror("read");
			return -1;
		}
		
		if( ret == 0 )
		{
			break;
		}
		
		if( dat == '\n' )
		{
			cbuf[did_read] = '\0';
			break;
		}
		else
		{
			cbuf[did_read] = dat;
		}
		
		did_read++;
	}
	
	return did_read;
}


/* Timers */


int __did_set_time_origin = 0;
static struct timespec __time_origin;

struct timespec allinea_get_current_time()
{
	struct timespec ret;
	
#if 0
	/* This is the implementation outside of MALP */
	if( __did_set_time_origin == 0 )
	{
		if( clock_gettime( CLOCK_REALTIME, &__time_origin) < 0 )
		{
			perror("clock_gettime");
			abort();
		}
		
		__did_set_time_origin = 1;
	}
	
		
	if( clock_gettime( CLOCK_REALTIME, &ret) < 0 )
	{
		perror("clock_gettime");
		abort();
	}
	
	
	ret.tv_sec -= __time_origin.tv_sec;
	ret.tv_nsec -= __time_origin.tv_nsec;

#else
	/* MALP synchronized clock */
	uint64_t ts = Timer_tsc();
	
	double time_in_sec = ts / Timer_second();
	
	long sec = floor( time_in_sec );
	long nsec = (time_in_sec - sec) * 1e9;

	ret.tv_sec = sec;
	ret.tv_nsec = nsec;

#endif


	return ret;
}
