/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

/**
 * @brief @addtogroup Data_Model_
 *
 *@{
 */

#ifndef EVENT_DESCRIPTOR_H
#define EVENT_DESCRIPTOR_H

#include <stdint.h>
#include "config.h"

#if HAVE_OMPT_H
   #include "OMPT_Events.h"
#endif /*HAVE_OMPT_H */

#if HAVE_MPI_T
   #include "MPIT_Events.h"
#endif /* HAVE_MPI_T */ 

#include "Stream_Descriptor.h"
#include "MPI_Events.h"
#include "Wrapped_Symbols.h"
#include "OTF_Event.h"
#include "MAP_Event.h"
#include "MPI_Section_evt.h"
#include "Locus.h"

/**
 * @brief enum defining event types
 */
typedef enum
{
	MALP_EVENT_NONE,     	 /**< @brief None type event */
	MALP_EVENT_META,     	 /**< @brief The event contains meta data */
	MALP_EVENT_MPI,      	 /**< @brief The event contains MPI informations */
	MALP_EVENT_WRAPPED,      /**< @brief The event contains wrapped function informations */
	MALP_EVENT_OTF,			 /**< @brief Enter + Leave events for OTF2 compatibility */
	MALP_EVENT_MAP,			 /**< @brief MAP event */
	MALP_EVENT_MPI_SECTION,	 /**< @brief MPI_Section events */
#if HAVE_OMPT_H
        MALP_EVENT_OMPT,          /**< @brief OMPT events */
#endif /* HAVE_OMPT_H */
#if HAVE_MPI_T
        MALP_EVENT_MPI_T,          /**< @brief OMPT events */
#endif /* HAVE_MPI_T */
	MALP_EVENT_TYPE_COUNT    /**< @brief The number of enums */
}MALP_Trace_event_type;

/**
 * @brief List of string constants describing event types
 */
static const char * const MALP_Trace_event_type_names[ MALP_EVENT_TYPE_COUNT ] =
{
	"None", /* MALP_EVENT_NONE */
	"Meta-Data", /* MALP_EVENT_META */
	"MPI event", /* MALP_EVENT_MPI */
	"Posix event", /* MALP_EVENT_WRAPPED */
	"OTF_Event", /* MALP_EVENT_OTF */
	"MAP Event", /* MALP_EVENT_MAP */
	"MPI_Section", /* MALP_EVENT_MPI_SECTION */
#if HAVE_OMPT_H
    	"OMPT"        /* MALP_EVENT_OMPT */
#endif /* HAVE_OMPT_H */
#if HAVE_MPI_T
        "MALP_EVENT_MPI_T",   /**< @brief MPI_T events */
#endif /* HAVE_MPI_T */
};


/**
 * @brief union defining which data can be contained into an event
 */
typedef union
{
	struct Stream_Descriptor stream_descriptor; /**< @brief See @link Stream_Descriptor @endlink */
	struct MPI_Event mpi_event;                   /**< @brief See @link MPI_Event @endlink */
	struct Wrapped_Event wrapped_event;           /**< @brief See @link Wrapped_Event @endlink */
	struct OTF_Event otf;						   /**< @brief Enter + Leave event implementation for OTF support */
	struct MAP_Event map;						   /**< @brief MAP sampled Event definition entry */
	struct MPI_Section_event mpi_section;		   /**< @brief Events associated with MPI_Sections */
#if HAVE_OMPT_H
        struct MALP_OMPT_event_s ompt;                             /**< @brief Events associated with OMPT */
#endif /* HAVE_OMPT_H */
#if HAVE_MPI_T
        struct MALP_MPI_T_event_s mpit;                             /**< @brief Events associated with MPI_T */
#endif /* HAVE_MPI_T */
} MALP_Trace_event_t;

/**
 * @brief Struct defining an event
 */
struct MALP_Trace_Event
{
	struct Event_Locus locus;     /**< @brief The event locator */
	MALP_Trace_event_type type; /**< @brief The type of the event */
	uint64_t timestamp;          /**< @brief Creation date */
	uint64_t duration;           /**< @brief Duration */
	uint64_t tid;                /**< @brief The stream unique id */
	uint64_t partition_id;       /**< @brief The sender partition id */
	int rank;                     /**< @brief The rank */
	MALP_Trace_event_t event;   /**< @brief The event data */
};


/**
 * @brief The size of a pack of events
 */
#define EVENT_PACK_SIZE 8192

/**
 * @brief Struct defining a pack of events
 */
struct Event_Pack
{
	struct Pack_Locus locus;                            /**< @brief The pack locator */
	struct MALP_Trace_Event events[EVENT_PACK_SIZE];    /**< @brief The events */
	uint32_t offset;                                    /**< @brief The number of packed events */
	uint8_t contains_metadata;                          /**< @brief Indicates whether the pack contains at least one MALP_EVENT_META event */
};


#endif /* EVENT_DESCRIPTOR_H */

/**
 * @}
 */
