#include "config.h"

#if HAVE_OMPT_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "Timer.h"
#include "Event_Desc.h"
#include "Event_Pack.h"
#include "Stream_MGR.h"
#include "MALP_atomics.h"

#include "ompt.h"
#include "OMPT_instr.h"

#define MALP_OMPT_FAIL    0
#define MALP_OMPT_SUCCESS 1

static MALP_atomics_int_t MALP_ompt_next_parallel_id = MALP_ATOMICS_INT_T_INIT(1); 
static MALP_atomics_int_t MALP_ompt_next_task_id = MALP_ATOMICS_INT_T_INIT(1); 
static MALP_atomics_int_t MALP_ompt_next_thread_id = MALP_ATOMICS_INT_T_INIT(1); 
static MALP_atomics_int_t MALP_ompt_next_lock_id = MALP_ATOMICS_INT_T_INIT(1); 

static int 
MALP_ompt_gen_parallel_id( void )
{
   return MALP_atomics_fetch_and_incr_int( &( MALP_ompt_next_parallel_id ) );
}

static int 
MALP_ompt_gen_task_id( void )
{
   return MALP_atomics_fetch_and_incr_int( &( MALP_ompt_next_task_id ) );
}

static int 
MALP_ompt_gen_thread_id( void )
{
   return MALP_atomics_fetch_and_incr_int( &( MALP_ompt_next_thread_id ) );
}

static int 
MALP_ompt_gen_lock_id( void )
{
   return MALP_atomics_fetch_and_incr_int( &( MALP_ompt_next_lock_id ) );
}

#define MALP_ompt_str(name) #name

#define MALP_ompt_register_cb( add_cb_fn, cb_name )                        		\
    do{                                                                                 \
        int retval = add_cb_fn( cb_name , (ompt_callback_t ) MALP_##cb_name );   	\
        if( ( retval == ompt_set_error ) || ( retval == ompt_set_never ) )              \
        {                                                                               \
            fprintf(stderr, "error: callback '" MALP_ompt_str(cb_name) "' \n" );        \
        }                                                                               \
    }while(0)

/****** TASK CALLBACK ******/

static MALP_ompt_rt_entries_t* MALP_ompt_rt_entries; 

static inline MALP_ompt_rt_entries_t*
MALP_get_ompt_rt_entries( void )
{
   return MALP_ompt_rt_entries;
}

static inline int
MALP_ompt_if_runtime_entries_init( ompt_function_lookup_t ompt_lookup_fn )
{
   if( !( MALP_ompt_rt_entries ) )
   {
      MALP_ompt_rt_entries = malloc( sizeof( MALP_ompt_rt_entries_t ) );
      assert( MALP_ompt_rt_entries );
   } 
  
   /* Register Runtime Entry Points */ 
   MALP_ompt_rt_entries->lookup_states_fn = ( ompt_enumerate_states_t ) ompt_lookup_fn( "ompt_enumerate_states" );
   MALP_ompt_rt_entries->lookup_mutex_impls_fn = ( ompt_enumerate_mutex_impls_t ) ompt_lookup_fn( "ompt_enumate_mutex_impls" );
   MALP_ompt_rt_entries->set_callback_fn = ( ompt_set_callback_t ) ompt_lookup_fn( "ompt_set_callback" );
   MALP_ompt_rt_entries->get_callback_fn = ( ompt_get_callback_t ) ompt_lookup_fn( "ompt_get_callback" );
   MALP_ompt_rt_entries->get_thread_data_fn = ( ompt_get_thread_data_t ) ompt_lookup_fn( "ompt_get_thread_data " ); 
   MALP_ompt_rt_entries->get_task_info_fn = ( ompt_get_task_info_t ) ompt_lookup_fn( "ompt_get_task_info" );
   MALP_ompt_rt_entries->get_parallel_info_fn = ( ompt_get_parallel_info_t ) ompt_lookup_fn( "ompt_get_parallel_info" );

   return MALP_OMPT_SUCCESS;
}

static inline MALP_ompt_thread_data_t*
MALP_ompt_get_thread_data( void )
{
   ompt_data_t* thread_data;
   MALP_ompt_rt_entries_t* entries;

   if( !( entries = MALP_get_ompt_rt_entries() ) )
      return NULL;
 
   if( !( entries->get_thread_data_fn ) )
      return NULL; 
   
   thread_data = entries->get_thread_data_fn();
   return (MALP_ompt_thread_data_t*) thread_data->ptr;
}


static inline MALP_ompt_task_data_t*
MALP_ompt_get_task_data( int ancestor_level, 
                         ompt_task_type_t *type,
                         ompt_frame_t** task_frame,
                         ompt_data_t** parallel_data,
                         int *thread_num )
{
   int ret, __thread_num, __type;
   ompt_frame_t * __task_frame;
   ompt_data_t *__parallel_data, *__task_data; 
   MALP_ompt_rt_entries_t* entries;

   assert( type && task_frame );

   if( !( entries = MALP_get_ompt_rt_entries() ) )
      return NULL;

   if( !( entries->get_task_info_fn ) )
      return NULL;
   
   ret = entries->get_task_info_fn( ancestor_level, &__type, &__task_data, &__task_frame, &__parallel_data, &__thread_num );

   if( !( ret ) ) return NULL;

   /* Update Other args */
   if( type ) *type = __type;
   if( task_frame ) *task_frame = __task_frame;
   if( thread_num ) *thread_num = __thread_num;
   if( parallel_data ) *parallel_data = ( ompt_data_t*) __parallel_data->ptr;  

   return ( MALP_ompt_task_data_t*) __task_data->ptr;
}

static inline MALP_ompt_task_data_t*
MALP_ompt_get_cur_task_data( ompt_task_type_t *type, 
                             ompt_frame_t** task_frame,  
                             ompt_data_t** parallel_data, 
                             int *thread_num )
{
   return MALP_ompt_get_task_data( 0, type, task_frame, parallel_data, thread_num ); 
}

static inline MALP_ompt_parallel_data_t*
MALP_ompt_get_parallel_info( int ancestor_level, int *team_size )
{
   int ret, __team_size;
   ompt_data_t* parallel_data;
   MALP_ompt_rt_entries_t* entries;

   if( !( entries = MALP_get_ompt_rt_entries() ) )
      return NULL;

   if( !( entries->get_parallel_info_fn ) )
      return NULL;

   ret = entries->get_parallel_info_fn( ancestor_level, &parallel_data, &__team_size );
    
   if( !( ret ) ) return NULL;
 
   if( team_size ) *team_size = __team_size;
   
   return (MALP_ompt_parallel_data_t*) parallel_data->ptr; 
}

static inline MALP_ompt_parallel_data_t*
MALP_ompt_get_cur_parallel_info( int *team_size )
{
   return MALP_ompt_get_parallel_info( 0, team_size );
}

static inline int
MALP_ompt_set_event_ctx_data( MALP_ompt_context_call_t* ctx )
{
    ompt_data_t* __ompt_data;
    int* parallel_ids, *task_ids;
    int par_ancestor_num, tsk_ancestor_num;
    MALP_ompt_task_data_t* __task_data; 
    MALP_ompt_thread_data_t* __thread_data; 
    MALP_ompt_parallel_data_t* __parallel_data; 
    assert( ctx );

    __thread_data = MALP_ompt_get_thread_data();

    par_ancestor_num = 0; parallel_ids = NULL;
    while( __parallel_data = MALP_ompt_get_parallel_info( par_ancestor_num, NULL ) )
    {
       int* tmp = ( int* ) realloc( parallel_ids, sizeof( int ) * ( par_ancestor_num + 1 ) );
       assert( tmp );
       parallel_ids = tmp;
       parallel_ids[par_ancestor_num] = __parallel_data->global_id;
       par_ancestor_num++;
    }

    tsk_ancestor_num = 0; task_ids = NULL;
    while( __task_data = MALP_ompt_get_task_data( tsk_ancestor_num, NULL, NULL, NULL, NULL ) )
    {
       int* tmp = ( int* ) realloc( task_ids, sizeof( int ) * ( tsk_ancestor_num + 1 ) );
       assert( tmp );
       task_ids = tmp;
       task_ids[tsk_ancestor_num] = __task_data->global_id;
       tsk_ancestor_num++;
    }

    assert( par_ancestor_num <= tsk_ancestor_num );

    /* 0 is undef variable so task / parallel start with 1 */
    memset( ctx, 0, sizeof( MALP_ompt_context_call_t ) );

    /* GENERIC CTX DATA*/
    assert( par_ancestor_num <= MALP_OMPT_MAX_PARALLEL_NESTED_LEVEL );
    memcpy( ctx->parallel_ids, parallel_ids, sizeof( int ) * par_ancestor_num );

    assert( tsk_ancestor_num <= MALP_OMPT_MAX_TASK_NESTED_LEVEL );
    memcpy( ctx->task_ids, task_ids, sizeof( int ) * tsk_ancestor_num );

    ctx->nested_level = par_ancestor_num;
    ctx->thread_id = __thread_data->global_id;

    return MALP_OMPT_SUCCESS;
}


/****** PARALLEL REGION CALLBACK ******/

static inline void
MALP_ompt_callback_task_create( ompt_data_t* parent_task_data,
                                const ompt_frame_t* parent_frame,
                                ompt_data_t* new_task_data,
				ompt_task_type_t type,
                                int has_dependences,
                                const void * codeptr_ra )
{
   struct MALP_Trace_Event event; 

   const uint64_t __begin = Timer_tsc ();   

   fprintf(stderr, "%s:%d -- nb deps: %d\n", __func__, __LINE__, has_dependences );

   event.event.ompt.scope = MALP_OMPT_SCOPE_BEGIN;
   event.event.ompt.type = MALP_OMPT_TASK_EVENT_CREATE;
   MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 ); 

   /* Submit event if OMPT instrumented */
   Event_Packer_push( &event );
}

static inline void
MALP_ompt_callback_task_dependences( ompt_data_t* task_data,
                                     const ompt_task_dependence_t* deps,
                                     int ndeps )
{

}

static inline void
MALP_ompt_callback_task_schedule( ompt_data_t* prior_task_data,
                                  ompt_task_status_t prior_task_status,
                                  ompt_data_t* next_task_data )
{

}

static inline void
MALP_ompt_callback_implicit_task( ompt_scope_endpoint_t endpoint,
                                   ompt_data_t* parallel_data,
                                   ompt_data_t* tast_data,
                                   unsigned int thread_num )
{
   fprintf(stderr, "%s:%d -- nb deps: %d\n", __func__, __LINE__, 0 );
}

static inline int
MALP_ompt_enable_task_report( void )
{
   ompt_set_callback_t register_cb;

   /* Get register callback from lookup */
   register_cb = MALP_get_ompt_rt_entries()->set_callback_fn;

   if( !( register_cb ) ) return MALP_OMPT_FAIL;

   MALP_ompt_register_cb( register_cb, ompt_callback_task_create );
   MALP_ompt_register_cb( register_cb, ompt_callback_task_dependences );
   MALP_ompt_register_cb( register_cb, ompt_callback_task_schedule );
   MALP_ompt_register_cb( register_cb, ompt_callback_implicit_task );
   
   return MALP_OMPT_SUCCESS;
}


static inline void 
MALP_ompt_callback_parallel_begin( ompt_data_t* parent_task_data,
                                   const ompt_frame_t* parent_frame,
                                   ompt_data_t* parallel_data,
                                   unsigned int requested_team_size,
                                   unsigned int actual_team_size,
                                   ompt_invoker_t invoker,
                                   const void * codeptr_ra )
{
   struct MALP_Trace_Event event;
   MALP_ompt_parallel_data_t* __parallel_data;
   
   const uint64_t __begin = Timer_tsc ();

   assert( parallel_data ); //&& *parallel_data != ompt_data_none );

   __parallel_data = malloc( sizeof( MALP_ompt_parallel_data_t ) );
   assert( __parallel_data );

/*
   event.event.ompt.type = MALP_OMPT_PARALLEL_EVENT;
   event.event.ompt.scope = MALP_OMPT_SCOPE_BEGIN;
   MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );
*/
   __parallel_data->start_time = __begin; 
   __parallel_data->global_id = MALP_ompt_gen_parallel_id(); 
   parallel_data->ptr = ( void* ) __parallel_data;	

   /* Submit event if OMPT instrumented */
//   Event_Packer_push( &event );
}

static inline void 
MALP_ompt_callback_parallel_end( ompt_data_t* parallel_data,
                                 ompt_data_t* task_data,
                                 ompt_invoker_t invoker,
                                 const void * codeptr_ra )
{
   struct MALP_Trace_Event event;
   MALP_ompt_parallel_data_t* __parallel_data;

   const uint64_t __end = Timer_tsc ();

   assert( parallel_data ); //&& *parallel_data != ompt_data_none );

   __parallel_data = ( MALP_ompt_parallel_data_t* ) parallel_data->ptr;
   const uint64_t __begin = __parallel_data->start_time;  

   event.event.ompt.type = MALP_OMPT_PARALLEL_EVENT;
   event.event.ompt.scope = MALP_OMPT_SCOPE_END;
   MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __end, __end - __begin );

   /* Submit event if OMPT instrumented */
   Event_Packer_push( &event );

   free( __parallel_data );
}

static inline int
MALP_ompt_enable_parallel_report( void )
{
   ompt_set_callback_t register_cb;

   /* Get register callback from lookup */
   register_cb = MALP_get_ompt_rt_entries()->set_callback_fn;

   if( !( register_cb ) ) return MALP_OMPT_FAIL;

   MALP_ompt_register_cb( register_cb, ompt_callback_parallel_begin );
   MALP_ompt_register_cb( register_cb, ompt_callback_parallel_end );

   return MALP_OMPT_SUCCESS;
}

/****** THREAD CALLBACK ******/

static inline void
MALP_ompt_callback_thread_begin( ompt_thread_type_t thread_type, 
				 ompt_data_t* thread_data )
{
   struct MALP_Trace_Event event;
   MALP_ompt_thread_data_t* __thread_data;
    
   const int __begin = Timer_tsc ();
   
   assert( thread_data && !( thread_data->value ) );

   __thread_data = malloc( sizeof( MALP_ompt_thread_data_t ) );
   assert( __thread_data );

   event.event.ompt.type = MALP_OMPT_THREAD_EVENT;
   event.event.ompt.scope = MALP_OMPT_SCOPE_BEGIN;
   MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 ); 

   __thread_data->start_time = __begin; 
   __thread_data->global_id = MALP_ompt_gen_thread_id();
   thread_data->ptr = ( void* ) __thread_data; 

   /* Submit event if OMPT instrumented */
   Event_Packer_push( &event );
}

static inline void
MALP_ompt_callback_thread_end( ompt_data_t* thread_data )
{
   struct MALP_Trace_Event event;
   MALP_ompt_thread_data_t* __thread_data;

   //TODO: Check Event after MPI_FInalize 	
   return;

   const int __end = Timer_tsc ();

   assert( thread_data && thread_data->ptr );

   __thread_data = ( MALP_ompt_thread_data_t* ) thread_data->ptr;
   const int __begin = __thread_data->start_time;   

   event.event.ompt.type = MALP_OMPT_THREAD_EVENT;
   event.event.ompt.scope = MALP_OMPT_SCOPE_END;
   MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __end, __end - __begin ); 

   /* Submit event if OMPT instrumented */
   Event_Packer_push( &event );

   free( __thread_data );
}

static inline int 
MALP_ompt_enable_thread_report( void )
{
   ompt_set_callback_t register_cb;

   /* Get register callback from lookup */
   register_cb = MALP_get_ompt_rt_entries()->set_callback_fn;

   if( !( register_cb ) ) return MALP_OMPT_FAIL;

   MALP_ompt_register_cb( register_cb, ompt_callback_thread_begin );
   MALP_ompt_register_cb( register_cb, ompt_callback_thread_end );
   return MALP_OMPT_SUCCESS;
}

/****** LOCK CALLBACK ******/

/* Register a new lock from OpenMP runtime */
static inline void
MALP_ompt_callback_lock_init( ompt_mutex_kind_t kind,
                              unsigned int hint,
                              unsigned int impl,
                              ompt_wait_id_t wait_id,
                              const void* codeptr_ra )
{
    MALP_OMPT_event_t* ompt;
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;

    const int __begin = Timer_tsc ();
    
    ompt =  &( event.event.ompt );
    ompt->type = MALP_OMPT_LOCK_EVENT_INIT; 
    //(void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );
    
    /* SPECIFIC EVT DATA */ 
    lock = &( ompt->data.lock ); 
    lock->type = MALP_ompt_translate_lock_type( kind );
    lock->hint = hint;
    lock->impl = impl;
    lock->wait_id = wait_id;

    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );   

    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );
}

static inline void
MALP_ompt_callback_lock_destroy( ompt_mutex_kind_t kind,
                                 ompt_wait_id_t wait_id,
                                 const void* codeptr_ra )
{
    MALP_OMPT_event_t* ompt;
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;

    const int __begin = Timer_tsc ();
    
    ompt =  &( event.event.ompt );
    ompt->type = MALP_OMPT_LOCK_EVENT_FINI; 
    (void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );
    
    /* SPECIFIC EVT DATA */ 
    lock = &( ompt->data.lock ); 
    memset( lock, 0, sizeof( MALP_ompt_lock_event_t ) ); 
    lock->type = MALP_ompt_translate_lock_type( kind );
    lock->wait_id = wait_id;

    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );   

    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );
}

static inline void 
MALP_ompt_callback_nest_lock( ompt_scope_endpoint_t endpoint,
                              ompt_wait_id_t  wait_id,
                              const void* codeptr_ra )
{
    MALP_OMPT_event_t* ompt;
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;

    const int __begin = Timer_tsc ();
    
    ompt =  &( event.event.ompt );
    ompt->type = MALP_OMPT_LOCK_EVENT_NEST; 
    (void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );
    
    /* SPECIFIC EVT DATA */ 
    lock = &( ompt->data.lock ); 
    memset( lock, 0, sizeof( MALP_ompt_lock_event_t ) ); 
    lock->wait_id = wait_id;

    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );   

    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );

}

static inline void 
MALP_ompt_callback_mutex_acquire( ompt_mutex_kind_t kind,
                                  unsigned int hint,
                                  unsigned int impl,
                                  ompt_wait_id_t  wait_id,
                                  const void* codeptr_ra )
{
    MALP_OMPT_event_t* ompt;
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;

    const int __begin = Timer_tsc ();
    
    ompt =  &( event.event.ompt );
    ompt->type = MALP_OMPT_LOCK_EVENT_WAIT; 
    (void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );
    
    /* SPECIFIC EVT DATA */ 
    lock = &( ompt->data.lock ); 
    memset( lock, 0, sizeof( MALP_ompt_lock_event_t ) ); 
    lock->type = MALP_ompt_translate_lock_type( kind );
    lock->hint = hint;
    lock->impl = impl;
    lock->wait_id = wait_id;

    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );   

    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );
}

static inline void 
MALP_ompt_callback_mutex_acquired( ompt_mutex_kind_t kind,
                                   ompt_wait_id_t  wait_id,
                                   const void* codeptr_ra )
{
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;
    MALP_OMPT_event_t* ompt;

    const int __begin = Timer_tsc ();
    
    ompt =  &( event.event.ompt );
    ompt->type = MALP_OMPT_LOCK_EVENT_INIT; 
    (void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );
    
    /* SPECIFIC EVT DATA */ 
    lock = &( ompt->data.lock ); 
    lock->type = MALP_ompt_translate_lock_type( kind );
    lock->wait_id = wait_id;

    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );   

    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );
}

static inline void 
MALP_ompt_callback_mutex_released( ompt_mutex_kind_t kind,
                                   ompt_wait_id_t  wait_id,
                                   const void* codeptr_ra )
{
    MALP_OMPT_event_t* ompt;
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;

    const int __begin = Timer_tsc ();
    
    ompt =  &( event.event.ompt );
    ompt->type = MALP_OMPT_LOCK_EVENT_INIT; 
    (void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );
    
    /* SPECIFIC EVT DATA */ 
    lock = &( ompt->data.lock ); 
    lock->type = MALP_ompt_translate_lock_type( kind );
    lock->wait_id = wait_id;

    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );   

    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );
}

static inline int
MALP_ompt_get_mutex_impls_list( ompt_function_lookup_t ompt_lookup_fn )
{
   const char *next_impl_name;  
   int current_impl, next_impl;
   ompt_enumerate_mutex_impls_t ompt_lock_lookup_fn;

   ompt_lock_lookup_fn = (ompt_enumerate_mutex_impls_t ) ompt_lookup_fn("ompt_enumerate_mutex_impls");
   assert( ompt_lock_lookup_fn );
#if 0   
   //Register mutex name type
   current_impl = ompt_mutex_impl_unknown;
   while( ompt_lock_lookup_fn( current_impl, &next_impl, &next_impl_name ) )
   {
      current_impl = next_impl; 
   } 
#endif 
   return MALP_OMPT_SUCCESS;
}
 
static inline int
MALP_ompt_enable_lock_report( void )
{
   ompt_set_callback_t register_cb;
   
   /* Get register callback from lookup */
   register_cb = MALP_get_ompt_rt_entries()->set_callback_fn;
   
   if( !( register_cb ) ) return MALP_OMPT_FAIL;
   
   MALP_ompt_register_cb( register_cb, ompt_callback_lock_init );
   MALP_ompt_register_cb( register_cb, ompt_callback_lock_destroy );
   MALP_ompt_register_cb( register_cb, ompt_callback_nest_lock );
   MALP_ompt_register_cb( register_cb, ompt_callback_mutex_acquire );
   MALP_ompt_register_cb( register_cb, ompt_callback_mutex_acquired );
   MALP_ompt_register_cb( register_cb, ompt_callback_mutex_released );

   return MALP_OMPT_SUCCESS;
}

static void
MALP_internal_ompt_work_event_loop_begin( ompt_data_t* parallel_data,
		                          ompt_data_t* task_data,
                                          int64_t count,
                                          const void* codeptr_ra )
{
    MALP_OMPT_event_t* ompt;
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;
    MALP_ompt_task_data_t* __task_data;
    MALP_ompt_parallel_data_t* __parallel_data;

    const int __begin = Timer_tsc ();

    ompt =  &( event.event.ompt );
    ompt->scope = MALP_OMPT_SCOPE_BEGIN;
    ompt->type = MALP_OMPT_WORK_EVENT_LOOP; 
    
    fprintf(stderr, "%s:%d -- Count value: %d\n", __func__, __LINE__, count );
    //(void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );

    MALP_ompt_push_work_timer_data( task_data, __begin );
    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, 0 );   

    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );
}

static void
MALP_internal_ompt_work_event_loop_end( ompt_data_t* parallel_data,
		                        ompt_data_t* task_data,
                                        uint64_t count,
                                        const void* codeptr_ra )
{ 
    MALP_OMPT_event_t* ompt;
    MALP_ompt_lock_event_t* lock;
    struct MALP_Trace_Event event;

    return;

    const int __end = Timer_tsc ();
    
    ompt =  &( event.event.ompt );
    ompt->scope = MALP_OMPT_SCOPE_END;
    ompt->type = MALP_OMPT_WORK_EVENT_LOOP; 
    //(void ) MALP_ompt_set_event_ctx_data( &( ompt->ctx ) );

    const int __begin = MALP_ompt_pop_work_timer_data( task_data );
    MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __begin, __end - __begin );   


    /* Submit event if OMPT instrumented */
    Event_Packer_push( &event );
}

static void 
MALP_internal_ompt_work_event_loop( ompt_scope_endpoint_t endpoint,
                                    ompt_data_t* parallel_data,
                                    ompt_data_t* task_data,
                                    uint64_t count,
                                    const void* codeptr_ra )
{
   switch( endpoint )
   {
      case ompt_scope_begin:
          MALP_internal_ompt_work_event_loop_begin( parallel_data, task_data, count, codeptr_ra );
          break;
      case ompt_scope_end:
          MALP_internal_ompt_work_event_loop_end( parallel_data, task_data, count, codeptr_ra );
          break;
      default:
          fprintf(stderr, "Skip Work Event Loop : unknown endpoint type : %d\n", endpoint );

   }
}

void MALP_ompt_callback_work( ompt_work_type_t wstype,
                              ompt_scope_endpoint_t endpoint,
                              ompt_data_t* parallel_data,
			      ompt_data_t* task_data,
			      uint64_t count,
                              const void* codeptr_ra )
{
    switch( wstype )
    {
       case ompt_work_loop:
          MALP_internal_ompt_work_event_loop( endpoint, parallel_data, task_data, count, codeptr_ra );
          break;
       case ompt_work_sections:
       case ompt_work_single_executor:
       case ompt_work_single_other:
       case ompt_work_workshare:
       case ompt_work_distribute:
       case ompt_work_taskloop:
       default:
          fprintf(stderr, "Skip Work Event: unknown wstype type : %d\n", wstype );
    }
}

static inline int
MALP_ompt_callback_master( ompt_scope_endpoint_t endpoint, 
			   ompt_data_t* parallel_data, 
                           ompt_data_t* task_data,
                    	   const void* codeptr_ra ) 
{
   MALP_OMPT_event_t* ompt;
   struct MALP_Trace_Event event;
   
   const uint64_t __end = Timer_tsc ();

   if( endpoint != ompt_scope_end ) return;
 
   ompt =  &( event.event.ompt );
   ompt->scope = MALP_OMPT_SCOPE_END;
   MALP_Trace_Event_fill( &event, MALP_EVENT_OMPT, __end, 0 );
   
   /* Submit event if MPI instrumented */
   Event_Packer_push(&event);
	
   return MPI_SUCCESS;
}


static inline int
MALP_ompt_enable_work_sharing_report( void )
{
   ompt_set_callback_t register_cb;

   //Get register callback from lookup
   register_cb = MALP_get_ompt_rt_entries()->set_callback_fn;

   MALP_ompt_register_cb( register_cb, ompt_callback_work );  
   MALP_ompt_register_cb( register_cb, ompt_callback_master );  

   return MALP_OMPT_SUCCESS;
}

/****** OMPT TOOL INIT/FINI CALLBACKS ******/

static inline int 
MALP_ompt_initialize( ompt_function_lookup_t ompt_lookup_fn, ompt_fns_t* fns )
{
   ompt_set_callback_t register_cb;

   assert( fns );
   assert( ompt_lookup_fn );

   MALP_ompt_if_runtime_entries_init( ompt_lookup_fn );

   /* Get register callback from lookup */
   register_cb = MALP_get_ompt_rt_entries()->set_callback_fn;
   assert( register_cb );

   (void) MALP_ompt_enable_thread_report();
   (void) MALP_ompt_enable_parallel_report();
   (void) MALP_ompt_enable_work_sharing_report();
   
   /* Get enumerate mutex type */
   ( void ) MALP_ompt_get_mutex_impls_list( ompt_lookup_fn );
   MALP_ompt_enable_lock_report();

   /* OMPT runtime initialize must return 1 */
   return 1;
}

static inline void
MALP_ompt_finalize( ompt_fns_t* fns )
{
   assert( fns );
   free( fns );
}


/** Weak pointer call by OpenMP runtime */
ompt_fns_t* ompt_start_tool( unsigned int omp_version, const char *runtime_version )
{
    ompt_fns_t* new_ompt_fns = ( ompt_fns_t* ) malloc( sizeof( ompt_fns_t ) );
    assert( new_ompt_fns );
    new_ompt_fns->finalize = MALP_ompt_finalize;
    new_ompt_fns->initialize = MALP_ompt_initialize;
    return new_ompt_fns;
}

#endif /* HAVE_OMPT_H */
