/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

#ifndef STREAM_MGT_H
#define STREAM_MGT_H

/**
 * @addtogroup Instrum_
 *
 * @{
 */

#include <stdint.h>

#include "VMPI.h"
#include "Posix_Wrapp.h"

#include "Event_Desc.h"
#include "Spinlock.h"
#include "Chained_List.h"
#include "Buffered_FIFO.h"

/**
 * @brief Struct defining a stream manager
 */
struct Stream_Manager
{
	struct Chained_List streams;    /**< @brief The list of managed streams*/
	uint64_t current_id;            /**< @brief a globally unique id for the streams */
	uint64_t partition_id;          /**< @brief The id of the partition */
	int rank;                       /**< @brief the rank (within the partition */
	MALP_Spinlock lock;             /**< @brief the lock for concurrent access */
};

/**
 * @brief The stream manager for current MPI process
 */
extern struct Stream_Manager __stream_Stream_Manager;

/**
 * @brief A getter for the stream manager
 * @return the current manager
 */
static inline struct Stream_Manager * Stream_Manager()
{
	return &__stream_Stream_Manager;
}

/**
 * @brief Initialises the manager for current MPI process
 */
void Stream_Manager_init();

/**
 * @brief Releases the manager for current MPI process (including all streams )
 */
void Stream_Manager_release();

/**
 * @brief Adds a stream into the current stream manager
 * @param parent Descriptor of wanted parent stream
 * @param type the stream type
 * @return descriptor of the newly added stream
 *
 * @todo Within MALP, parent is alwayse NULL and type is alwayse MALP_PROCESS
 */
struct Stream_Infos * Stream_MGR_new_stream( struct Stream_Descriptor *parent, Stream_type type );

/**
 * @brief Signals end of stream
 */
void Stream_MGR_end_stream();

/**
 * @brief Defines the maximum pending events
 */
#define MAX_PENDING_EVENT 4096

/**
 * @brief Structure describing a stream
 *
 * is_in_mpi_call and mpi_wrapped_event_buffer are used to avoid looping in MPI events when calling MPI from MPI in order to flush events to MPI streams. 
 * Instead events are stored in a fifo and flushed when leaving the event. 
 */
struct Stream_Infos
{
	struct Stream_Descriptor infos;                 /**< @brief The stream descriptor */
	volatile uint8_t is_in_mpi_call;                /**< @brief 1 if we are in a MPI call */
	struct Buffered_FIFO mpi_wrapped_event_buffer;  /**< @brief FIFO of events for MPI calls */
};

/**
 * @brief Initializes a Stream_Infos
 * @param info the Stream_Infos to initialize
 */
void Stream_Infos_init( struct Stream_Infos *info );

/**
 * @brief Releases a Stream_Infos
 * @param info the Stream_Infos to release
 */
void Stream_Infos_release( struct Stream_Infos *info );

/**
 * @brief This is a global Stream_Infos (not shared between threads)
 */
extern __thread struct Stream_Infos * __trace_module;

/**
 * @brief Getter on the current Stream_Infos
 * @return the current Stream_Infos
 */
static inline struct Stream_Infos * Stream_infos_get()
{
	return __trace_module;
}

/**
 * @brief Setter on the current Stream_Infos
 * @param sd the new Stream_Infos
 */
static inline void Stream_infos_set(struct Stream_Infos *sd)
{
	__trace_module = sd;
}

/**
 * @brief Getter on the current Stream_Descriptor
 * @return the current Stream_Descriptor
 */
static inline struct Stream_Descriptor * Stream_desc_get()
{
	return &__trace_module->infos;
}

/**
 * @brief Indicates to the current Stream_Infos (if any) that we are enterning a MPI call
 */
static inline void Stream_Infos_enter_mpi()
{
	if( !__trace_module )
		return;

	__trace_module->is_in_mpi_call = 1;
}

/**
 * @brief Indicates to the current Stream_Infos (if any) that we are leaving a MPI call
 */
static inline void Stream_Infos_leave_mpi()
{
	if( !__trace_module)
		return;

	__trace_module->is_in_mpi_call = 0;
}

/**
 * @brief Checks if we are currently inside a MPI call
 * @return 1 if we are inside a MPI call, 0 otherwise
 */
static inline int Stream_Infos_is_in_mpi()
{
	if( !__trace_module )
		return 0;

	return __trace_module->is_in_mpi_call;
}

static inline void MALP_Trace_Event_fill( struct MALP_Trace_Event *event,
						MALP_Trace_event_type type,
						uint64_t timestamp,
						uint64_t duration )
{
	struct Stream_Descriptor *stream_infos = Stream_desc_get();

	event->type = type;
	event->timestamp = timestamp;
	event->duration = duration;

	event->partition_id = VMPI_Get_partition_id();
	
	if( !stream_infos )
	{
		event->tid = 0; /* Undefined IDS are set to 0 */
		return;
	}

	event->tid = stream_infos->tid;
	
	event->rank = stream_infos->mpi_rank;
}


/**
 * @}
 */

#endif /* STREAM_MGT_H */
