/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#ifndef DATA_ENTRY_H
#define DATA_ENTRY_H


#ifdef __cplusplus
  extern "C"
  {
#endif

#include <stddef.h>
#include "CRC64.h"
#include "Chained_List.h"
#include "Spinlock.h"

/**
 * @ingroup Data_Entry_
 * @brief Struct defining a piece of data that can exist on the blackboard
 */
struct Data_entry
{
	uint64_t key;           /**< @brief the key, identifying the type of data */

	uint64_t ref_count;     /**< @brief the number of users of this piece of data */
	MALP_Spinlock lock;     /**< @brief a lock for concurrent access */
	short been_resubmitted; /**< @brief 1 if it has been re-submitted */

	size_t size;            /**< @brief the size of the data available in payload */
	char payload[0];        /**< @brief the actual data */
};

/**
 * @ingroup Data_Entry_
 * @brief Creates a new Data_entry
 * @param key the data key
 * @param payload the actual data
 * @param size the size of data
 * @return the newly created Data_entry
 */
struct Data_entry * Data_entry_init( uint64_t key, void *payload, size_t size );

/**
 * @ingroup Data_Entry_
 * @brief Clones a Data_entry
 * @param de the Data_entry to clone
 * @return the cloned Data_entry
 */
struct Data_entry * Data_entry_clone( struct Data_entry *de );

/**
 * @ingroup Data_Entry_
 * @brief Clears a Data_entry
 * @param de the Data_entry to clear
 */
void Data_entry_release( struct Data_entry *de );

/**
 * @ingroup Data_Entry_
 * @brief Resets the ref_count of a Data_entry
 * @param de the Data_entry
 */
static inline void Data_entry_reset_ref_count( struct Data_entry *de )
{
	MALP_Spinlock_lock( &de->lock );
	{
		de->ref_count = 0;
	}
	MALP_Spinlock_unlock( &de->lock );
}

/**
 * @ingroup Data_Entry_
 * @brief Getter on the ref_count of a Data_entry
 * @param de the Data_entry
 * @return the ref_count of the Data_entry
 */
static inline uint64_t Data_entry_get_ref_count( struct Data_entry *de )
{
	uint64_t ret;
	MALP_Spinlock_lock( &de->lock );
	{
		ret = de->ref_count;
	}
	MALP_Spinlock_unlock( &de->lock );

	return ret;
}

/**
 * @ingroup Data_Entry_
 * @brief Increments the ref_count of a Data_entry by a given value
 * @param de the Data_entry
 * @param val the value to add to ref_count
 */
static inline void Data_entry_inc_ref_count( struct Data_entry *de, uint64_t val)
{
	MALP_Spinlock_lock( &de->lock );
	{
		de->ref_count+= val;
	}
	MALP_Spinlock_unlock( &de->lock );
}

/**
 * @ingroup Data_Entry_
 * @brief Decrements the ref_count of a Data_entry
 * @param de the Data_entry
 * @return the new ref_count
 */
static inline uint64_t Data_entry_dec_ref_count( struct Data_entry *de)
{
	uint64_t ret = 0;
	MALP_Spinlock_lock( &de->lock );
	{
		ret = --de->ref_count;
	}
	MALP_Spinlock_unlock( &de->lock );
	return ret;
}

/**
 * @ingroup Data_Entry_
 * @brief Getter on the Data_entry key
 * @param de the Data_entry
 * @return the key of the Data_entry
 */
static inline uint64_t Data_entry_get_key(struct Data_entry *de)
{
	return de->key;
}

/**
 * @ingroup Data_Entry_
 * @brief Getter on the Data_entry payload size
 * @param de the Data_entry
 * @return the size of actual data
 */
static inline size_t Data_entry_get_payload_size( struct Data_entry *de )
{
	return de->size;
}

/**
 * @ingroup Data_Entry_
 * @brief Getter on the actual data of a Data_entry
 * @param de the Data_entry
 * @return the actual data
 */
static inline void *Data_entry_get_payload( struct Data_entry *de )
{
	return (void *)de->payload;
}


#ifdef __cplusplus
  }
#endif

#endif
