/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#ifndef KEY_ENTRY_H
#define KEY_ENTRY_H

#ifdef __cplusplus
extern "C" {
#endif

#include "Spinlock.h"
#include "Chained_List.h"

struct Knowledge_system;

/**
 * @ingroup Key_Entry_
 * @brief The Key_entry struct
 */
struct Key_entry
{
	struct Chained_List target_ks;  /**< @brief the list of Knowledge_system identified by this key entry */
	MALP_Spinlock lock;             /**< @brief a lock for concurrent access */
	uint64_t key;                   /**< @brief the key identifier */
	uint8_t freed;                  /**< @brief set to one when the entry is released, 0 otherwise */
};


/**
 * @ingroup Key_Entry_
 * @brief Init a key entry with given key 
 * @param ke the key entry to initialize
 * @param key the key
 */
void Key_entry_init( struct Key_entry *ke, uint64_t key );

/**
 * @ingroup Key_Entry_
 * @brief Clears a key entry 
 * @param ke the key entry to release
 */
void Key_entry_release( struct Key_entry *ke );

/**
 * @ingroup Key_Entry_
 * @brief associates a Knowledge_system to a key entry
 * @param ke the key entry 
 * @param ks the Knowledge_system to associate to the key
 */
void Key_entry_push_ks( struct Key_entry *ke, struct Knowledge_system *ks );

/**
 * @ingroup Key_Entry_
 * @brief Removes an association between a Knowledge_system and its key
 * @param ke the key entry where to remove association
 * @param ks the Knowledge_system where to remove association
 */
void Key_entry_pop_ks( struct Key_entry *ke, struct Knowledge_system *ks );

/**
 * @ingroup Key_Entry_
 * @brief Search in the key entry if a Knowledge_system is associated with it.
 * @param ke the key entry where to find the Knowledge_system
 * @param ks the ks to search
 * @return NULL if the Knowledge_system is not associated with the given ke
 */
struct Knowledge_system **  Key_entry_get_ks( struct Key_entry *ke, struct Knowledge_system *ks );

/**
 * @ingroup Key_Entry_
 * @brief Clear all references to Ks in this entry 
 * @param ke the key entry to clear
 */
void Key_entry_clear_ks( struct Key_entry *ke );

/**
 * @ingroup Key_Entry_
 * @brief Get the number of Ks in this entry 
 * @param ke the key entry where to count
 * @return the number of Ks in this key entry
 */
int Key_entry_get_count( struct Key_entry *ke );

/**
 * @ingroup Key_Entry_
 * @brief Check if a KS type is present in a key entry
 * @param ke the key entry
 * @param type the type to search
 * @return 1 if a Ks of type type is found in ke
 */
int Key_entry_has_KS_type( struct Key_entry *ke, uint64_t type );

/**
 * @ingroup Key_Entry_
 * @brief Call a handler on every Ks of a key entry
 * @param ke the key entry where to finc the set of Knowledge_system
 * @param handler the function to run on all Knowledge_system
 * @param arg an argument to giver to handler
 */
void Key_entry_walk( struct Key_entry *ke, int (*handler)( struct Knowledge_system *, void *), void *arg );

#ifdef __cplusplus
}
#endif

#endif


