#include "Red_Tree.h"

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "Socket_Utils.h"


struct RT_Ctx
{
	int socket;
	struct Red_Tree *rt;
};

void * Red_Tree_read_loop( void *prt )
{
	struct RT_Ctx *ctx = (struct RT_Ctx *)prt;
	
	struct Red_Tree *rt = ctx->rt;
	int socket = ctx->socket;
	
	free( ctx );
	
	struct Data_pack pack;
	
	int ret = 0;
	
	do
	{
		ret = Socket_safe_read( socket, &pack, sizeof( struct Data_pack ) );
		
		if( ret < 0 )
		{
			break;
		}
		
		if( ret )
		{
			if( rt->pa_socket < 0 )
			{
				//Data_pack_print( &pack );
				char * data_ptr = Data_Unpacker_push( &rt->unpacker, &pack );
				
				if( data_ptr && rt->data_handler )
				{
					(rt->data_handler)(rt, data_ptr, pack.size );
				}
				
				free( data_ptr );
			}
			else
			{
				if( Red_Tree_push_pack( rt, &pack ) )
				{
					printf("Error sending pack\n");
					break;
				}
			}
		}
		
	}while( ret != 0 );
	
	
	printf("Read loop disconnect \n");
	
	pthread_exit(NULL);
}



void * Red_Tree_accept_thread( void *prt )
{
	struct Red_Tree *rt = (struct Red_Tree *)prt;
	
	/* This thread accepts two connections then return */
	int i = 0;
	
	while( i < 2 )
	{
		
		int client_fd = Socket_wait_for_a_client(rt->listening_socket);
		pthread_t * th = NULL;
		
		if( 0 < client_fd )
		{
			struct RT_Ctx *ctx = malloc( sizeof( struct RT_Ctx ) );
			
			if( !ctx )
			{
				perror("malloc");
				abort();
			}
			
			ctx->socket = client_fd;
			ctx->rt = rt;
			
			
			if( i== 0 )
			{
				printf("%s accepted a client LC\n", rt->hostname );
				/* Left C */
				rt->lc_fd = client_fd;
				th = &rt->lc_progress_thread;
			}
			if( i == 1 )
			{
				printf("%s accepted a client RC\n", rt->hostname );
				/* Right C */
				rt->rc_fd = client_fd;
				th = &rt->rc_progress_thread;
			}
			
			/* We can now start the reading thread on client fd */
			if( pthread_create(th, NULL, Red_Tree_read_loop, (void*)ctx ) != 0 )
			{
				perror("pthread_create");
				abort();
			}
		}
		
		i++;
	}
	
	pthread_exit(NULL);
	
	
}


int Red_Tree_connect( struct Red_Tree *rt, struct Connection_ctx * target )
{
	/* Start a client socket 
	 be very patient if we are in parallel ....*/
	int cnt = 0;
	
	rt->pa_socket = -1;
	
	while( cnt < 1500 )
	{
		rt->pa_socket =  Socket_connect_to_server( target->addr,target->port);
		
		if( 0 < rt->pa_socket )
		{
			break;
		}
		
		
		/* Set an increasing timeframe
		 from 10 ms to 1.5 sec*/
		int timeval = cnt * 1000;
		timeval = (timeval < 10000)?10000:timeval;
		
		usleep(timeval);
		cnt+=100;
	}
	
	/* Check that we are connected */
	
	if( rt->pa_socket < 0 )
	{
		return 1;
	}
	
	printf("Connected to %s:%s\n", target->addr, target->port );
	
	return 0;
}


int Red_Tree_init( struct Red_Tree *rt, char * (target_host_handler)(struct Red_Tree *), int local_port )
{
	
	int real_port = local_port;

	rt->lc_fd = -1;
	rt->rc_fd = -1;
	rt->pa_socket = -1;
	rt->instance_id = 0;
	rt->data_handler = NULL;
	
	/* Retrieve hostname */
	if( gethostname(rt->hostname, 512) < 0 )
	{
		perror("gethostname");
		return 1;
	}
	
	/* Setup pa_socket_lock mutex */
	if( pthread_mutex_init(&rt->parent_socket_lock, NULL) != 0 )
	{
		perror("pthread_mutex_init");
		return 1;
	}
	
	/* Init node list */
	Node_list_init( &rt->node_list );
	
	/* Init the unpacker */
	Data_Unpacker_init( &rt->unpacker, 4241 );
  
	/* Start a listening server on any port (if port is 0) */
	rt->listening_socket = Socket_listen_on_free_port( &real_port );
	
	printf("%s now listening on %d\n", rt->hostname, real_port );
	
	if( rt->listening_socket < 0 )
	{
		return 1;
	}
	
	/* We can now initalize listening socket CTX */
	Connection_ctx_init_static( &rt->listening_socket_ctx, rt->hostname, real_port );
	
	/* Start accept thread on the listening socket */
	if( pthread_create(&rt->accept_thread, NULL, Red_Tree_accept_thread, (void*)rt ) != 0 )
	{
		perror("pthread_create");
		return 1;
	}
	
	

	/* Client socket (Parent link) */
	struct Connection_ctx target_node;
	
	memset(&target_node, 0, sizeof(struct Connection_ctx) );
	
	/* Parse host informations if present */
	rt->pa_socket = -1;
	
	
	char * target_host = NULL;
	
	if( target_host_handler )
		target_host = target_host_handler( rt );
	
	if( target_host )
	{
		if( Connection_ctx_init( &target_node, target_host ) )
		{
			printf("Failed to parse address %s\n", target_host );
			return 1;
		}
		
		
		if( Red_Tree_connect( rt, &target_node ) )
		{
			printf("Failed to connect to %s\n", target_host );
			return 1;
		}
		
	}
	
	free( target_host );
	
	return 0;
}

int Red_Tree_release( struct Red_Tree *rt )
{
  printf("RELEASE\n");
  /* Wait for children */
  if( 0 < rt->lc_fd )
  {
	pthread_join(rt->lc_progress_thread, NULL);
	shutdown(rt->lc_fd, SHUT_RDWR );
	close(rt->lc_fd);
  }
  printf("LC\n");
 
  if( 0 < rt->rc_fd )
  {
	pthread_join(rt->rc_progress_thread, NULL);
	shutdown(rt->rc_fd, SHUT_RDWR );
	close(rt->rc_fd);
  }
  printf("RC\n");
  
  /* Close listening socket */
  shutdown(rt->listening_socket, SHUT_RDWR );
  close( rt->listening_socket );
  pthread_join(rt->accept_thread, NULL);
  printf("LstSCK\n");
  
  /* Close parent socket */
  if( 0 < rt->pa_socket )
  {
	shutdown(rt->pa_socket, SHUT_RDWR );
	close( rt->pa_socket );
  }
  printf("PA\n");
  
  
  Connection_ctx_release( &rt->listening_socket_ctx );
  Data_Unpacker_release(&rt->unpacker );
  
  return 0;
}

int Red_Tree_set_handler( struct Red_Tree *rt, void (*data_handler)( struct Red_Tree *, char * , size_t ) )
{
	rt->data_handler = data_handler;
	return 0;
}


int Red_Tree_push_pack( struct Red_Tree *rt, struct Data_pack *pack )
{
	int ret = 0;
	
	pthread_mutex_lock(&rt->parent_socket_lock);
	
	if( Socket_safe_write( rt->pa_socket, pack, sizeof( struct Data_pack ) ) < 0 )
	{
		perror("Send");
		ret = 1;
	}
	
	pthread_mutex_unlock(&rt->parent_socket_lock);
	
	return ret;
}




int Red_Tree_push_data( struct Red_Tree* rt, void* data, int size )
{
	if( !rt )
	{
		printf("Warning NULL rt\n");
		return 1;
	}
	
	if( rt->pa_socket < 0 )
	{
		/* Directly call the handler as we are the root */
		if( rt->data_handler )
			(rt->data_handler)( rt, data, size );
		
		return 0;
	}
	
	/* We have to split in packs */
	struct Data_pack pack;
	
	int remaining_size = size;
	int to_write = PACK_SIZE_UNIT;
	int offset = 0;
	
	/* Initial pack setup */
	pack.is_eof = 0;
	pack.source_id = rt->instance_id;
	pack.total_size = size;
	
	do
	{
		to_write = (remaining_size<PACK_SIZE_UNIT)?remaining_size:PACK_SIZE_UNIT;
		
		/* Fill in pack infos */
		pack.size = to_write;
		/* Copy data segment */
		memcpy(pack.data, data + offset, to_write);
		
		/* Push pack in the socket */
		if( Red_Tree_push_pack( rt, &pack ) )
		{
			printf("Error pushing pack\n");
			return 1;
		}
		
		offset += to_write;
		remaining_size -= to_write;
	}while( remaining_size );
	
	return 0;
}




