/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

#ifndef VMPI
#define VMPI


#include <mpi.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

/**
 * @ingroup Virtual_MPI_
 * @brief Defines of magic values
 */
#define VMPI_NG_ARBITRARY_VALUE 0x1234567llu
/**
 * @ingroup Virtual_MPI_
 * @brief Maximum number of tags allowed
 */
#define VMPI_NG_MAXIMUM_TAGS 1000000
/**
 * @ingroup Virtual_MPI_
 * @brief This value is used to refer to MPI_COMM_UNIVERSE from the host application
 */
#define MPI_COMM_UNIVERSE ((MPI_Comm)VMPI_NG_ARBITRARY_VALUE)

/**
 * @ingroup Virtual_MPI_
 * @brief Enum for error handling
 */
enum VMPI_Return_Status
{
	VMPI_EAGAIN=-3, 	/**< @brief The call cannot pursue retry */
	VMPI_ERROR=-2,		/**< @brief Something went wrong */
	VMPI_MEM=-1,		/**< @brief Failed to allocate some memory */
	VMPI_SUCCESS=0		/**< @brief All OK */
};

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief indicates if VMPI is enabled
 * @return 1 if enabled, 0 otherwise
 */
int VMPI_Enabled();

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief Initializes VMPI
 * @param argc
 * @param argv
 * @return a @link VMPI_Return_Status @endlink
 */
int VMPI_Init(int * argc, char ***argv);

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief Releases VMPI
 * @return a @link VMPI_Return_Status @endlink
 */
int VMPI_Release();

/**
 * @ingroup Virtual_MPI_
 * @brief Maximum size of a program name
 */
#define VMPI_PNAME_LEN 200
/**
 * @ingroup Virtual_MPI_
 * @brief Maximum command line name
 */
#define VMPI_COMMAND_LEN 1024

/**
 * @ingroup VMPI_Partitions_
 * @brief Structure holding partitions descriptions
 */
typedef struct 
{
	int id; 								/**< @brief Unique identifier of the partition */
	int root_id; 							/**< @brief Identifier of the root of the partition */
	int size; 								/**< @brief Number of tasks involved in the partition */
	char name[200];							/**< @brief Name of the partition (could be NONE) */
	char program_name[VMPI_PNAME_LEN];		/**< @brief Program name as stated in argv[0] */
	char command_line[VMPI_COMMAND_LEN];	/**< @brief Complete command used to invoke the program */
	MPI_Comm partition_comm;                /**< @brief The communicator of the partition */
}VMPI_Partition_desc;


/**
 * @ingroup Virtual_MPI_
 * @brief Global status handling
 */
struct VMPI_Status
{
	int mypartition;                            /**< @brief Current partition id */
	int numpartitions;                          /**< @brief Partitions number*/
	int (*vmpi_trampoline)( int argc, char **argv );
	char trampoline_partition_name[200];
	char partition_name[200];                   /**< @brief Current partition name*/
	int current_tag;							/**< @brief Counter for local process tags */
	int vmpi_enabled;                           /**< @brief VMPI is enabled */
	MPI_Comm vmpi_communicator;                 /**< @brief current MPI communicator */
	VMPI_Partition_desc *partition_descs;       /**< @brief Partitions descriptions*/
};

/**
 * @ingroup Virtual_MPI_
 * @brief global VMPI status
 */
extern struct VMPI_Status __vmpi_status;

/**
 * @ingroup VMPI_Partitions_
 * @brief Returns the unique ID of the current partition
 *
 * Current partition ID is between 0 and VMPI_Get_partition_count()
 * @return The current partition unique ID
 */
static inline int VMPI_Get_partition_id()
{
	return __vmpi_status.mypartition;
}


/**
 * @ingroup VMPI_Partitions_
 * @brief Returns current partition's descriptions
 * @return The VMPI_Partition_desc corresponding to the current partition
 */
static inline VMPI_Partition_desc *VMPI_Get_desc()
{
	int id = VMPI_Get_partition_id();
	return &__vmpi_status.partition_descs[id];
}


/**
 * @ingroup VMPI_Partitions_
 * @brief Return the number of partitions
 * @return The number of partitions
 */
static inline int VMPI_Get_partition_count()
{
	return __vmpi_status.numpartitions;
}

/**
 * @ingroup VMPI_Partitions_
 * @brief Returns current partition's communicator
 * @return The MPI_Comm corresponding to the current partition
 */
static inline MPI_Comm VMPI_Get_partition_comm()
{
	return VMPI_Get_desc()->partition_comm;
}

/**
 * @ingroup VMPI_Partitions_
 * @brief Used to setup partition name for a given executable
 *
 * IT MUST be called before MPI_Init which will dispatch
 * this information to every processes.
 * @param name The new name for the current partition
 * @return VMPI_SUCCESS
 */
static inline int VMPI_Set_partition_name( char *name )
{
	size_t ret = snprintf(__vmpi_status.partition_name, 200, "%s" , name );
        assert( ret < 200 );
        __vmpi_status.partition_name[ret] = '\0';
	return VMPI_SUCCESS;
}

static inline int VMPI_partition_Test_partition_name( char* test_name )
{
   const size_t partition_name_len = strlen( __vmpi_status.partition_name );

   if( strlen( test_name ) != partition_name_len )
      return 0;
   
   if( strcmp( test_name, __vmpi_status.partition_name ) )
      return 0;

   return 1; 
}


static inline int VMPI_Set_trace_trampoline( char *trampoline_partition,
										     int (*vmpi_trampoline)( int argc, char **argv ) )
{
	if( !vmpi_trampoline || !trampoline_partition )
		return 1;
		
	
	strncpy( __vmpi_status.trampoline_partition_name, trampoline_partition, 200 );
	__vmpi_status.vmpi_trampoline = vmpi_trampoline;
	
	return 0;
}

static inline int VMPI_In_trampoline()
{
	if( __vmpi_status.vmpi_trampoline != NULL )
	{
		if( VMPI_Get_partition_id() == 1 )
			return 1;
	}
	
	return 0;
}
										    


/**
 * @ingroup VMPI_Partitions_
 * @brief Find a descriptions by its name
 * @param name The wanted partition's name
 * @return The VMPI_Partition_desc corresponding to the partition named "name". NULL if not found
 */
static inline VMPI_Partition_desc *VMPI_Get_desc_by_name(char *name)
{
	int i;

	for( i = 0 ; i < VMPI_Get_partition_count() ; i++ )
	{
		if( !strcmp( __vmpi_status.partition_descs[i].name, name ) )
			return &__vmpi_status.partition_descs[i];
	}

	return NULL;
}

/**
 * @ingroup VMPI_Partitions_
 * @brief Find a descriptions by its ID
 * @param id The wanted partition's ID
 * @return The VMPI_Partition_desc corresponding to the partition with ID "id". NULL if not found
 */
static inline VMPI_Partition_desc *VMPI_Get_desc_by_id(int id)
{
	if( VMPI_Get_partition_count() <= id )
		return NULL;

	return &__vmpi_status.partition_descs[id];

}



/**
 * @ingroup VMPI_Partitions_
 * @brief Prints all the fields of a partition descriptions
 * @param desc The description to be printed
 */
void VMPI_Display_desc(VMPI_Partition_desc *desc);

/**
 * @ingroup VMPI_Partitions_
 * @brief Display all the partitions descriptions in order
 */
void VMPI_Display_descs();

/**
 * @ingroup VMPI_Maps_
 * @brief Defines mapping method
 */
typedef enum
{
	VMPI_MAP_FIXED,			/**< @brief Associates each process to one process */
	VMPI_MAP_RANDOM,		/**< @brief Randomly associates processes */
	VMPI_MAP_ROUND_ROBIN	/**< @brief Map processes in a round robin fashion */
}VMPI_Map_mode;


/**
 * @ingroup VMPI_Maps_
 * @brief Defines a map
 */
typedef struct
{
	int count;              /**< @brief Number of ranks */
	int *ranks;             /**< @brief Target ranks */
}VMPI_Map;

/**
 * @ingroup VMPI_Maps_
 * @brief Clears a map
 *
 * should be called before any call to VMPI_Map_partition
 * @param map The map to be cleared
 */
static inline void VMPI_Map_clear( VMPI_Map *map )
{
	map->count = 0;
	map->ranks = NULL;
}

/**
 * @ingroup VMPI_Maps_
 * @brief Map to another partition (can be additive)
 * @param target_partition ID of the target partition
 * @param mode mapping mode as stated in VMPI_Map_mode
 * @param map output map
 * @return a @link VMPI_Return_Status @endlink
 *
 * This functions performs a "real" collective call. It means that not only one program will do that, but every partitions (ie. every programs given to mpirun)
 *
 */
int VMPI_Map_partitions( int target_partition, VMPI_Map_mode mode, VMPI_Map *map );
void VMPI_Print_map( VMPI_Map *map );
/**
 * @ingroup VMPI_Internal_
 * @brief Get VMPI's copy of MPI_COMM_WORLD
 */
static inline MPI_Comm VMPI_Get_vmpi_comm()
{
	if( !VMPI_Enabled())
		return MPI_COMM_WORLD;

	return __vmpi_status.vmpi_communicator;
}

//FIXME: partition name can change
static inline int VMPI_partition_partition_is_analyzer( void )
{
   return VMPI_partition_Test_partition_name( "Analyzer" );
}

/**
 * @ingroup VMPI_Internal_
 * @brief Book a tag on VMPI's communicator for this process
 * @return the newly booked tag. VMPI_ERROR if something goes wrong. 
 */
int VMPI_get_new_tag();

#include "VMPI_stream.h"

#ifdef __cplusplus
}
#endif

#endif /*VMPI*/
