/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

var fs = require('fs');
var spawn = require('child_process').spawn;

/*************************************************/
/* Extend the string class with useful functions */
/*************************************************/

String.prototype.contains = function( needle )
{
	var match = this.match(".*" + needle +".*");

	if( match )
		return 1;
	
	return 0;
}

String.prototype.trim = function()
{
	return this.replace(/^\s+|\s+$/g, '').replace(/\s+/g, ' ');
}

String.prototype.toComment = function()
{
	return "/*** " + this + " ***/";
}

String.prototype.regExp = function() {
  return this.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\$&");
}

/*************************************************/
/*           Help and exit handlers              */
/*************************************************/

function printHelp()
{
	console.log("/************************************************************/");
	console.log("/* posix_wrap [CC] [CFLAGS] [SCRIPT] [OUTPUT] [HEADERLIST]  */");
	console.log("/* CC : 	  C compiler                                     */");
	console.log("/* CFLAGS : C compiler compilation flags                    */");
	console.log("/* SCRIPT : Optionnal for mode Generate                     */");
	console.log("/************************************************************/");
}


function exitError( error )
{
	printHelp();
	
	if( error )
		console.log( error.toComment() );
		
	process.exit(1);
}

/*************************************************/
/*         Argument parsing and validation       */
/*************************************************/

var args = process.argv.splice(2);

if( args.length < 3 )
{
	exitError("Missing arguments");
}

var cc = args[0].trim();
var cflags = args[1].trim();

var script = undefined;

if( 2 < args.length )
{
	script = args[2].trim();
}


var output_file = "./out.c"

if( 3 < args.length )
{
	output_file = args[3].trim();
}


/* Process header list */

var header_list="stdio.h," +
				"stdlib.h," +
				"errno.h," +
				"unistd.h";

if( 4 < args.length )
{
	header_list = args[4].trim();
}

var tmp_header_list = header_list.split(',');

var i = 0;
header_list = "";
for( i = 0 ; i < tmp_header_list.length ; i++ )
{
	header_list += "#include <" + tmp_header_list[i] + ">\n";
}

/*************************************************/
/* Wrapper script loading                        */
/*************************************************/
/* READ SCRIPT */
var script_data = "";

try
{
	script_data = fs.readFileSync( script, { encoding : "ascii" }  );
}
catch(e)
{
	exitError("Failed to open script :" + script);
}
	
var script_lines = script_data.split('\n');


/***********************************************************/
/*         Add configuration in the generated file         */
/***********************************************************/

console.log(("Compiler : " + cc).toComment());
console.log(("CFLAGS : " + cflags).toComment());

if( script )
{
	console.log(("Script : " + script).toComment());
}

/***********************************************************/
/*     Utils function to interact with compiler            */
/***********************************************************/

/* Write a c code and returns the file path */
function writeAtestFile( code )
{
	//console.log( code );
	/* Generate a temporary file name */
	var tmp_file = "/tmp/temp" + Math.round(Math.random() * 1e6) + ".c";

	console.log( tmp_file );

	/* Write the test file */
	try
	{
		fs.writeFileSync(tmp_file, code);
	}
	catch(e)
	{
		exitError("Failed to write test program");
	}
	
	return tmp_file;
}

function generateMain( body )
{
	return "#define _GNU_SOURCE\n"
			+ header_list
			+ "int main(int argc, char **argv){" + body + "}";
	
}

/* Compile the code then call an handler handler( preprocessed_output ) */
function preprocessCode( file, cc, cflags, handler )
{
	var compiler_data = "";
	var compiler_process;
	var compiler_args = [];

	//console.log(file);

	compiler_args = compiler_args.concat(["-E", file]).concat( cflags.split(" ")  );
		
	try
	{
		compiler_process = spawn(cc, compiler_args);
	}
	catch(e)
	{
		exitError("Failed to launch command " + cc );
	}

	compiler_process.stdout.on('data', function (data) {
	  compiler_data += data;
	});


	compiler_process.stderr.on('data', function (data) {
	  console.log(data);
	});

	compiler_process.on('close', function (code)
	{
		fs.unlinkSync(file);
		
		if( code != 0 )
		{
			exitError("Compiler returned an error code:" + code );
		}

		if( handler )
			handler( compiler_data );

	});
}


/* Compile the code with given cflags and calls an handler with the return code */
function compileCode( file, cc, cflags, handler )
{
	var compiler_data = "";
	var compiler_process;
	var compiler_args = [];
	
	//console.log(file);
	
	compiler_args = compiler_args.concat([file, "-ldl", "-rdynamic", "-Werror"]).concat( cflags.split(" ")  );
	

	try
	{
		compiler_process = spawn(cc, compiler_args);
	}
	catch(e)
	{
		exitError("Failed to launch command " + cc );
	}

	compiler_process.stderr.on('data', function (data) {
	  compiler_data += data;
	});

	compiler_process.on('close', function (code)
	{
		/* Delete source file */
		fs.unlinkSync(file);
		
		/* If program compiled now try to run it */
		var run_process;
		
		try
		{
			var run_process =  spawn("./a.out", []);
		}
		catch(e)
		{
			console.log("Failed to launch binary");
			return;
		}
		
		run_process.on('close', function (code)
		{
				if( handler )
					handler( code, compiler_data );
		});
		
		



	});
	
	
}



/*********************************************************************/
/* Wrapper script compilation                                        */
/* BE CAREFUL THIS SECTION IS DUPLICATED in the two wrappers         */
/*********************************************************************/

/* Script entity */

function script_action( type, line, filter, params )
{	
	this.type = type;
	this.line = line;

	this.is_whitelisted = 0;
	this.whitelist = {};
	this.blacklist = {};
	
	if( filter != "" )
	{
		var sp = filter.split(",");
		
		var i;
		
		for( i = 0; i < sp.length ; i++ )
		{
			sp[i] = sp[i].trim();
	
			if( sp[i][0] == "!" )
			{
				
				if( this.is_whitelisted )
				{
					console.log("You cannot whitelist while blacklisting: " + filter );
					process.exit(1);
				}

				this.blacklist[ sp[i].slice(1) ] = 1;
			}
			else
			{
				this.is_whitelisted = 1;
				this.whitelist[ sp[i] ] = 1;
			}
			
		}
	}

	this.params = params;
	
	this.actions = [];
}

/* Parser object */

function script_parser()
{
	this.is_in_wrapped_section = 0;
	this.actions = [];
	
	this.output = "";
}

script_parser.prototype.parseline = function( line )
{
	/* Section Start */
	var match = line.match("<MALPW(.*)>");
	
	if( match != null )
	{
		var filter = match[1];
		
		this.is_in_wrapped_section = 1;
		
		this.actions.push( new script_action( "section", line, filter ) );
		
		return;
	}
	
	match = line.match("</MALPW>");
	
	if( match != null )
	{
		this.is_in_wrapped_section = 0;
		
		return;
	}
	
	/* We do actions only in wrapped sections */
	if( this.is_in_wrapped_section )
	{
		var current_action = this.actions[this.actions.length - 1 ];

		if( current_action.type != "section" )
		{
			console.log("You can only put commands in sections " + script + ":" + (i + 1) );
			process.exit(1); 
		}
		
		match = line.match("MALPW_PROTOTYPE\\(\\);");
		
		if( match != null )
		{
			current_action.actions.push( new script_action( "prototype", "", "" ) );
			return;
		}

		match = line.match("MALPW_GEN_RET\\(\\);");
		
		if( match != null )
		{
			current_action.actions.push( new script_action( "ret", "", "" ) );
			return;
		}
		
		match = line.match("MALPW_NAME\\(\\);");
		
		if( match != null )
		{
			current_action.actions.push( new script_action( "name", "", "" ) );
			return;
		}

		match = line.match("MALPW_WRAP\\((.*),(.*),(.*)\\);");
		
		if( match != null )
		{
			if( match.length != 4 )
			{
				console.log("Missing arguments to MALPW_WRAP" + script + ":" + (i + 1) );
				process.exit(1);
			}
			
			var type = match[1];
			var filter = match[2];
			var macro = match[3];
			
			current_action.actions.push( new script_action( "wrap", "", "", { "type" : type , "filter" : filter, "macro" : macro } ) );
			return;
		}
		
		match = line.match("MALPW_FOOTPRINT\\((.*)\\);");
		
		if( match != null )
		{
			var prefix = "";
			
			if( match.length == 2 )
			{
				prefix = match[1].trim();				
			}
			
			current_action.actions.push( new script_action( "footprint", "", "", { "prefix" : prefix } ) );
			return;
		}
		
				
		match = line.match("MALPW_WNAMECASE_UP\\((.*)\\);");
		
		if( match != null )
		{
			var prefix = "";
			
			if( match.length == 2 )
			{
				prefix = match[1].trim();				
			}
			
			current_action.actions.push( new script_action( "caseup", "", "", { "prefix" : prefix } ) );
			return;
		}
		
		match = line.match("MALPW_WNAME\\((.*)\\);");
		
		if( match != null )
		{
			var prefix = ""
			
			if( match.length == 2 )
			{
				prefix = match[1].trim();
			}

			current_action.actions.push( new script_action( "wname", "", "", { "prefix" : prefix } ) );
			return;
		}		

		match = line.match("MALPW_CALL\\((.*)\\);");
		
		if( match != null )
		{

			var prefix = "";
			
			if( match.length == 2 )
			{
				prefix = match[1].trim();
			}
			
			current_action.actions.push( new script_action( "call", "", "", { "prefix" : prefix } ) );
			return;
		}
		
		match = line.match("MALPW_HAS_ARG_B\\((.*),(.*)\\);");
		
		if( match != null )
		{

			if( match.length != 3 )
			{
				console.log("Missing arguments to MALPW_HAS_ARG_B" + script + ":" + (i + 1) );
				process.exit(1);
			}
			
			var type = match[1];
			var name = match[2];
			
			current_action.actions.push( new script_action( "argb", "", "", { "type" : type, "name" : name } ) );
			return;
		}
		
		match = line.match("MALPW_HAS_ARG_E(.*);");
		
		if( match != null )
		{

			var type = match[1];
			var name = match[2];
			
			current_action.actions.push( new script_action( "arge", "", "", { "type" : type, "name" : name } ) );
			return;
		}
		
		match = line.match("MALPW_CALL_SIZE\\((.*)\\);");
		
		if( match != null )
		{
			var var_name = "";
			
			if( match.length == 2 )
			{
				var_name = match[1].trim();				
			}
			
			current_action.actions.push( new script_action( "size_for_call", "", "", { "var_name" : var_name } ) );
			return;
		}

		match = line.match("MALPW_CALL_PTR\\((.*)\\);");
		
		if( match != null )
		{
			var var_name = "";
			
			if( match.length == 2 )
			{
				var_name = match[1].trim();				
			}
			
			current_action.actions.push( new script_action( "ptr_for_call", "", "", { "var_name" : var_name } ) );
			return;
		}

		/* If we did not match we push text */
		current_action.actions.push( new script_action( "text", line, "" ) );
		return;
	}
	
	/* Actions which are out of wrapped sections */
	match = line.match("MALPW_GEN_HEADERS\\(\\);");

	if( match != null )
	{
		this.actions.push( new script_action( "headers", "", "" ) );
		return;
	}
	
	
	/* IF it did not match it is text */
	this.actions.push( new script_action( "text", line, "" ) );
	
}


script_parser.prototype.do_action = function( action, current_func )
{
	var ret = "";
	var i;
	
	switch( action.type )
	{
			case "text" :
				ret = action.line;
				break;
			case "prototype" :
				if( current_func == undefined )
					return;
					
				ret = current_func.ret + " " + current_func.name + "( ";
				
				for( i=0; i < current_func.args.length ; i++ )
				{
					ret += current_func.args[i][0] + " " + current_func.args[i][1];
					if( i != (current_func.args.length - 1 ) )
						ret += ", ";
					else
						ret += " ";
				}
				
				ret += ")"
				
				break;
			case "wrap" :
				if( current_func == undefined )
					return;

				for( i=0; i < current_func.args.length ; i++ )
				{
					if( current_func.args[i][0].trim() == action.params['type'].trim()
						&& ( (current_func.args[i][1] == action.params['filter'].trim()) || action.params['filter'].trim() == "*" )
					  )
					{
						ret += "\t" + action.params['macro'] + "(" + current_func.args[i][1].trim() + ")";
					}
				}
				
				break;
			case "name" :
					ret += "\t" + current_func.name.trim();
				break;
			case "wname" :
				if( action.params['prefix'] != "" )
					ret += "\t" + action.params['prefix'] + "(" + current_func.name.trim() + ")";
				else
					ret += "\t" + current_func.name.trim();
				break;
			case "caseup" :
				if( action.params['prefix'] != "" )
					ret += "\t" + action.params['prefix'] + "(" + current_func.name.toUpperCase().trim() + ")";
				else
					ret += "\t" + current_func.name.toUpperCase().trim();
				break;
			case "footprint" :
				if( current_func == undefined )
					return;
			
				if( action.params['prefix'] != "" )
					ret = "#define " + action.params['prefix'].trim() + current_func.name.trim() + " ";
					ret += current_func.ret + " (*)(";
					
					for( i=0; i < current_func.args.length ; i++ )
					{
						ret += current_func.args[i][0];
						if( i != (current_func.args.length - 1 ) )
							ret += ", ";
						else
							ret += " ";
					}
					
					ret += ")";
					
				break;
				break;		
			case "call" :
				if( current_func.ret != "void" )
					ret = "\t" + current_func.ret + " ret = ";
				
				if(  action.params['prefix'] != "" )
					ret += action.params['prefix'] + "(" + current_func.name + ") ( ";
				else
					ret += current_func.name + "( ";
				
				for( i=0; i < current_func.args.length ; i++ )
				{
					ret += current_func.args[i][1].replace(/\[.*\]/g, '');
					if( i != (current_func.args.length - 1 ) )
						ret += ", ";
					else
						ret += " ";
				}
				
				ret += ");"
				break;			
			case "argb" :
				if( current_func == undefined )
					return;

				var found = 0;

				for( i=0; i < current_func.args.length ; i++ )
				{
					if( current_func.args[i][0].trim() == action.params['type'].trim()
						&& ( (current_func.args[i][1] == action.params['name'].trim()) || action.params['name'].trim() == "*" )
					  )
					{
						found = 1;
						break;
					}
				}
				
				ret = "\t#if " + found;
				break;
			case "size_for_call" :
				ret = "\tint64_t " + action.params['var_name'].trim() + " = " + current_func.size_management + ";";
				break;
			case "ptr_for_call" :
				ret = "\tvoid * " + action.params['var_name'].trim() + " = " + current_func.ptr_management + ";";
				break;
			case "arge" :
				ret = "\t#endif";
				break;
			case "ret" :
				if( current_func.ret != "void" )
					ret = "\t" + "return ret;";
				break;		
			case "headers" :
					ret = "#define _GNU_SOURCE\n" + header_list;
				break;		
	}
	
	this.output += ret + "\n";
}	


script_parser.prototype.gen = function(functions)
{
	var i;
	var j;
	var k;
	
	for( i = 0 ; i < this.actions.length ; i++ )
	{
		if( this.actions[i].type != "section" )
		{
				this.do_action( this.actions[i] );
		}
		else
		{

			for( j = 0 ; j < functions.length ; j++ )
			{
				if( functions[j].name == undefined )
					continue;
				
				if( this.actions[i].is_whitelisted )
				{
					if( this.actions[i].whitelist[ functions[j].name ] == undefined )
						continue;
				}
				else
				{
					if( this.actions[i].blacklist[ functions[j].name ] != undefined )
					{
						continue;
					}
				}
					
				for( k = 0 ; k < this.actions[i].actions.length ; k++ )
				{
					this.do_action( this.actions[i].actions[k],  functions[j] );
				}
			}
			
		}
		
		
	}
	
	
	
	
}


/***********************************************************/
/*              Generate Knowledge BASE                    */
/***********************************************************/

/* List of supported types for parsing */
var type_array = 
'char|int|size_t|void|void\\*|char\\*|H5Z_filter_t|float|FILE\\*|__ssize_t|va_list'+
'|__off64_t|unsigned long long int|unsigned long int|unsigned int|short|unsigned short'+
'|unsigned short int\\*|double|long double|ssize_t|__pid_t|pid_t|const char\\*|long double|sem_t\\*'+
'|long int|long long int|div_t|ldiv_t|lldiv_t|__useconds_t|useconds_t|__locale_t|struct group\\*'+
'|pthread_condattr_t\\*|pthread_cond_t\\*|const pthread_rwlockattr_t\\*|pthread_rwlockattr_t\\*|pthread_rwlock_t\\*'+
'|pthread_mutexattr_t\\*|const pthread_mutexattr_t\\*|pthread_mutex_t\\*|pthread_attr_t\\*|struct utmpx\\*|wctrans_t' +
'|pthread_t|pthread_t\\*|const sigset_t\\*|sigset_t\\*|siginfo_t|siginfo_t\\*|__sighandler_t|imaxdiv_t|wctype_t' +
'|pthread_spinlock_t\\*|herr_t|htri_t|hid_t|H5S_sel_type|hssize_t|haddr_t|wint_t|wchar_t\\*|__uid_t|__gid_t'+
'|struct tm\\*|const time_t\\*|time_t\\*|clock_t|speed_t|const struct termios\\*|struct servent\\*|struct protoent\\*'+
'|struct netent\\*|struct hostent\\*|struct rpcent\\*|in_addr_t|struct in6_addr\\*|socklen_t'+
'|uint8_t|uint16_t|uint32_t|uint64_t|int8_t|int16_t|int32_t|int64_t|uint8_t\\*|uint16_t\\*|struct stat64\\*'+
'|uint32_t\\*|uint64_t\\*|int8_t\\*|int16_t\\*|int32_t\\*|int64_t\\*|struct cmsghdr\\*|DIR|DIR\\*|struct dirent|struct dirent\\*' +
'|struct dirent64|struct dirent64\\*|double complex|long double complex|iconv_t|reg_syntax_t|struct passwd|struct passwd\\*|__off_t|float complex';

/* Here we define how size related informations
   are parsed in dynamically wrapped functions
   
   Syntax is #1-N -> function arguments left to right
             #ret -> Value returned from the call

   Result is stored in an int64_t
   Within scripts use the MALPW_CALL_SIZE macro
   only after the call
*/	

var sized_functions_KB = 
{ "fread" : "#ret * #1",
  "fwrite" : "#ret * #1",
  "fgetc" : "(#ret != EOF)?1:0",
  "fgets" : "(#ret)?strlen(ret):0",
  "getc" : "(#ret != EOF)?1:0",
  "getchar" : "(#ret != EOF)?1:0",
  "ungetc" : "(#ret != EOF)?1:0",
  "fputc" : "(#ret != EOF)?1:0",
  "fputs" : "(#ret != EOF)?strlen(#0):0",
  "putc" : "(#ret != EOF)?1:0",
  "putw" : "(#ret != EOF)?4:0",
  "getw" : "(#ret != EOF)?4:0",
  "putchar" : "(#ret != EOF)?1:0",
  "puts" : "(#ret != EOF)?strlen(#0):0",
  "getc_unlocked" : "1",
  "getchar_unlocked" : "1",
  "putc_unlocked" : "1",
  "putchar_unlocked" : "1",
  "perror" : "strlen(#0) + 2 + strlen( sys_errlist[ errno ] )",
  "trucate" : "#1",
  "ftrucate" : "#1",
  "pread" : "(0 <= #ret != EOF)?#ret:0",
  "pwrite" : "(0 <= #ret != EOF)?#ret:0",
  "read" : "(0 <= #ret != EOF)?#ret:0",
  "write" : "(0 <= #ret != EOF)?#ret:0",
  "malloc" : "#0",
  "calloc" : "#0 * #1",
  "realloc" : "#1"
}

/* Here we define what is the argument which will
   be put in the the pointer section of the performance
   event note that it is a void *, syntax is the same
   as what has been described for the size i

   Within scripts us the MALPW_CALL_PTR macro
   only after the call has been emitted
*/

var pointer_functions_KB = 
{
  "malloc" : "#ret",
  "calloc" : "#ret",
  "realloc" : "#ret",
  "free" : "#0",
  "fopen" : "#ret",
  "fdopen" : "#ret",
  "freopen" : "#ret",
  "creat" : "#ret"
}

/* First we launch the compiler in purpose of getting its outpout */

function posixFunction(ret, name , args)
{
	/* 
	 * Function initialization 
	 * 
	 * */
	
	this.ret = ret.trim();
	this.name = name.trim();

	/* Here we filter the main */
	if( this.name == "main")
	{
		this.name = undefined;
		return;
	}
	
	/* Here we filter the main */
	if( this.name.match(".*_chk") )
	{
		this.name = undefined;
		return;
	}
	
	this.args = args;
	
	var tmp_array = [];
	var i;
	for( i = 0 ; i < this.args.length ; i++ )
	{
		var cur = this.args[i].trim();
		
		if( cur.length != 0 && cur != 'void' )
		{
			tmp_array.push( cur );
		}
	}
	
	this.args = tmp_array;
	var cnt = 0;
	
	/* We now have to parse arguments to become [0] type [1] name */
	for( i = 0 ; i < this.args.length ; i++ )
	{
		var current_arg = this.args[i];
		var last_star_index = current_arg.lastIndexOf("*") + 1;
		var last_space_index = current_arg.lastIndexOf(" ") + 1;
		
		if( last_space_index < last_star_index )
		{
			this.name = undefined;
			return;
		}
		

		var name = current_arg.substring( last_space_index );
		var type = current_arg.substring( 0, last_space_index );
		
		
		/* At this point we want to make sure that name is not a type */
		
		var ret = name.match( "^(" + type_array.regExp() + ")$" );
		
		if( ret != null )
		{
			/*name is a type */
			type += name.trim();
			name = "var" + cnt++;
		}
		else
		{
			/* Make sure that name is alphanumeric */
			ret = name.match( "^[A-Za-z0-9_]+$" );
		
			if( ret == null )
			{
				this.name = undefined;
				return;
			}

		}
		
		
		this.args[i] = [];
		
		this.args[i][0] = type;
		this.args[i][1] = name;
	}
	
	/* Here all args have been parsed */
		
	/* For size generation */
	
	/* See if the function is in the "size KB" */
	this.size_management = "0";
	
	var size_element = sized_functions_KB[ this.name ];
	
	if( size_element )
	{
		/* We have the string matching the size expression */
		this.size_management = size_element;
	}
	
	/* See if the function is in the "pointer KB" */
	this.ptr_management = "(void *)-1";
	
	var ptr_element = pointer_functions_KB[ this.name ];

	if( ptr_element )
	{
		/* We have the string matching the ptr expression */
		this.ptr_management = ptr_element;
	}
		
	/* Proceed with substitution */
	for( var i = 0 ; i < this.args.length ; i++ )
	{
		var arg = this.args[i];
		
		if( ! arg )
			continue;

		/* Insert Argument */
		this.size_management = this.size_management.replace (new RegExp("#" + i, "g"), arg[1] );
		this.ptr_management = this.ptr_management.replace (new RegExp("#" + i, "g"), arg[1] );
	}
	
	this.size_management = this.size_management.replace (new RegExp("#ret", "g"), "ret" );	
	this.ptr_management = this.ptr_management.replace (new RegExp("#ret", "g"), "ret" );	
}



function KB()
{
	this.entries = [];
	
	KB.prototype.pushEntry = function( line )
	{
		if( line.contains( "typedef" ) )
			return;
		
		if( line.contains( "__gnuc_va_list" ) || line.contains( "__extension__" )  )
			return;
		
		var ret = line.match("("+ type_array + ") +([A-Za-z0-9_]+) *\\((.*)\\)");
		
		if( ret != null )
		{
			var func = new posixFunction( ret[1] , ret[2], ret[3].split(',') );
			
			if( func.name )
			{
				this.entries.push( func );	
			}
			
		}
		else
		{
			/* For debug to see missing arguments when a function
			   is not catched */
			//console.log("\t\t: No match for ->>" + line + "<<-" );
		}
		
	}
	
	KB.prototype.print = function()
	{
		var i = 0;
		
		for( i = 0 ; i < this.entries.length ; i++ )
		{
			console.log(this.entries[i].ret + " " + this.entries[i].name);
		}
	}
	
	KB.prototype.checkCompilation = function(callback, arity, depth)
	{
		if( depth == undefined )
			depth = 0;
			
		if( arity == undefined )
			arity = 1;
		
		if( this.entries.length <= depth )
		{
			if( callback )
				callback();
			return;
		}
		
		/* Check if the code compiles */
		var my_test_script = new script_parser();
		var current_depth = depth;
		var target_funcs = [];

		var i = 0;
		for( i = 0 ; i < arity ; i++ )
		{
			if( this.entries[ depth + i ] )
				if(this.entries[ depth + i ].checked != true)
					target_funcs.push( this.entries[ depth + i ] );
		}
		
		if( target_funcs.length == 0 )
		{
			this.checkCompilation( callback, arity, current_depth + arity );
			return;
		}
		
		my_test_script.parseline("<MALPW>");
		my_test_script.parseline("void * VAR(");
		my_test_script.parseline("MALPW_NAME();");	
		my_test_script.parseline(") = NULL;\n");
		my_test_script.parseline("MALPW_FOOTPRINT(FOOTPRINT_);");
		my_test_script.parseline("MALPW_PROTOTYPE();");
		my_test_script.parseline("{");
		my_test_script.parseline("if( im_inside_a_function ){ puts(\"Looping DETECTED\\n\"); abort();}\n");
		my_test_script.parseline("im_inside_a_function = 1;\n");
		my_test_script.parseline("void ** local = &VAR(");
		my_test_script.parseline("MALPW_NAME();");
		my_test_script.parseline(");");		
		my_test_script.parseline("if(!*local) *local = dlsym(RTLD_NEXT,");
		my_test_script.parseline("TS(");
		my_test_script.parseline("MALPW_NAME();");
		my_test_script.parseline(")");
		my_test_script.parseline(");");
		my_test_script.parseline("MALPW_CALL(REAL);");
		my_test_script.parseline("im_inside_a_function = 0;\n");
		my_test_script.parseline("MALPW_GEN_RET();");
		my_test_script.parseline("}");
		my_test_script.parseline("/<MALPW>");
		
		my_test_script.gen(target_funcs);
		
		
		var test_code =   "#define _GNU_SOURCE\n"
							+ "#include <dlfcn.h>\n"
							+ header_list
							+ "#define VAR(a) fn_addr ## a \n"
							+ "#define REAL(a) (( FOOTPRINT_ ## a ) VAR( a ) )\n"
							+ "#define TS(a) #a\n"
							+ "int im_inside_a_function = 0;\n"
							+ my_test_script.output + "\n\n"
							+ "int main(int argc, char **argv){\n";
		
			
	   for( var i = 0 ; i <  target_funcs.length ; i++)
	   {
		   test_code += "void * a" + target_funcs[i].name +"  = dlsym( RTLD_NEXT, \"" + target_funcs[i].name +"\" );\n"
		   test_code += "printf(\"%p\\n\", a" + target_funcs[i].name + ");\n"
		   test_code += "if( ! a" + target_funcs[i].name + ") return 1;\n";
		   test_code += "if( a" + target_funcs[i].name + " == " + target_funcs[i].name + " ) return 2;\n";
	   }
		
		test_code += "return 0;\n}\n";
		
		var tmp_file = writeAtestFile ( test_code );


											
		var current_kb = this;
		
		
		compileCode( tmp_file, cc, cflags, function( code, compiler_data ) 
		{
			var i = 0;
			
			if( code == 139 )
			{
				compiler_data += "Wrap loop detected (SEGFAULT) avoiding ...";
			}
			
			if( code == 134 )
			{
				compiler_data += "Wrap loop detected avoiding ...";
			}
			
			if( code == 2 )
			{
				compiler_data +=  "Wrap loop detected (SELF) avoiding ...";
			}
			
			if( code == 1 )
			{
				compiler_data +=  "dlsym lookup test failed avoiding ...";
			}
			
			if( code != 0 )
			{
				/* deactivate function only if we compiled a single one */
				if( arity == 1 )
				{
					/* Here we do some pretty printing */
					console.log("======================"+target_funcs[0].name+"======================");
					console.log( "\033[1;30m" + compiler_data + "\033[0m");
					console.log("Wrapping " + target_funcs[0].name + "... NO" );
					console.log("======================" + "======================" + target_funcs[0].name.replace(/[A-Za-z-0-9_]/g, '=') );
					target_funcs[0].name = undefined;
				}
			}
			else
			{
				/* However if things went well we can check all functions */
				for( i = 0 ; i < arity ; i++ )
				{
					if( !target_funcs[i] )
						continue;

					console.log("Wrapping " + target_funcs[i].name + "... YES" );
					target_funcs[i].checked = true;
				}
			}
			
			
			
			current_kb.checkCompilation( callback, arity, current_depth + arity );
		});

	}
	
	
}

var knowledge_base = new KB();


function generateKB()
{
	var tmp_file = writeAtestFile (generateMain(""));


	/* Launch the preprocessor on the program and fill the knownledge base */
	preprocessCode( tmp_file, cc, cflags, function( output )
	{
		/* Filter out empty lines and lines beginning with # */
		var tmp_lines = output.split('\n');
		var clean_output = "";
		var i;
		
		for( i = 0 ; i <  tmp_lines.length ; i++ )
		{
			if( tmp_lines[i].contains('#') )
				continue;
			
			if( (tmp_lines[i].trim()).length == 0 )
				continue;
			
			clean_output += tmp_lines[i].trim();
		}
	
		clean_output = clean_output.replace(/;/g, ';\n');
		
		/* Remove multiple spaces */
		clean_output = clean_output.replace( / +/g, ' ' );
			
		/* Remove attributes, extern  __extension__ */
		clean_output = clean_output.replace(/[\s]*__attribute__[\s]*\(.*\)[\s]*/g, ' ');
		clean_output = clean_output.replace(/[\s]*__asm__[\s]*\(.*\)[\s]*/g, ' ');
		clean_output = clean_output.replace(/extern/g, ' ').replace(/__restrict/g, ' ');
		//clean_output = clean_output.replace(/__extension__/g, '');
		clean_output = clean_output.replace(/_Complex/g, 'complex');
		
		/* Move start to the left */
		clean_output = clean_output.replace(/ +\*/g, '* ');
		
		/* And separate lines */
		clean_output = clean_output.replace(/\n/g,'').split(";");
		
		//console.log( clean_output );
		
		/* Push lines in the KB */
		var i = 0;
		for( i = 0 ; i <  clean_output.length ; i++ )
		{
			knowledge_base.pushEntry( clean_output[i].trim() );
			//console.log( "--> " + clean_output[i].trim() );
		}
		
		
		
		
		knowledge_base.pushEntry( "int __xstat64(int ver, const char * path, struct stat64 * stat_buf)" );
		knowledge_base.pushEntry( "int __lxstat64(int ver, const char * path, struct stat64 * stat_buf)");
		knowledge_base.pushEntry( "int __fxstat64(int ver, int fildes, struct stat64 * stat_buf)" );


		/* Here we process functions in blocks of decreasing size
		 * in order to speedup the compilation process
		 * as we expect our parsing to be very reliable =) */
		knowledge_base.checkCompilation( function()
		{
			knowledge_base.checkCompilation(function()
			{
				knowledge_base.checkCompilation(function()
				{
					/*************************************************/
					/* Final output generation                       */
					/*************************************************/

					/* First make sure that entries are unique by name */
					var i,j;
					var unique_entries = [];
					
					for( i = 0 ; i < knowledge_base.entries.length ; i++ )
					{
						if( knowledge_base.entries[i].name == undefined )
						{
							continue;
						}
						
						var is_dup = 0;
						
						for( j = 0 ; j < knowledge_base.entries.length ; j++ )
						{
							if( knowledge_base.entries[j].name == undefined )
							{
								continue;
							}
								
							if( j == i )
								continue;
							
							if( knowledge_base.entries[j].name.trim() == knowledge_base.entries[i].name.trim() )
							{
								is_dup = 1; 
								break;
							}
						}
						
						if( !is_dup )
						{
							unique_entries.push( knowledge_base.entries[i] );
						}
						
					}
					
					knowledge_base.entries = [];
					knowledge_base.entries = unique_entries;

					/* And now proceed with script generation */

					var sc = new script_parser();

					var cnt = 0;
					
					for( i = 0 ; i < knowledge_base.entries.length ; i++ )
					{
						if( knowledge_base.entries[i].checked == true )
							cnt++;
					}

					console.log("\033[1;31mSucessfully wrapped " + cnt + " posix functions !\033[0m");

					for( i = 0 ; i < script_lines.length ; i++ )
					{
						sc.parseline( script_lines[i] );
					}

					sc.gen(knowledge_base.entries);

					//console.log( sc.output );
					try
					{
						fs.writeFileSync(output_file, sc.output);
						fs.writeFileSync("kb.json", JSON.stringify(knowledge_base.entries));
					}
					catch(e)
					{
						console.log("Failed to write output program");
						console.log(e);
						process.exit(1);
					}
				}, 1 );
			}, 3);
		}, 25);
		
		
	});
}

/*************************************************/
/* Knowledge base generation                     */
/*************************************************/


generateKB();




